{********************************************************************}
{E_Stats.pas
 This program copyright 1996
 Written by Peter E. Midford, with extensive copying from PDTREE.PAS
 for Theodore Garland, Jr.
 Last modified 13 May 1996.}
{********************************************************************}

{This unit contains all the statistical routines for PDERROR.}

Unit e_stats;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

Uses e_utils; {type definitions, constants, and utility functions.}

Type bLConArray = array [1..maxTips,1..2] of extended; {Array of contrasts for each trait}

Procedure computeTipStats(simData : nodeDataPtr;
                          Var tipStats : statPak);

Procedure computeStatistics(tree :nodePtr;
                            Var stats :statPak;
                            Var correlations :corrArray);

Procedure correlationScore(tree :nodePtr;
                           cost :costType;
                           Var trait1Cor, trait2Cor :extended);
                           Var workingStats, stats :statPak;

Procedure calcBLContrasts(tree1,tree2 :nodePtr;
                          Var bLCons :bLConArray;
                          Var cCount :integer);

Procedure bLContrastStats(bLCons :bLConArray;
                          size :integer;
                          tree1,tree2 :nodePtr;
                          t1xform,t2xform :extended;
                          t1xformtype,t2xformtype :char;
                          Var stpk :statPak;
                          Var correlations :corrArray);

Implementation

{**************************************************************************}
{This procedure calculates tip means, after any transforms have been
 performed.}
{**************************************************************************}
Procedure computeTipStats(simData : nodeDataPtr;
                          Var tipStats : statPak);
   Var
      curNode : nodeDataPtr;
   Begin
      tipStats.number := 0;
      tipStats.n1 := 1;
      tipStats.n2 := 2;
      tipStats.mean[1] := 0;
      tipStats.mean[2] := 0;
      While simData <> NIL Do Begin
         inc(tipStats.number);
         tipStats.mean[1] := tipStats.mean[1] + simData^.data[1];
         tipStats.mean[2] := tipStats.mean[2] + simData^.data[2];
         simData := simData^.next^.next;
      End; {while}
      tipStats.mean[1] := tipStats.mean[1]/tipStats.number;
      tipStats.mean[2] := tipStats.mean[2]/tipStats.number;
   End; {procedure computeTipStats}


{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
 It then fills in the height of the node and goes further up the tree until
 all the nodes have their height put in the temp field.}
{**************************************************************************}
Procedure tempHeight(nd :nodePtr; ht :extended);
   Begin
      nd^.temp := ht + nd^.length;
      If (nd^.sibL<>nil) Then Begin
         tempheight(nd^.sibL,nd^.temp);
         tempheight(nd^.sibR,nd^.temp)
      End {then}
   End; {procedure tempheight}

{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
 It then fills in the height of the node and goes further up the tree until
 all the nodes have their height put in the temp field.}
{**************************************************************************}
Procedure temp2Height(nd :nodePtr; ht :extended);
   Begin
      nd^.temp2 := ht+nd^.length;
      If (nd^.sibL<>nil) Then Begin
         temp2height(nd^.sibL,nd^.temp2);
         temp2height(nd^.sibR,nd^.temp2)
      End {then}
   End; {procedure temp2height}

{*************************************************************************}
{The following two procedures are like the preceeding height procedureds,
 except that they fill in "contrast corrected" heights, rather than simple
 heights based on lengths. These procedures assume that computeContrasts has
 been called to fill in the variance fields.
{*************************************************************************}

{*************************************************************************}
{The corrected length version of tempHeight.  Like tempHeight, the sum of
 adjusted lengths of nodes leading back to the root is assigned to the
 node's temp field.}
{*************************************************************************}
Procedure tempCorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Begin
      If nd <> rt
      Then nd^.temp := ht+nd^.variance
      Else nd^.temp := 0;
      If nd^.sibL<>nil Then Begin
         tempCorrectedHeight(nd^.sibL,rt,nd^.temp);
         tempCorrectedHeight(nd^.sibR,rt,nd^.temp);
      End; {then}
   End; {procedure tempAdjustedHeight}

{*************************************************************************}
{The corrected length version of tempAveHeight.  Like
 tempAve2Height, the average of the heights of the node's two daughters
 is assigned to the node's temp2 field.}
{*************************************************************************}
Procedure temp2CorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Begin
      If nd <> rt
      Then nd^.temp2 := ht+nd^.variance
      Else nd^.temp2 := 0;
      If nd^.sibL<>nil Then Begin
         temp2CorrectedHeight(nd^.sibL,rt,nd^.temp2);
         temp2CorrectedHeight(nd^.sibR,rt,nd^.temp2);
      End; {then}
   End; {procedure temp2CorrectedHeight}

{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
 It then fills in the temp field with the average height of the siblings of
 this node.  If the siblings of this node has siblings, then it recurses up
 the tree.}
{**************************************************************************}
Procedure tempAveHeight(nd :nodePtr; ht :extended);
   Begin
      ht := ht+nd^.length;
      nd^.temp := ht+(nd^.sibL^.length+nd^.sibR^.length)/2;
      If (nd^.sibL^.sibL<>nil) Then tempAveheight(nd^.sibL,ht);
      If (nd^.sibR^.sibR<>nil) Then tempAveheight(nd^.sibR,ht);
   End; {procedure tempaveheight}

{*************************************************************************}
{This procedure computes the standardized contrasts for each internal node.
 It also computes the predicted trait values for those nodes by Felsenstein's
 method.  It stores these values in the tree. Taken from PDTREE v2.0}
{*************************************************************************}
Procedure computeCon(nd :nodePtr);
   Var
      v1,v2 :extended;
   Begin
      If (nd^.sibL=nil) Then nd^.variance := nd^.length
      Else Begin
         computeCon(nd^.sibL);
         computeCon(nd^.sibR);
         v1 := nd^.sibL^.variance;
         v2 := nd^.sibR^.variance; 
         nd^.variance := nd^.length+(v1*v2/(v1+v2));
         nd^.sdcon := sqrt(v1+v2);
         nd^.contrast[1] := (nd^.sibL^.dat[1]-nd^.sibR^.dat[1])/nd^.sdcon;
         nd^.contrast[2] := (nd^.sibL^.dat[2]-nd^.sibR^.dat[2])/nd^.sdcon;
         nd^.dat[1]:=(v1*nd^.sibR^.dat[1] + v2*nd^.sibL^.dat[1])/(v1+v2);
         nd^.dat[2]:=(v1*nd^.sibR^.dat[2] + v2*nd^.sibL^.dat[2])/(v1+v2);
      End {else}
   End; {procedure computecon}

{**************************************************************************}
{This procedure calculates the statistics for the contrasts vs. contrasts
 plot.  It uses several recursive procedures to help calculate the stats just
 like in the procedures above.  The main difference is that it calculates
 the stats through the origin so it artificially sets the means to zero.} 
{**************************************************************************}
Procedure statCon(nd :nodePtr;Var stpk :statPak);
   Var sx2,sxy :extended;
  {************************************************************************}
  Procedure getNum(nd :nodePtr; Var n :integer);
     Begin
        If (nd^.sibL<>nil) Then Begin
           inc(n);
           getNum(nd^.sibL,n);
           getNum(nd^.sibR,n)
        End {then}
     End; {procedure getnum}
  {************************************************************************}
  Procedure getVar(nd :nodePtr; Var s :statPak);
     Begin
        If (nd^.sibL<>nil) Then Begin
           s.variance[1] := s.variance[1]+sqr(nd^.contrast[1]);
           s.variance[2] := s.variance[2]+sqr(nd^.contrast[2]);
           s.corr := s.corr+(nd^.contrast[1])*(nd^.contrast[2]);
           getVar(nd^.sibL,s);
           getVar(nd^.sibR,s)
        End {then}
     End; {procedure getvar}
  {************************************************************************}
  Procedure getNumCon(nd :nodePtr; Var s :statPak);
     Var t2 : integer;
     Begin
        t2 := s.n2;
        If nd^.sibL<>nil Then Begin
           If nd^.contrast[t2]>0 Then inc(s.conypos)
           Else
              If nd^.contrast[t2]<0 Then inc(s.conyneg)
              Else inc(s.cony0);
           getNumCon(nd^.sibL,s);
           getNumCon(nd^.sibR,s)
        End {then}
     End; {procedure getnumcon}
  {************************************************************************}
  Procedure getSums(Var sxy,sx2 :extended; nd :nodePtr;t1,t2 : integer);
     Begin
        If (nd^.sibL<>nil) Then Begin
           sxy := sxy+ nd^.contrast[t1]*nd^.contrast[t2];
           sx2 := sx2+ sqr(nd^.contrast[t1]);
           getSums(sxy,sx2,nd^.sibL,t1,t2);
           getSums(sxy,sx2,nd^.sibR,t1,t2)
        End {then}
     End; {procedure getsums}
  {************************************************************************}
  {This procedure deserves some mention.  It positivizes the contrasts.  If
  the contrast on the X axes is negative, then it filps its sign as well as
  the sign of the other contrast at that node.}
  {************************************************************************}
  Procedure positivise(nd :nodePtr; whichTrait : integer);
     Begin
        If (nd^.sibL<>nil) Then Begin
           If (nd^.contrast[whichTrait]<0) Then Begin
              nd^.contrast[1] := nd^.contrast[1]*-1;
              nd^.contrast[2] := nd^.contrast[2]*-1
           End; {then}
           positivise(nd^.sibL,whichTrait);
           positivise(nd^.sibR,whichTrait)
        End {then}
    End; {procedure positivise}
{**************************************************************************}
  Begin {procedure statcon}
    stpk.badStats := false;
    positivise(nd,stpk.n1);
    stpk.number := 0;
    stpk.mean[1] := 0;
    stpk.mean[2] := 0;
    getNum(nd,stpk.number);
    stpk.variance[1] := 0;
    stpk.variance[2] := 0;
    stpk.corr := 0;
    getVar(nd,stpk);
    stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
    stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
    stpk.corr := stpk.corr/(stpk.number-1);
    stpk.cov := stpk.corr;
    If ((stpk.variance[1]<>0) And (stpk.variance[2]<>0)) Then
      stpk.corr := stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
    Else stpk.corr := 0;
    sxy := 0;
    sx2 := 0;
    getSums(sxy,sx2,nd,stpk.n1,stpk.n2);
    If sx2 <> 0
    Then stpk.ls := sxy/sx2
    Else Begin
       stpk.ls := 0;
       stpk.badStats := true;
    End; {else}
    stpk.lsy := 0.0;
    finishStatOrg(stpk);
    stpk.df := stpk.df+1;
    stpk.conypos := 0;
    stpk.conyneg := 0;
    stpk.cony0 := 0;
    getNumCon(nd,stpk)
  End; {procedure statcon} 

{***************************************************************************}
{This procedure computes the statistics (Corr[1] and Corr[2] for a set of
 data.} 
{***************************************************************************}
Procedure statStdVsd(nd :nodePtr;t :integer;Var stpk :statPak);
  {************************************************************************}
  {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getMean(t :integer;nd :nodePtr;Var n :integer;Var m :traits);
     Begin
        If (nd^.sibL<>nil) Then Begin 
           inc(n);
           m[1] := m[1]+nd^.sdcon;
           m[2] := m[2]+abs(nd^.contrast[t]);
           getmean(t,nd^.sibL,n,m);
           getmean(t,nd^.sibR,n,m)
        End {then}
     End; {procedure getmean}
  {************************************************************************}
  {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getVar(t :integer;nd :nodePtr;Var s :statPak);
     Begin
        If (nd^.sibL<>nil) Then Begin
           s.variance[1] := s.variance[1]+sqr(s.mean[1]-nd^.sdcon);
           s.variance[2] := s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
           s.corr := s.corr+(s.mean[1]-nd^.sdcon)*(s.mean[2]-abs(nd^.contrast[t]));
           getVar(t,nd^.sibL,s);
           getVar(t,nd^.sibR,s)
        End {then}
     End; {procedure getvar}
  {***********************************************************************}
  Begin {procedure statstdvsd}
     stpk.number := 0;
     stpk.mean[1] := 0;
     stpk.mean[2] := 0;
     getMean(t,nd,stpk.number,stpk.mean);
     stpk.mean[1] := stpk.mean[1]/stpk.number;
     stpk.mean[2] := stpk.mean[2]/stpk.number;
     stpk.variance[1] := 0;
     stpk.variance[2] := 0;
     stpk.corr := 0;
     getVar(t,nd,stpk);
     stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
     stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
     stpk.corr := stpk.corr/(stpk.number-1);
     stpk.cov := stpk.corr;
     If ((stpk.variance[1]>0) And (stpk.variance[2]>0)) Then
        stpk.corr := stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
     Else stpk.corr := 0
  End; {procedure statstdvsd}

{***************************************************************************}
{This procedure computes the statistics (screen 3,4) for a set of data.} 
{***************************************************************************}
Procedure statNodeCon(nd :nodePtr;t :integer;Var stpk :statPak);
  {************************************************************************}
  {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getMean(t :integer;nd :nodePtr;Var n :integer;Var m :traits);
    Begin
      If (nd^.sibL<>nil) Then Begin
        n := n+1;
        m[1] := m[1]+nd^.dat[t];
        m[2] := m[2]+abs(nd^.contrast[t]);
        getMean(t,nd^.sibL,n,m);
        getMean(t,nd^.sibR,n,m)
      End {then}
    End; {procedure getmean}
  {************************************************************************}
  {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getVar(t :integer;nd :nodePtr;Var s :statPak);
    Begin
      If (nd^.sibL<>nil) Then Begin
        s.variance[1] := s.variance[1]+sqr(s.mean[1]-nd^.dat[t]);
        s.variance[2] := s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
        s.corr := s.corr+(s.mean[1]-nd^.dat[t])*(s.mean[2]-abs(nd^.contrast[t]));
        getvar(t,nd^.sibL,s);
        getvar(t,nd^.sibR,s)
      End {then}
    End; {procedure getvar}
{**************************************************************************}
  Begin {procedure statnodecon}
    stpk.number := 0;
    stpk.mean[1] := 0;
    stpk.mean[2] := 0;
    getMean(t,nd,stpk.number,stpk.mean);
    stpk.mean[1] := stpk.mean[1]/stpk.number;
    stpk.mean[2] := stpk.mean[2]/stpk.number;
    stpk.variance[1] := 0;
    stpk.variance[2] := 0;
    stpk.corr := 0;
    getVar(t,nd,stpk);
    stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
    stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
    stpk.corr := stpk.corr/(stpk.number-1);
    stpk.cov := stpk.corr;
    If ((stpk.variance[1]>0) and (stpk.variance[2]>0)) Then
       stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
    Else stpk.corr:=0
  End; {procedure statnodecon}

{***************************************************************************}
{This procedure computes the statistics (screen 5,6) for a set of data.} 
{***************************************************************************}
Procedure statHeightNode(nd :nodePtr;t :integer;Var stpk :statPak);
  {************************************************************************}
  {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
   Procedure getMean(t :integer;nd :nodePtr;Var n :integer;Var m :traits);
      Begin
         If (nd^.sibL<>nil) Then Begin
            inc(n);
            m[1] := m[1]+nd^.temp;
            m[2] := m[2]+nd^.dat[t];
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m)
         End {then}
      End; {procedure getmean}
  {************************************************************************}
  {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
  {************************************************************************}
   Procedure getVar(t :integer;nd :nodePtr;var s :statPak);
      Begin     
         If (nd^.sibL<>nil) Then Begin
            s.variance[1] := s.variance[1]+sqr(s.mean[1]-nd^.temp);
            s.variance[2] := s.variance[2]+sqr(s.mean[2]-nd^.dat[t]);
            s.corr := s.corr+(s.mean[1]-nd^.temp)*(s.mean[2]-nd^.dat[t]);
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s)
         End {then}
      End; {procedure getvar}
   {*************************************************************************}
   Begin {procedure statheightnode}
      stpk.number := 0;
      stpk.mean[1] := 0;
      stpk.mean[2] := 0;
      getMean(t,nd,stpk.number,stpk.mean);
      stpk.mean[1] := stpk.mean[1]/stpk.number;
      stpk.mean[2] := stpk.mean[2]/stpk.number;
      stpk.variance[1] := 0;
      stpk.variance[2] := 0;
      stpk.corr := 0;
      getvar(t,nd,stpk);
      stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
      stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
      stpk.corr := stpk.corr/(stpk.number-1);
      stpk.cov := stpk.corr;
      If ((stpk.variance[1]>0) And (stpk.variance[2]>0))
      Then stpk.corr := stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr := 0
   End; {procedure statheightnode}    

{***************************************************************************}
{This procedure computes the statistics (Screen 7,8) for a set of data.} 
{***************************************************************************}
Procedure statHeightCon(nd :nodePtr;t :integer;Var stpk :statPak);
  {************************************************************************}
  {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(t :integer;nd :nodePtr;Var n :integer;Var m :traits);
      Begin
         If (nd^.sibL<>nil) Then Begin
            inc(n);
            m[1] := m[1]+nd^.temp;
            m[2] := m[2]+abs(nd^.contrast[t]);
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m)
         End {then}
      End; {procedure getMean}
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
    covariance for the data.  It finds the sums of the deviations.
    Notice that it is local to a parent procedure, and therefore, can coexist
    in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t :integer;nd :nodePtr;Var s :statPak);
      Begin
         If (nd^.sibL<>nil) Then Begin
            s.variance[1] := s.variance[1]+sqr(s.mean[1]-nd^.temp);
            s.variance[2] := s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
            s.corr := s.corr+(s.mean[1]-nd^.temp)*(s.mean[2]-abs(nd^.contrast[t]));
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s)
         End {then}
      End; {procedure getvar}
   {*************************************************************************}
   Begin {procedure statheightcon}
      stpk.number := 0;
      stpk.mean[1] := 0;
      stpk.mean[2] := 0;
      getMean(t,nd,stpk.number,stpk.mean);
      stpk.Mean[1] := stpk.mean[1]/stpk.number;
      stpk.mean[2] := stpk.mean[2]/stpk.number;
      stpk.variance[1] := 0;
      stpk.variance[2] := 0;
      stpk.corr := 0;
      getVar(t,nd,stpk);
      stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
      stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
      stpk.corr := stpk.corr/(stpk.number-1);
      stpk.cov := stpk.corr;
      If ((stpk.variance[1]>0) And (stpk.variance[2]>0))
      Then stpk.corr := stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0
   End; {procedure statheightcon}

{**************************************************************************}
{This recursive procedure is passed a node, and the least squares parameters.
 It then fills in the temp field with the residuals.  Stolen from PDTREE
 v2.0.}
{**************************************************************************}
Procedure tempAbsResidual(nd :nodePtr; m,y0 :extended;t,t2 :integer);
   Begin
      If (nd^.sibL<>nil) Then Begin
         nd^.temp := abs(nd^.contrast[t2]-(m*nd^.contrast[t]+y0));
         tempAbsResidual(nd^.sibL,m,y0,t,t2);
         tempAbsResidual(nd^.sibR,m,y0,t,t2)
      End {then}
   End; {procedure tempresidual}
{**************************************************************************}
{This recursive procedure is passed a node, and the least squares parameters.
 It then fills in the temp field with the residuals.  Stolen from PDTREE
 v2.0.}
{**************************************************************************}
Procedure tempAbs2TreeResidual(tree1,tree2 :nodePtr; m,y0 :extended;t,t2 :integer);
   Begin
      If (tree1^.sibL<>nil) Then Begin
         tree1^.temp := abs(tree2^.contrast[t2]-(m*tree1^.contrast[t]+y0));
         tempAbs2TreeResidual(tree1^.sibL,tree2^.sibL,m,y0,t,t2);
         tempAbs2TreeResidual(tree1^.sibR,tree2^.sibR,m,y0,t,t2)
      End {then}
   End; {procedure tempresidual}
{**************************************************************************}
{This procedure computes the values returned by the PlotCon procedure in
 PDTREE v 2.0.  It does no file or graphics output.}
{**************************************************************************}
  Procedure noPlotCon(nd:nodeptr;var m,y0:extended;t,t2:integer);
  Var stpk:statpak;
  Begin
    stpk.n1 := t;
    stpk.n2 := t2;
    statcon(nd,stpk);
    m := stpk.ls;
    y0 := stpk.lsy;
  End; {procedure noplotcon}
{**************************************************************************}
{This procedure computes the values returned by the PlotCon procedure in
 PDTREE v 2.0.  It does no file or graphics output.}
{**************************************************************************}
Procedure noPlot2TreeCon(tree1,tree2 :nodeptr;var m,y0:extended;t,t2:integer);
  Var sxy,sx2 :extended;
      stpk1,stpk2 :statpak;
  {************************************************************************}
  Procedure getSums(Var sxy,sx2 :extended; tree1,tree2 :nodePtr);
     Begin
        If (tree1^.sibL<>nil) Then Begin
           sxy := sxy + tree1^.contrast[1]*tree2^.contrast[2];
           sx2 := sx2 + sqr(tree1^.contrast[1]);
           getSums(sxy,sx2,tree1^.sibL,tree2^.sibL);
           getSums(sxy,sx2,tree1^.sibR,tree2^.sibR)
        End {then}
     End; {procedure getsums}
  {************************************************************************}
  Begin {procedure noplot2Treecon}
    stpk1.n1 := t;
    stpk1.n2 := t2;
    stpk2.n1 := t;
    stpk2.n2 := t2;
    statcon(tree1,stpk1);
    statcon(tree2,stpk2);
    sxy := 0;
    sx2 := 0;
    getSums(sxy,sx2,tree1,tree2);
    If (sx2 = 0)
    Then m := -9E99
    Else m := sxy/sx2;
    y0 := 0.0;
  End; {procedure noplot2Treecon}
{***************************************************************************}
{This procedure computes the statistics (Correlation 11) for a set of data. }
{***************************************************************************}
Procedure statResidualHeight(nd :nodePtr;Var stpk :statPak);
  {************************************************************************}
  {This recursive procedure does the first step in finding the mean values
  for the data.  It finds the sums and the number of data points.
  Notice that it is local to a parent procedure, and therefore, can coexist
  in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getMean(nd :nodePtr;Var n :integer;Var m :traits);
     Begin
        If (nd^.sibL<>nil) Then Begin 
           n := n+1;
           m[1] := m[1]+nd^.temp2;
           m[2] := m[2]+nd^.temp;
           getMean(nd^.sibL,n,m);
           getMean(nd^.sibR,n,m)
        End {then}
     End; {procedure getmean}
  {************************************************************************}
  {This recursive procedure does the first step in finding the variance and
  covariance for the data.  It finds the sums of the deviations.
  Notice that it is local to a parent procedure, and therefore, can coexist
  in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getVar(nd :nodePtr;Var s :statPak);
     Begin
        If (nd^.sibL<>nil) Then Begin
           s.variance[1] := s.variance[1]+sqr(s.mean[1]-nd^.temp2);
           s.variance[2] := s.variance[2]+sqr(s.mean[2]-nd^.temp);
           s.corr := s.corr+(s.mean[1]-nd^.temp2)*(s.mean[2]-nd^.temp);
           getVar(nd^.sibL,s);
           getVar(nd^.sibR,s)
        End {then}
     End; {procedure getvar}
{**************************************************************************}
   Begin {procedure statresidualheight}
      stpk.number := 0;
      stpk.mean[1] := 0;
      stpk.mean[2] := 0;
      getMean(nd,stpk.number,stpk.mean);
      stpk.mean[1] := stpk.mean[1]/stpk.number;
      stpk.mean[2] := stpk.mean[2]/stpk.number;
      stpk.variance[1] := 0;
      stpk.variance[2] := 0;
      stpk.corr := 0;
      getVar(nd,stpk);
      stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
      stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
      stpk.corr := stpk.corr/(stpk.number-1);
      stpk.cov := stpk.corr;
      If ((stpk.variance[1]>0) and (stpk.variance[2]>0)) Then
         stpk.corr := stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr := 0
   End; {procedure statresidualheight}

{*******************************************************************}
{This procedure computes all the statistics for the tree passed in
 tree.  Stats contains the "standard" statistics, while correlations
 contains the various correlations in columns 11-20.}
{*******************************************************************}
Procedure computeStatistics(tree :nodePtr;
                            Var stats :statPak;
                            Var correlations :corrArray);
  Var workingStats :statPak;
      i :integer;
      m,y0 :extended;         {Slope and intercept}
      t,t2 :integer;
  Begin
    stats.badStats := false;
    workingStats.badStats := false;
    stats.n1 := 1;        {probably the correct assumptions to start with}
    stats.n2 := 2;
    computeCon(tree);     {Fill the tree with contrast values}
    statCon(tree,stats);  {Compute the "standard" contrast statistics}
    For i := 1 To 11 Do correlations[i] := 0.0;  {Initialize correlations}
    tempheight(tree,0);
    statStdvsd(tree,1,workingStats);
    correlations[1] := workingStats.corr;
    statStdvsd(tree,2,workingStats);
    correlations[2] := workingStats.corr;
    statNodeCon(tree,1,workingStats);
    correlations[3] := workingStats.corr;
    statNodeCon(tree,2,workingStats);
    correlations[4] := workingStats.corr;
    tempCorrectedHeight(tree,tree,0);
    statHeightCon(tree,1,workingStats);
    correlations[5] := workingStats.corr;
    statHeightCon(tree,2,workingStats);
    correlations[6] := workingStats.corr;
    statHeightNode(tree,1,workingStats);
    correlations[7] := workingStats.corr;
    statHeightNode(tree,2,workingStats);
    correlations[8] := workingStats.corr;
    noPlotCon(tree,m,y0,1,2);                {will we want to switch vars}
    tempAbsResidual(tree,m,y0,1,2);          {in some cases?}
    temp2CorrectedHeight(tree,tree,0);
    statResidualHeight(tree,workingStats);
    correlations[11] := workingStats.corr;
    If workingStats.badStats Then stats.badStats := true;
  End;{procedure computeStatistics}


{*******************************************************************}
{This procedure calculates separate correlation scores for each of
the two traits represented on a tree.}
{*******************************************************************}
Procedure correlationScore(tree :nodePtr;
                           cost :costType;
                           Var trait1Cor, trait2Cor :extended);
   Var workingStats, stats :statPak;
       m,y0 : extended;
   Begin
      workingStats.n1 := 1;
      workingStats.n2 := 2;
      stats.n1 := 1;
      stats.n2 := 2;
      computeCon(tree);
      statCon(tree,stats);
      tempheight(tree,0);
      If (cost = SdcBL) Then Begin
         statStdvsd(tree,1,workingStats);
         trait1Cor := workingStats.corr;
         statStdvsd(tree,2,workingStats);
         trait2Cor := workingStats.corr
      End; {then}
      If (cost = SdcNV) Then Begin
         statNodeCon(tree,1,workingStats);
         trait1Cor := workingStats.corr;
         statNodeCon(tree,2,workingStats);
         trait2Cor := workingStats.corr;
      End; {then}
      If (cost = SdcHt) Then Begin
         tempCorrectedHeight(tree,tree,0);
         statHeightCon(tree,1,workingStats);
         trait1Cor := workingStats.corr;
         statHeightCon(tree,2,workingStats);
         trait2Cor := workingStats.corr;
      End; {then}
      If (cost = NvHt) Then Begin
         tempCorrectedHeight(tree,tree,0);
         statHeightNode(tree,1,workingStats);
         trait1Cor := workingStats.corr;
         statHeightNode(tree,2,workingStats);
         trait2Cor := workingStats.corr;
      End; {then}
         
 {     If (cost = ResXH) Then Begin
          noPlotCon(tree,m,y0,1,2);        
          tempAbsResidual(tree,m,y0,1,2);
          temp2CorrectedHeight(tree,0);
          statResidualHeight(tree,workingStats);
          trait1Cor := workingStats.corr;
          noPlotCon(tree,m,y0,2,1);
          tempAbsResidual(tree,m,y0,2,1);  
          temp2CorrectedHeight(tree,0);
          statResidualHeight(tree,workingStats);
          trait2Cor := workingStats.corr;
       End;} {then}
   End; {procedure correlationScore}

{*******************************************************************}
{This procedure takes two isomorphic trees and collects the contrast
values into an n-by-2 array.}
{*******************************************************************}
Procedure collectContrasts(tree1,tree2 :nodePtr;
                           Var bLCons :bLConArray;
                           Var index :integer);
   Begin
      If (tree1^.sibL<>nil) Then Begin
         collectContrasts(tree1^.sibL,tree2^.sibL,bLCons,index);
         collectContrasts(tree1^.sibR,tree2^.sibR,bLCons,index);
         index := index + 1;
         bLCons[index,1] := tree1^.contrast[1];
         bLCons[index,2] := tree2^.contrast[2]
      End {then}
   End; {procedure collectContrasts}

{*******************************************************************}
{This procedure computes contrasts for two trees and collects then
contrast values into an array. }
{*******************************************************************}
Procedure calcBLContrasts(tree1,tree2 :nodePtr;
                          Var bLCons :bLConArray;
                          Var cCount :integer);
   Begin
     computeCon(Tree1);
     computeCon(Tree2);
     collectContrasts(tree1,tree2,bLCons,cCount);
   End; { procedure calcBLContrasts}

{*******************************************************************}
{This procedure calculates the statistics for the contrast array.  This
procedure is laid out more or less like the other ones in T_Felsen.

getbounds determines the max and min.

getmean summs up each variable and counts them.  From this one can
  easily deterime the mean.

getvar summs up the variance and covaraince deviations from which the
  variance and covariance can easily be determined.

getsumms gets the summs needed for the least squares regression.
  Here regressions are not done through the origin.}
{*******************************************************************}
Procedure bLContrastStats(bLCons :bLConArray;
                          size :integer;
                          tree1,tree2 :nodePtr;
                          t1xform,t2xform :extended;
                          t1xformtype,t2xformtype :char;
                          Var stpk :statPak;
                          Var correlations :corrArray);
  Var
     sy,sx,sx2,sy2,sxy,d :extended;
     rt :nodePtr;
     dum :integer;
     workingstats :statPak;
     i :integer;
     m,y0 :extended;         {Slope and intercept}
     t,t2 :integer;

  {*******************************************************************}
  {This computes the maximum and minimum contrast values for each trait
  from the branch length contrast array.}
  {*******************************************************************}
  Procedure getBounds(bLCons :bLConArray; size :integer; Var min,max :traits);
    Var i :integer;
    Begin
       For i := 1 To size Do Begin
           If (bLCons[i,1]>max[1]) Then max[1] := BLCons[i,1];
           If (bLCons[i,1]<min[1]) Then min[1] := BLCons[i,1];
           If (bLCons[i,2]>max[2]) Then max[2] := BLCons[i,2];
           If (bLCons[i,2]<min[2]) Then min[2] := BLCons[i,2]
       End {for}
    End; {procedure getbounds}
  {*******************************************************************}
  {This computes the variance of and correlations between contrasts
  stored in the branch length contrast array.}
  {*******************************************************************}
  Procedure getVar(bLCons :bLConArray; size :integer; Var s :statpak);
     Var i :integer;
     Begin
        For i := 1 To size Do Begin
           s.variance[1] := s.variance[1] + sqr(bLCons[i,1]);
           s.variance[2] := s.variance[2] + sqr(bLCons[i,2]);
           s.corr := s.corr + (bLCons[i,1] * bLCons[i,2])
        End {for}
     End; {procedure getvar}
  {*******************************************************************}
  {This computes the sums and sums-of-squares of contrasts stored in
  the branch length contrast array.}
  {*******************************************************************}
  Procedure getSumms(bLCons :bLConArray;
                     size :integer;
                     Var sx,sy,sx2,sy2,sxy :extended);
     Var i :integer;
     Begin
        For i := 1 to size Do Begin
           sx := sx + BLCons[i,1];
           sx2 := sx2 + sqr(BLCons[i,1]);
           sy2 := sy2 + sqr(BLCons[i,2]);
           sy := sy + BLCons[i,2];
           sxy := sxy + BLCons[i,1] * BLCons[i,2]
        End {for}
     End; {procedure getsumms}  
  {*******************************************************************}
  {This procedure counts the number of contrasts in trait 2 greater,
  equal, or less than zero.  It uses the contrasts stored in the branch
  length contrast array.}
  {*******************************************************************}
  Procedure getNumCon(BLCons :BLConArray; size :integer; var s :statPak);
     Var i :integer;
     Begin
        For i := 1 to size Do
           If (BLCons[i,1]*BLCons[i,2]>0) Then
              s.conypos := s.conypos+1
             Else If (BLCons[i,1]*BLCons[i,2]<0) Then
                s.conyneg := s.conyneg+1
               Else                     {one or both is 0}
                 If BLCons[i,2] = 0 Then s.cony0 := s.cony0+1
                 Else If BLCons[i,2]<0 Then s.conyneg := s.conyneg+1
                      Else s.conypos := s.conypos+1;
    End; {procedure getnumcon}
  {********************************************************************}
  {This calculates the rma, ma ,r squared called r2, t statistic
  called ts,the f statistic, and the degrees of freedom called df.
  Both rma and ma are forced through the origin, so rmay and may are 0
  Note: This version, stolen from the tips module of pdtree, has been
  drastically simplified.  If necessary, use the original as a template
  for new code.
  {********************************************************************}
  Procedure finishStatOrg(Var stpk :statpak);
     Begin
        If stpk.corr <> 0
        Then stpk.rma := stpk.ls/stpk.corr
        Else stpk.rma := 0;
        If (((stpk.corr>0) and (stpk.rma<0)) or
            ((stpk.corr<0) and (stpk.rma>0))) Then
           stpk.rma := stpk.rma*-1;
        stpk.rmay:=0.0;

        stpk.ma := sqr(stpk.variance[1]+stpk.variance[2]) -
             4*(stpk.variance[2]*stpk.variance[1] - sqr(stpk.cov));
        stpk.ma := 0.5*(stpk.variance[2]+stpk.variance[1]+
             sqrt(stpk.ma));
        If (stpk.ma = stpk.variance[2]) Then
           stpk.ma := -9.9999999999E-9
        Else
           stpk.ma := stpk.cov/(stpk.ma-stpk.variance[2]);
        stpk.may := 0.0;

        stpk.r2 := sqr(stpk.corr);


      If (stpk.r2 >= 1.0) Then Begin
         stpk.ts := -9.9999999999E-9;
         stpk.f := -9.9999999999E-9
      End
      Else Begin
         stpk.ts:=stpk.corr*sqrt((stpk.number-1)/(1-stpk.r2));
         stpk.f:=sqr(stpk.ts)
      End;
      stpk.df := stpk.number-2
      End; {procedure finishstatorg}
  {*************************************************************************}
  Begin {procedure bLContrastStats }
     stpk.badStats := false;
     stpk.n1 := 1;         {These may change, but at least they're}
     stpk.n2 := 2;         {initialized}
     workingStats.n1 := 1;
     workingStats.n2 := 2;
     stpk.max[1] := BLCons[1,1];
     stpk.min[1] := stpk.max[1];
     stpk.max[2] := BLCons[1,2];
     stpk.min[2] := stpk.max[2];
     getBounds(BLCons,size,stpk.min,stpk.max);
     stpk.mean[1] := 0;
     stpk.mean[2] := 0;
     stpk.number := size;
     stpk.variance[1] := 0;
     stpk.variance[2] := 0;
     stpk.corr := 0;
     getVar(BLCons,size,stpk);
     stpk.variance[1] := stpk.variance[1]/(stpk.number-1);   
     stpk.variance[2] := stpk.variance[2]/(stpk.number-1);  
     stpk.cov := stpk.corr/(stpk.number-1);
     sx := 0;
     sy := 0;
     sx2 := 0;
     sy2 := 0;
     sxy := 0;
     getSumms(BLCons,size,sx,sy,sx2,sy2,sxy);
     if (sx2*sy2 > 0)
     Then stpk.corr := stpk.corr/sqrt(sx2*sy2)
     Else stpk.corr := 0;
     If (sx*sx2) = 0
     Then Begin
        d := 0;
        If (sx = 0)
        Then Begin
           stpk.ls := -9E99;
           stpk.badStats := true;
           writeln('No variance in trait 1, most stats set to -9');
        End
        Else stpk.ls := sxy/sx2;
     End
     Else Begin
        d := 1/(stpk.number*sx2-sqr(sx));
        stpk.ls := sxy/sx2;
     End; {else}
     stpk.lsy := 0.0;
     If not(stpk.badStats)
     Then finishStatOrg(stpk);
     stpk.conypos:=0;
     stpk.conyneg:=0;
     stpk.cony0:=0;
     getNumCon(BLCons,size,stpk);
     For i := 1 to 11 Do correlations[i] := 0.0;
     tempHeight(tree1,0);
     tempHeight(tree2,0);
     statStdvsd(tree1,1,workingStats);
     correlations[1] := workingStats.corr;
     statStdvsd(tree2,2,workingStats);
     correlations[2] := workingStats.corr;
     statNodeCon(tree1,1,workingStats);
     correlations[3] := workingStats.corr;
     statNodeCon(tree2,2,workingStats);
     correlations[4] := workingStats.corr;
     tempCorrectedHeight(tree1,tree1,0);
     tempCorrectedHeight(tree2,tree2,0);
     statHeightCon(tree1,1,workingStats);
     correlations[5] := workingStats.corr;
     statHeightCon(tree2,2,workingStats);
     correlations[6] := workingStats.corr;
     statHeightNode(tree1,1,workingStats);
     correlations[7] := workingStats.corr;
     statHeightNode(tree2,2,workingStats);
     correlations[8] := workingStats.corr;
     noPlot2TreeCon(tree1,tree2,m,y0,1,2);         {this may have to change!}
     tempAbs2TreeResidual(tree1,tree2,m,y0,1,2);
     temp2CorrectedHeight(tree1,tree1,0);
     statResidualHeight(tree1,workingstats);
     correlations[11] := workingStats.corr
   End; {procedure BlContrastStats}

End. {Unit e_stats}