{***********************************************************************}
{This unit contains definitions specific for reading Nexus files.  Trees}
{are represented using a "light weight" node structure because this }
{application should be used for contrast calculations.  It also defines}
{structures for data and taxon matrices.  These are all stored in a}
{nexusStruct, which can hold multiple trees, taxa sets and data sets.}
{The structure allows the program to determine which combinations of taxa,}
{trees and characters are compatible for writing to a PDI file.}
{***********************************************************************}

Unit n_util1;

Interface

Uses crt;  {clrscr and readkey for UI}

Const
   UNINITIALIZED = -9999;   {TG style code for bogus value}
   NNAMELENGTH   = 255;     {The length of nexus names - this may need}
                            {tuning.}
   TAXACOUNT     = 936;     {set to 26*36; may need to shrink}

Type
   flstring	 = string[30];           {maybe expand this}
   typstring	 = string[3];
   pdinamestr    = string[2];            {the two character PDI namestring}
   nnamestr	 = string[NNAMELENGTH];  {allow full length names}
   traits	 = array[1..2] of extended;
   traitsPtr	 = ^traits;
				 
   nnodePtr    = ^nnode;   {pointer for new lightweight nexus node}
   
   nnode = Record
	      Ancestor : nnodeptr;
	      dat      : traits;     {Tip data}
	      SibL     : nnodeptr;   {Children of node}
	      SibR     : nnodeptr;
	      tipL     : nnodeptr;   {Pointer to nodes within tree.  Originally
				      tip to tip.}           
	      tipR     : nnodeptr;
	      length   : extended;
              pdiname  : pdinamestr; {short, 2 char name}
              tipnum   : integer;    {position of tip, not the same as index}
	      index    : integer;    {Number assigned to node.  First N numbers are tips
				    (left to right) next N-1 are interior nodes.}
	      
	   End; {Record nnode}

   {Currently a taxon is simply represented by its name}
   taxonPtr = ^taxon;
   taxon = Record
              name : string;
           End; {record taxon}

   {Would be nice to do something more flexible about taxablocks, instead
    of this array. }
   taxaBlockPtr = ^taxaBlock;
   taxaBlock = Record
                  taxa : Array [1..TAXACOUNT] of taxonPtr;
                  ntaxa : integer;
                  limit : integer;
                  next : taxaBlockPtr;
               End; {record taxaBlock}


   {Character (and data) matrices frequently store categorical data as
    packed strings that hold a row of values.  charMatrixRow is a
    container for the uninterpreted string and a pointer to the taxon
    associated with the data.}
   charMatrixRowPtr = ^charMatrixRow;
   charMatrixRow = Record
                      myTaxon : taxonPtr;
                      dataString : string;  {raw data}
                   End; {record charMatrixRow}

   {These enumerates define possible ways of parsing the strings that
    contain the data for a row of the character matrix. }
   itemType = (states);
   statesType = (statesPresent);
   dataTypeType = (standard,continuous);  {all we need for now}

   charBlockPtr = ^characterBlock;
   characterBlock = Record
                       taxa : taxaBlockPtr;
                       matrix : Array [1..TAXACOUNT] of charMatrixRowPtr;
                       nchar : integer;
                       items : itemType;
                       statesFormat : statesType;
                       dataTypeSpec : dataTypeType;
                       symbols : string;
                       mapDigits : boolean; {true to interpret single digits as integer values}
                       zeroBased : boolean;
                       next : charBlockPtr;
                    End; {characterBlock}

   treeEntryPtr = ^treeEntry;
   treeEntry = Record
                  name   : nnamestr; {name of the tree}
                  tree   : nnodePtr; {the tree}
                  next   : treeEntryPtr; {next tree}
                  rooted : boolean; {false if tree marked as unrooted}
               End; {treeEntry}

   treeBlockPtr = ^treeBlock;
   treeBlock = Record
                  name  : nnamestr;        {name of the ?block}
                  trees : treeEntryPtr;    {the trees in the block}
                  next  : treeBlockPtr;
               End; {treeList}

   translatePairPtr = ^translatePair;
   translatePair = Record
                      index : integer;
                      curtaxa : taxaBlockPtr;
                      taxonindex : integer;
                      charBlock : charBlockPtr;
                      next : translatePairPtr;
                   End; {type translatePair}

   translateBlockPtr = ^translateBlock;
   translateBlock = Record
                         index   : integer;
                         tlength : integer;
                         myData  : charBlockPtr;
                         data    : translatePairPtr;
                         next    : translateBlockPtr;
                      End; {type translationEntry}


   {This type holds the entire content of the nexus file.}
   nexusPtr      = ^nexusRecord;
   nexusRecord = Record
                   trees :  treeBlockPtr;
                   translations : translateBlockPtr;
                   taxa : taxaBlockPtr;
                   characters : charBlockPtr;
                   {selectedTree : treeEntryPtr; }
                   selectedTreeBlock : treeBlockPtr;
                   selectedTranslateBlock : translateBlockPtr;
                   selectedCharBlock : charBlockPtr;
                 End; {nexusRecord}

   Procedure pdnexus_error(message:string);

   Procedure pdnexus_warning(message:string);

   Procedure init_errors;

   Procedure nexusRecord_dispose(Var toDie : nexusPtr);

   Function nexusRecord_new : nexusPtr;

   Function taxaBlock_new : taxaBlockPtr;

   Function charBlock_new : charBlockPtr;

   Function dataBlock_addTaxon(block : charBlockPtr;
                               name : string) : integer;

   Procedure dataBlock_addCharData(block : charBlockPtr;
                                   index : integer;
                                   data : string);

   Function taxaBlock_addTaxon(block : taxaBlockPtr;
                               name : string) : integer;

   Function translateBlock_new : translateBlockPtr;

   Procedure updateTranslateBlocks(blockList : translateBlockPtr; 
                                   taxa : taxaBlockPtr);

   Function treeBlock_new : treeBlockPtr;

   Procedure treeBlock_addTree(block : treeBlockPtr;
                               name : nNameStr;
                               rooted : boolean;
                               Var result : treeEntryPtr);

   Function lookupTaxon(block : taxaBlockPtr; name : String) : integer;

   Function nnode_new : nnodeptr;

   Procedure extendWrite(Var outFile :text;num :extended);

   Procedure CheckPDIName(nod:nnodeptr;
                          name:pdinamestr;
                          Var notfound:boolean;
                          Var want:nnodeptr);

   Procedure prepareTips(rt: nnodeptr; Var total:integer; Var c:nnodeptr);

   Procedure analyzeContents(contents : nexusPtr;
                             rootName : flString;
                             Var characterBlockIndex : integer;
                             Var treeIndex : integer);


Implementation

   Var error_count : integer;

{*************************************************************************}
{This is a simple error reporter.                                         }
{*************************************************************************}
   Procedure pdnexus_error(message:string);
      Begin
         If ((error_count mod 8 = 0) And (error_count > 0))
         Then Begin
            writeln;
            write('Press enter to continue');
            readln;
         End; {then}
         writeln('Error - ',message);
         writeln;
         inc(error_count);
      End; {procedure pdnexus_error}

{*************************************************************************}
{This is a simple warning reporter.                                       }
{*************************************************************************}
   Procedure pdnexus_warning(message:string);
      Begin
         If ((error_count mod 8 = 0) And (error_count > 0))
         Then Begin
            writeln;
            write('Press enter to continue');
            readln;
         End; {then}
         writeln('Warning - ',message);
         writeln;
         inc(error_count);
      End; {procedure pdnexus_warning}

{*************************************************************************}
{ This just initializes the line counter for the error messages           }
{*************************************************************************}
   Procedure init_errors;
      Begin
         error_count := 0;
      End; {proceudre init_errors}

{*************************************************************************}
{ Ersatz constructor function; this allocates a node and initializes the  }
{ fields appropriately, to avoid initialization errors.                   }
{*************************************************************************}
   Function nnode_new : nnodeptr;
      Var
         result : nnodeptr;
      Begin
         result := new(nnodeptr);
         result^.Ancestor := nil;
         result^.dat[1] := UNINITIALIZED;
         result^.dat[2] := UNINITIALIZED;
         result^.SibL := nil;
         result^.SibR := nil;
         result^.tipL := nil;
         result^.tipR := nil;
         result^.length := 1.0; {better default length; allows polytomy testing}
         result^.index := UNINITIALIZED;
         result^.tipnum := -9;   {possibly unnecessary}
         result^.pdiname := '';
         nnode_new := result;
      End; {function nnode_new}

{*************************************************************************}
{ This disposes of an nnode and its descendents.                          }
{*************************************************************************}
   Procedure nnode_dispose(Var node : nnodeptr);
      Begin
          If (node^.sibl <> NIL)
          Then nnode_dispose(node^.sibl);
          If (node^.sibr <> NIL)
          Then nnode_dispose(node^.sibr);
          dispose(node); {done}
      End; {procedure nnode_dispose}

{*************************************************************************}
{ Ersatz constructor function; this allocates a trait and initializes the }
{ fields appropriately, to avoid initialization errors.                   }
{*************************************************************************}
   Function traits_new : traitsPtr;
      Var
	 result	: traitsPtr;
      Begin
	 result := new(traitsPtr);
	 result^[1] := 0.0;
	 result^[2] := 0.0;
         traits_new := result;
      End; {function traits_new}

{*************************************************************************}
{ Ersatz constructor function; this allocates a nexusRecord and           }
{ initializes the fields appropriately, to avoid initialization errors.   }
{*************************************************************************}
   Function nexusRecord_new : nexusPtr;
      Var
         result : nexusPtr;
      Begin
         result := new(nexusPtr);
         result^.trees := nil;
         result^.translations := nil;
         result^.taxa := nil;
         result^.characters := nil; 
         {result^.selectedTree := nil;}
         result^.selectedTranslateBlock := nil; 
         result^.selectedCharBlock :=nil;  {add fields as needed}
         nexusRecord_new := result;
      End; {function nexusRecord_new}

{*************************************************************************}
{ Ersatz constructor function; this allocates a taxa Block and            }
{ initializes the fields appropriately, to avoid initialization errors.   }
{*************************************************************************}
   Function taxaBlock_new : taxaBlockPtr;
      Var
         result : taxaBlockPtr;
      Begin
         result := new(taxaBlockPtr);
         result^.ntaxa := 0;
         result^.limit := 0;
         result^.next := NIL;
         taxaBlock_new := result;
      End; {function taxaBlock_new}

{*************************************************************************}
{ Disposal function for taxaBlocks.                                       }
{*************************************************************************}
   Procedure taxaBlock_dispose(Var toDie: taxaBlockPtr);
      Var count : integer;
      Begin
         For count := 1 to toDie^.ntaxa Do Begin
            If (toDie^.taxa[count] <> NIL)
            Then dispose(toDie^.taxa[count]);
         End; {for}
         {more to come}
         dispose(toDie);
      End; {taxaBlock_dispose}

{*************************************************************************}
{ Returns the integer index of the taxon in block with the name name.     }
{ Returns -1 if no such taxon is found.                                   }
{*************************************************************************}
   Function lookupTaxon(block : taxaBlockPtr; name : String) : integer;
      Var
         count : integer;
         found : boolean;
      Begin
         count := 1;
         found := false;
         While ((count <= block^.ntaxa) And (not found)) Do Begin
            If (block^.taxa[count]^.name = name)
            Then found := true
            Else inc(count);
         End; {while}
         If found
         Then lookupTaxon := count
         Else lookupTaxon := -1;
      End; {function lookupTaxon}

{*************************************************************************}
{ Creates a new taxon called name at the end of the taxon block.          }
{ Trying to duplicate an existing name generates a warning.               }
{*************************************************************************}
   Function addTaxon(block : taxaBlockPtr; name : string) : integer;
      Var result : integer;
      Begin
         result := lookupTaxon(block,name);  {check if already present}
         If result=-1 Then Begin
            inc(block^.ntaxa);
            new(block^.taxa[block^.ntaxa]);
            block^.taxa[block^.ntaxa]^.name := name;
            result := block^.ntaxa;
         End
         Else
            pdnexus_warning('duplicate taxon name; there may be a problem');
         addTaxon := result;
      End; {function addTaxon}

{*************************************************************************}
{ Ersatz constructor function; this allocates a character block and       }
{ initializes the fields appropriately, to avoid initialization errors.   }
{*************************************************************************}
   Function charBlock_new : charBlockPtr;
      Var result : charBlockPtr;
      Begin
         result := new(charBlockPtr);
         result^.taxa := nil;
         result^.nchar := 0;
         result^.items := states;
         result^.statesFormat := statesPresent;
         result^.dataTypeSpec := standard;
         result^.symbols := '';
         result^.mapDigits := false;
         result^.zeroBased := false;
         result^.next := NIL;
         charBlock_new := result;
      End; {function dataBlock_new}

{*************************************************************************}
{ This should free all the memory allocated for the dataBlock and its     }
{ parts.                                                                  }
{*************************************************************************}
   Procedure charBlock_dispose(Var toDie : charBlockPtr);
      Var i : integer;
      Begin
         If ((toDie^.taxa <> NIL) And
             (toDie^.nchar > 0))
         Then For i := 1 To toDie^.taxa^.limit Do
            If (toDie^.matrix[i] <> NIL)
            Then dispose(toDie^.matrix[i]);
         {more to come here}
         dispose(toDie);
      End; {procedure charBlock_dispose}

{*************************************************************************}
{This adds the character data string for one row of a character matrix.   }
{*************************************************************************}
   Procedure addCharData(block : charBlockPtr;
                         index : integer;
                         data : string);
      Var newData : charMatrixRowPtr;
      Begin
         new(newData);
         newData^.dataString := data;
         newData^.myTaxon := block^.taxa^.taxa[index];
         block^.matrix[index] := newData;
      End; {procedure addCharData}

{*************************************************************************}
{ Data blocks are character matrices that simultaneously define taxa sets.}
{ They are used in MacClade 3.00-3.07, so this needs to support them.     }
{ The abstraction will be implemented with a couple of functions that work}
{ with a characterBlock and its associated taxa block at the same time.   }
{*************************************************************************}
   Function dataBlock_new : charBlockPtr;
      Var
         result : charBlockPtr;
      Begin
         result := charBlock_new;
         result^.taxa := taxaBlock_new;
         dataBlock_new := result;
      End; {function dataBlock_new}

{*************************************************************************}
{ Function for adding a taxon in a data matrix.                           }
{*************************************************************************}
   Function dataBlock_addTaxon(block : charBlockPtr;
                               name : string) : integer;
      Begin
         dataBlock_addTaxon := addTaxon(block^.taxa,name);
      End; {function dataBlock_addTaxon}

{*************************************************************************}
{ Function for adding a taxon in a data matrix.                           }
{*************************************************************************}
   Procedure dataBlock_addCharData(block : charBlockPtr;
                                   index : integer;
                                   data : string);
      Begin
         addCharData(block,index,data);
      End; {procedure dataBlock_addTaxon}

{*************************************************************************}
{ Function for adding a taxon in a taxalabels list.                       }
{*************************************************************************}
   Function taxaBlock_addTaxon(block : taxaBlockPtr;
                               name : string) : integer;
      Begin
         taxaBlock_addTaxon := addTaxon(block,name);
      End; {function taxaBlock_addTaxon}

{*************************************************************************}
{ Procedure for cleaning up a list of translatePairs.                     }
{*************************************************************************}
   Procedure translatePair_dispose(toDie : translatePairPtr);
      Begin
         If (toDie^.next <> NIL)
         Then translatePair_dispose(toDie^.next);
         dispose(toDie);
      End;  {procedure translatePair_dispose}

{*************************************************************************}
{ Ersatz constructor function; this allocates a translateBlock and        }
{ initializes the fields appropriately, to avoid initialization errors.   }
{*************************************************************************}
   Function translateBlock_new : translateBlockPtr;
      Var
         result : translateBlockPtr;
      Begin
         result := new(translateBlockPtr);
         result^.index := 0; {TBD}
         result^.tlength := 0; {fill in length}
         result^.data := nil;
         result^.next := nil;
         translateBlock_new := result;
      End; {function translateBlock_new}

{*************************************************************************}
{ This should free all the memory allocated for the nexusRecord and its   }
{ parts.                                                                  }
{*************************************************************************}
   Procedure translateBlock_dispose(Var toDie : translateBlockPtr);
      Begin
         If (toDie^.next) <> nil
         Then translateBlock_dispose(toDie^.next);
         If (toDie^.data) <> NIL
         Then translatePair_dispose(toDie^.data);
         {more to come here}
         dispose(toDie);
      End; {procedure translateBlock_dispose}

{*************************************************************************}
{ This should update any translation blocks to refer to the most recently }
{ read taxon block if necessary.  Is this necessary?                      }
{*************************************************************************}
   Procedure updateTranslateBlocks(blockList : translateBlockPtr; 
                                   taxa : taxaBlockPtr);
   Begin
      {writeln('updateTranslateBlocks called');
      readln;}
   End; {procedure updateTranslateBlocks} 

{*************************************************************************}
{ Allocates a entry in a tree block to hold a single tree.                }
{*************************************************************************}
   Function treeEntry_new : treeEntryPtr;
      Var
         result : treeEntryPtr;
      Begin
         result := new(treeEntryPtr);
         result^.name := '';         {name of the tree}
         result^.tree := NIL;        {root of the tree}
         result^.next := NIL;        {next tree in the list}
         result^.rooted := true;     {rooted by default}
         treeEntry_new := result
      End; {function treeEntry_new}

{*************************************************************************}
{ This disposes of a tree entry and its tree.                             }
{*************************************************************************}
   Procedure treeEntry_dispose(entry : treeEntryPtr);
      Begin
         If (entry^.next <> NIL)
         Then treeEntry_dispose(entry^.next);
         If (entry^.tree <> NIL)
         Then nnode_dispose(entry^.tree);
         dispose(entry);
      End; {procedure treeEntry_dispose}

{*************************************************************************}
{ Ersatz constructor function; this allocates a treeBlock and initializes }
{ the fields appropriately, to avoid initialization errors.               }
{*************************************************************************}
   Function treeBlock_new : treeBlockPtr;
      Var
         result : treeBlockPtr;
      Begin
         result := new(treeBlockPtr);
         result^.name  := '';   {name of the block, if any}
         result^.trees := nil;  {list of tree entries in the block}
         result^.next  := nil;  {could be multiple tree blocks}
         treeBlock_new := result
      End; {function treeBlock_new}

{*************************************************************************}
{ This calls treeEntry_new and adds the new tree Entry to the treeBlock.  }
{*************************************************************************}
   Procedure treeBlock_addTree(block : treeBlockPtr;
                               name : nNameStr;
                               rooted : boolean;
                               Var result : treeEntryPtr);
      Begin
         result := treeEntry_new;
         result^.name := name;
         result^.rooted := rooted;
         result^.tree := nnode_new;   {preload tree?}
         result^.next := block^.trees;
         block^.trees := result;
      End; {procedure treeBlock_addTree}

{*************************************************************************}
{ This should free all the memory allocated for the treeBlock and its     }
{ parts.                                                                  }
{*************************************************************************}
   Procedure treeBlock_dispose(Var toDie : treeBlockPtr);
      Begin
         If (toDie^.trees) <> nil
         Then treeEntry_dispose(toDie^.trees);
         If (toDie^.next) <> nil
         Then treeBlock_dispose(toDie^.next);
         {more to come here}
         dispose(toDie);
      End; {procedure treeBlock_dispose}

{*************************************************************************}
{ This should free all the memory allocated for the nexusRecord and its   }
{ parts.                                                                  }
{*************************************************************************}
   Procedure nexusRecord_dispose(Var toDie : nexusPtr);
      Begin
         If (toDie^.trees <> nil)
         Then treeBlock_dispose(toDie^.trees);
         If (toDie^.translations <> nil)
         Then translateBlock_dispose(toDie^.translations);
         If (toDie^.taxa <> nil)
         Then taxaBlock_dispose(toDie^.taxa);
         If (toDie^.characters <> nil)
         Then charBlock_dispose(toDie^.characters);
         {more to come here}
         dispose(toDie);
      End; {procedure nexusRecord_dispose}

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a file without the
    extra exponent and mantessa digits.}
   {************************************************************************}
   Procedure extendWrite(Var outFile :text;num :extended);
      Var outStr :string[19];
          userChar : char;
      Begin
         If (abs(num) < 9.9E99) Then Begin
            If ((abs(num) > 1E-99) Or (Abs(num) = 0)) Then Begin
               str(num:19:-11,outStr);
               delete(outStr,16,2);
               write(outFile,outStr);
            End
            Else Begin
                clrscr;
               writeln;
               writeln('Error:  The program tried to write a number with an absolute value smaller');
               writeln('        than 1E-99.  The program will now exit.  You should go into');
               writeln('        PDTREE and rescale your tree, for example, if the height of your');
               writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
               writeln('        you may need to rescale tip values.');
               writeln;
               write('Press a key to exit the program');
               userChar := readkey;
               halt
            End; {else}
         End
         Else Begin
            clrscr;
            writeln;
            writeln('Error:  The program tried to write a number with an absolute value larger than');
            writeln('        9.9999999999E+99.  The program will now exit.  You should go into');
            writeln('        PDTREE and rescale your tree, for example, if the height of your');
            writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
            writeln('        you may need to rescale tip values.');
            writeln;
            write('Press a key to exit the program');
            userChar := readkey;
            halt
         End; {else}
      End; {procedure extendwrite}

{*************************************************************************}
{This recursive procedure checks a name that is passed with the rest of   }
{the tree to see if it is already in the tree.  If it is, it sets notfound}
{to false.  It calls nothing but itself.                                  }
{*************************************************************************}
   Procedure CheckPDIName(nod:nnodeptr;
                          name:pdinamestr;
                          Var notfound:boolean;
                          Var want:nnodeptr);
      Begin
         If nod^.pdiname=name Then Begin
            notfound:=false;
            want:=nod;
         End {then}
         Else Begin
            If (nod^.sibL<>nil)
            Then checkpdiname(nod^.sibL,name,notfound,want);
            If (nod^.sibR<>nil) and notfound
            Then checkpdiname(nod^.sibR,name,notfound,want);
         End; {else}
      End; {procedure checkname}

{*****************************************************************}
{This procedure assigns left to right ordering to the tree's tips.
 Besides adding ordering numbers to the tips, the procedure links
 the tips in order, useful in building the correspondence between
 tips and one or more traslate blocks. }
{******************************************************************}
  Procedure prepareTips(rt: nnodeptr; Var total:integer; Var c:nnodeptr);
     Var
        num :integer;
        previous,first : nnodeptr;
  {****************************************************************}
  {This recursive procedure traverses the tree and sets the tip
   pointers tipR,tipL, and the tip number tipnum.}
  {****************************************************************}
     Procedure traverse(nd : nnodeptr;
                        Var prev,first: nnodeptr;
                        Var num:integer);
        Begin
           If (nd^.SibL <> NIL) Then Begin
              traverse(nd^.SibL,prev,first,num);
              traverse(nd^.SibR,prev,first,num);
           End {then}
           Else Begin
              inc(num);
              If (num=1)
              Then first:= nd;
              nd^.tipnum:=num;
              nd^.tipL:=prev;
              If (prev<>NIL)
              Then prev^.tipR:=nd;
              prev:=nd;
           End; {else}
        End; {procedure traverse}
  {****************************************************************}
     Begin
        num:=0;
        previous:=NIL;
        traverse(rt,previous,first,num);
        previous^.tipR:=NIL;

        total:= previous^.tipnum;
        c:=first
     End; {prepareTips}


{*****************************************************************}
{This is a support procedure for several procedures called by     }
{analyzeContents.  It reports the details for one taxa block.     }
{It is needed because a taxa block may occur by itself or be      }
{implicitly defined in a data (or some other) block.              }
{*****************************************************************}
   Procedure reportOneTaxaBlock(block : taxaBlockPtr);
      Begin
      End; {reportOneTaxaBlock}


{*****************************************************************}
{This is a support procedure for analyzeContents.  It reports     }
{details on the character (and data) blocks found in the file.    }
{*****************************************************************}
   Procedure reportCharBlocks(contents : nexusPtr);
      Begin
      End; {reportCharBlocks}

{*****************************************************************}
{This is a support procedure for analyzeContents.  It reports     }
{details on the taxa blocks found in the file.                    }
{*****************************************************************}
   Procedure reportTaxaBlocks(contents : nexusPtr);
      Begin
      End; {reportTaxaBlocks}

{*****************************************************************}
{Display the trees in the file.                                   }
{*****************************************************************}
   Procedure displayTrees(blocks: treeBlockPtr;
                          Var blockCount : integer;
                          Var treeCount : integer);
      Var curBlock : treeBlockPtr;
          curTree : treeEntryPtr;
          temp      : nnodePtr;
          tipCount  : integer;
          anyUnroots: boolean;
      Begin
         treeCount := 0;
         anyUnroots := false;
         curBlock := blocks;
         writeln('Tree Summary');
         writeln;
         writeln('Number  Rooted?  Tips   Name');
         While (curBlock <> NIL) Do Begin
            curTree := curBlock^.trees;
            While (curTree <> NIL) Do Begin
               inc(treeCount);
               tipCount := 0;
               prepareTips(curTree^.tree,tipCount,temp);
               If (curTree^.rooted)
               Then writeln('   ',treeCount:3,'  Yes      ',tipCount,'    ',curTree^.name)
               Else Begin
                  writeln('   ',treeCount:3,'  No       ',tipCount,'    ',curTree^.name);
                  anyUnroots := true;
               End; {else}
               curTree := curTree^.next;
            End; {while}
            curBlock := curBlock^.next;
         End; {while}
         writeln;
         If anyUnroots
         Then pdnexus_warning('Unrooted trees encountered - they have been rooted arbitrarily');
      End; {procedure displayTrees}

{*****************************************************************}
{Select a tree to save, from the list generated by displayTrees.  }
{*****************************************************************}
   Procedure selectTree(Var selectedTree : integer;
                        treeCount : integer);
      Begin
         selectedTree := 0;
         Repeat
            writeln('Select a tree to save from the list by entering its number');
            write('Tree Number: ');
            readln(selectedTree);
            If (selectedTree <= 0)
            Then writeln('Tree number (',selectedTree,') is too small.');
            If (selectedTree > treeCount)
            Then writeln('Tree number (',selectedTree,') is too large.');
         Until ((selectedTree > 0) And (selectedTree <= treeCount));
         writeln;
      End;


{*****************************************************************}
{This counts the number of tree blocks in the file and the total  }
{number of trees in those blocks.                                 }
{*****************************************************************}
   Function countTaxa(blocks: taxaBlockPtr) : integer;
      Var count : integer;
      Begin
         count := 0;
         While (blocks <> NIL) Do Begin
            blocks := blocks^.next;
            inc(count);
         End; {while}
         countTaxa := count;
      End; {function countTaxa}

{*****************************************************************}
{This counts the number of tree blocks in the file and the total  }
{number of trees in those blocks.                                 }
{*****************************************************************}
   Function countCharBlocks(blocks : charBlockPtr) : integer;
      Var count : integer;
      Begin
         count := 0;
         While (blocks <> NIL) Do Begin
            blocks := blocks^.next;
            inc(count);
         End; {while}
         countCharBlocks := count;
      End; {function countCharBlocks}

{*****************************************************************}
{This procedure lists the trees and character blocks in the file, }
{and should let the user select which one to save to a pdi file.  }
{*****************************************************************}
   Procedure analyzeContents(contents : nexusPtr;
                             rootName : flString;
                             Var characterBlockIndex : integer;
                             Var treeIndex : integer);
      Var userch : char;
          blockCount : integer;
          treeCount : integer;
          ch : char;
      Begin
         {clrscr;}
         writeln('Summary of Nexus file: ',rootName,'.NEX');
         writeln;
         If (contents^.trees <> NIL) Then Begin
            displayTrees(contents^.trees,blockCount,treeCount);
            If (treeCount > 1)
            Then selectTree(treeIndex,treeCount) {let the user select a tree here}
            Else treeIndex := 1;
            blockCount := countTaxa(contents^.taxa);
            writeln('File contained ',blockCount,' taxa blocks.');
            writeln;
            blockCount := countCharBlocks(contents^.characters);
            writeln('File contained ',blockCount,' character blocks.');
            {Should let the user select a character block here}
            characterBlockIndex := 1;
            write('Press any key to continue');
            ch := readkey;
         End
         Else Begin {no trees}
            writeln('No trees were loaded, exiting');
            treeIndex := 0;
         End; {else}
      End; {procedure analyzeContents}

End. {Unit n_util1}
