{************************************************************************}
{This program reports descriptive statistics, and performs analysis of
 variance (ANOVA) and analysis of covariance (ANCOVA) on multiple sets
 of simulated characters.

 Modification history:
     14 May 1999 - Fixed lack of spaces between columns in output files
                   when value was negative. PEM
      3 Nov 1999 - Added compiler switch block.  PEM
      5 Jan 2000 - Added feature to add constant to each set
     11 Aug 2000 - Repeat list of species after each set is defined. PEM
     15 Jan 2001 - Allows user to choose species with 2-letter names,
                   instead of numbers.  PEM

 Last Modified 15 Jan 2001 Peter E. Midford                              }
{************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65500,0,655360} {big stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Program PDANOVA; {perform ANOVA on multiple sets of simulated characters}

   Uses dos;

   Const
      MaxNoSpp = 255; {number of terminal taxa!!!}
      MaxNoSets = 20;

   Type
      NameType          = Array[1..MaxNoSpp] of String[2];  {two character name}
      TraitArType       = Array [1..MaxNoSpp] of extended;  {values of one trait}
      SpValType         = Array[1..2] of TraitArType;       {values for two traits}
      SpSetType         = Set of 1..MaxNoSpp;               {subset of species to compare}
      SpSetArType       = Array[0..MaxNoSets] of SpSetType; {array of subsets}
      SetAttributesType = Array[0..MaxNoSets] of extended;      {e.g., subset means}
      SetPairType       = Array[1..2] of setattributestype;
      SNType            = Array[1..MaxNoSets] of byte; {samplesize of each set}
      SumSquaresType = Record
                          w,            {sum of squares within a set}
                          a,            {sum of squares among sets}
                          t : extended; {total sum of squares}
                       End; {record}
      DfType = Record
                  w,           {degrees of freedom within a set}
                  a,           {degrees of freedom among sets}
                  t : integer; {total degrees of freedom}
               End; {record}
      TraitNameType = Array[1..2] of String[7];
      setNameType = Array[0..MaxNoSets] of String[7];

      {define large buffers to reduce disk access}
      BufferType    = Array[1..Maxint Div 4] of char;
      BigBufferType = Array[1..Maxint Div 2] of char;

   Var
      Infile       : text;
      DATfile      : text;
      VARfile      : text;
      AVEfile      : text;
      CORfile      : text;
      MIMfile      : text;
      MEDfile      : text;
      Ffile        : text;
      LEVfile      : text;

      InBuf        : ^BigBuffertype;
      MIMbuf       : ^BigBuffertype;

      varbuf       : ^Buffertype;
      avebuf       : ^Buffertype;
      medbuf       : ^Buffertype;
      fbuf         : ^Buffertype;
      levbuf       : ^Buffertype;
      corbuf       : ^Buffertype;

      SS           : sumsquarestype;
      SXY          : sumsquarestype;
      Devyx        : sumsquarestype;

      TipAr        : SpValType; {holds actual tip values}
      AbsDev       : SpValType; {holds relative deviations from median or
                                 mean for Levene's test}
      SetConsts    : SpValType; {holds constants to be added to tip values
                                 based on set membership}

      SetVals      : setPairType; {holds constants for writing to .DAT file}
      Residuals    : TraitArType; {to store regression-corrected values}

      Mean         : SetAttributesType;
      Median       : SetAttributesType;

      SN           : SNType;

      Lo           : extended;
      Hi           : extended;
      MeanDev      : extended; {low and high value}
      SSTr1        : extended;
      SSTr2        : extended;

      SpSet        : SpSetArType;  {array of sets of taxa}


      T            : TraitNameType; {names for traits 1 and 2}

      setNames     : setNameType; {names for the sets}

      df           : dftype;


      I            : Integer;
      Sim          : Integer;
      SetNo        : Integer;
      NoSpp        : Integer;
      NoSets       : Integer;
      Sp           : Integer;
      Trait        : Integer;
      Ecode        : Integer;
      Name         : NameType;  {array of two-character symbols representing
                                 species}
      inFilePath   : pathStr;   {full name of input file}
      inFileDir    : dirStr;    {directory component}
      inFileName   : nameStr;   {name of input file}
      inFileExt    : extStr;    {extension component}

      CX           : Byte;      {independent variable in ANCOVA}
      CY           : Byte;      {dependent variable in ANCOVA}
      userChar     : char;      {character for yes/no questions}
      BlankLine    : String[21];
      MeanZero     : Boolean;
      MedianZero   : Boolean;   {used to test for zero-valued statistics}
      badResiduals : Boolean;

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a file without the
    extra exponent and mantessa digits.}
   {************************************************************************}
   Procedure extendWrite(Var outFile :text;num :extended);
      Var outStr :string[19];
      Begin
         str(num:19:-11,outStr);
         delete(outStr,16,2);
         write(outFile,outStr)
      End; {procedure extendwrite}

   {************************************************************************}
   {This procedure writes a line of asterisks.                              }
   {************************************************************************}
   Procedure asterisks(Var outFile : text);
      Begin
         writeln(Outfile,'*************************************' +
                   '*************************************');
      End; {procedure asterisks}

   {***********************************************************************}
   {This function returns true if the file exists and might contain a SIM
    format set of data, false otherwise. }
   {***********************************************************************}
   Function checkFile(fname : string) : boolean;
      Var testFile : text;
          testSize : longInt;
          valCode  : integer;
          testBuf  : string;
      Begin
         {$I-}   {turn off i/o checking}
         assign(testFile,fname);
         reset(testFile);
         If (IOResult = 0) Then Begin
            testSize := 0;
            readln(testFile);
            readln(testFile);
            readln(testFile);
            readln(testFile,testBuf);
            val(copy(testBuf,24,2),testSize,valCode);
            If (valCode <> 0) Then Begin
               writeln('The file ',fname, ' does not appear to be in .SIM format; try again.');
               checkFile := false;
            End
            Else checkFile := true;
            close(testFile);
         End
         Else Begin
            writeln('The file ',fname, ' does not exist; try again.');
            checkFile := false;
         End; {else}
         {$I+}
      End; {function checkFile}

   {************************************************************************}
   {This procedure extracts the date and time and writes them to the DAT
    file.                                                                   }
   {************************************************************************}
   Procedure DateTime;
      Var Year, Month, Day, DayofWeek, Hour, Minute, Second, Sec100 : word;
      Begin
         Write(DATfile, 'Output from PDANOVA run on ');
         GetDate(Year, Month, Day, DayofWeek);
         Write(DATfile, Day,' ');
         Case Month of
            1 : Write(DATfile, 'Jan ');
            2 : Write(DATfile, 'Feb ');
            3 : Write(DATfile, 'Mar ');
            4 : Write(DATfile, 'Apr ');
            5 : Write(DATfile, 'May ');
            6 : Write(DATfile, 'Jun ');
            7 : Write(DATfile, 'Jul ');
            8 : Write(DATfile, 'Aug ');
            9 : Write(DATfile, 'Sep ');
            10 : Write(DATfile, 'Oct ');
            11 : Write(DATfile, 'Nov ');
            12 : Write(DATfile, 'Dec ');
         End; {case month of}
         Write(DATfile, Year,' at ');
         GetTime(Hour, Minute, Second, Sec100);
         Write(DATfile, Hour, ':');
         If Minute < 10 Then Write(DATfile,'0');
         write(DATfile, Minute,':');
         If Second < 10 Then write(DATfile,'0');
         writeln(DATfile,Second,'  ');
         Writeln(DATfile, 'Input data file for this run was ', inFileName+inFileExt,
                          ', which contains ',NoSpp,' tips.');
         Writeln(DATfile);
      End; {proc datetime}

   {************************************************************************}
   {This program displays the intro screen.}
   {************************************************************************}
   Procedure DoIntro;
      Var I : Byte;
          j : char;
      Begin
         Writeln('       ********* Phenotypic Diversity Analysis Program *********');
        { Writeln;}
         Writeln('                            PDANOVA Module');
         Writeln;
         Writeln('V. 3.0 Copyright 15 Jan 2001 by Allan W. Dickerman, Peter E. Midford,');
         Writeln('                        and Theodore Garland, Jr.');
         Writeln;
         Writeln('     Contact T.G. at the Department of Zoology, 430 Lincoln Drive,');
         Writeln('              University of Wisconsin, Madison, WI 53706-1381');
         Writeln('            Telephone:  608-262-4437; Facsimile:  608-265-6320');
         Writeln('                   E-Mail: tgarland@facstaff.wisc.edu');
         Writeln;
         Writeln('This program performs analysis of variance and covariance, and Levene''s');
         Writeln('test, on multiple simulated data sets produced by PDSIMUL, CMSIMUL, or ');
         Writeln('PDSIMMOD, PDTIPS or PDRANDOM.  Output is like PDSINGLE, but in multiple files.');
         Writeln('Procedures are described in the following paper:');
         Writeln('Garland, T., Jr., A. W. Dickerman, C. M. Janis, and J. A. Jones.');
         Writeln('     1993. Phylogenetic analysis of covariance by computer');
         Writeln('     simulation. Systematic Biology 42:265-292.');
         writeln;
      End; {proc dointro}

   {************************************************************************}
   {This function returns the mean for one subset for one trait.}
   {************************************************************************}
   Function GetMean (OneSet : SpSetType;         {subset of species}
                     Nospp : Integer;            {total species in sim set}
                     Value : TraitArType)        {set of trait values}
                                          : extended;
      Var Sp, N : Integer;
          Total : extended;
      Begin
         N := 0;
         Total := 0;
         For Sp := 1 to NoSpp Do {calculate mean}
           If Sp In OneSet Then Begin
              Inc(N);
              Total := Total + Value[Sp];
           End; {if sp in spset}
         GetMean := Total/N;
      End; {function getmean}

   {************************************************************************}
   {This function calculates SSwithin for one subset.                       }
   {************************************************************************}
   Function GetSS (OneSet : SpSetType;
                   Nospp : integer;
                   Value : TraitArType)
                                        : extended;
      Var Sp : Integer;
          Mean, SS : extended;
      Begin {func}
         Mean := GetMean(Oneset, Nospp, Value);
         SS := 0;
         For Sp := 1 to NoSpp Do    {calculate squared deviations}
           If Sp In OneSet Then
             SS := SS + (Sqr(Mean - Value[Sp]));
         Getss := ss;
      End; {function getsumsquares}

   {************************************************************************}
   {This procedure calculates the sum of cross products for two traits for
    one subset. SCP = Sum of Cross Products}
   {************************************************************************}
   Function GetSCP (OneSet : SpSetType; Nospp : integer;
                    Value1, Value2 : TraitArType) : extended;
      Var Mean1, Mean2, SCP : extended;
          Sp : Integer;
      Begin {func}
         Mean1 := GetMean(Oneset, Nospp, Value1);
         Mean2 := GetMean(Oneset, Nospp, Value2);
         SCP := 0;
         For Sp := 1 to NoSpp Do    {calculate squared deviations}
            If Sp In OneSet Then
               SCP := SCP + (Mean1 - Value1[Sp]) * (Mean2 - Value2[Sp]);
         GetSCP := SCP;
      End; {function getscp}

   {************************************************************************}
   {This function calcuates the sums of squares for the analysis of
    variance.                                                               }
   {************************************************************************}
   Procedure DoAnova(SpSet : SpSetArType;
                     noSets, noSpp : Integer;
                     Value : TraitArType;
                     Var SS : SumSquaresType);
      Var grandMean : extended;
          setno, sp, n : integer;
      Begin
         With ss Do Begin
            w := 0;
            a := 0;
            t := 0;
            grandmean := getmean(spset[0], nospp, value);
            For setno := 1 To noSets Do Begin
               n := 0;
               For sp := 1 To nospp Do
                 If sp In spset[setno] Then inc(n);
               w := w + getss(spset[setno], nospp, value);
               a := a + sqr(grandmean - getmean(spset[setno], nospp, value)) * n;
            End; {for setno}
            t := getss(spset[0], nospp, value);
         End {with ss}
      End; {procedure doanova}

   {************************************************************************}
   {This procedure calculates sums of cross products and corresponding
    degrees of freedom.                                                     }
   {************************************************************************}
   Procedure GetSXY(SpSet : SpSetArType;
                    noSets, noSpp : Integer;
                    TipArray : SpValType;
                    Var SXY : SumSquaresType);
      Var grandmeanx, grandmeany : extended;
          setno, sp, n : integer;
      Begin
         With sxy Do Begin
            w := 0;
            a := 0;
            t := 0;
            grandmeanx := getmean(spset[0], nospp, tipar[1]);
            grandmeany := getmean(spset[0], nospp, tipar[2]);
            For setno := 1 to nosets Do Begin
               n := 0;
               For sp := 1 To nospp Do
                 If sp In spset[setno] Then inc(n);
               w := w + getscp(spset[setno], nospp, tipar[1], tipar[2]);
               a := a + (grandmeanx - getmean(spset[setno], nospp, tipar[1])) *
                      (grandmeany - getmean(spset[setno], nospp, tipar[2])) * n;
            End; {for setno}
            t := getscp(spset[0], nospp, tipar[1], tipar[2]);
         End {with sxy}
      End; {procedure getsxy}

   {************************************************************************}
   {This procedure performs the Analysis of Covariance.                     }
   {************************************************************************}
   Procedure DoAnCova (Spset : SpsetArType;
                       NoSets : Integer;
                       Tipar : SpValType;
                       Cy, Cx : Byte;
                       Var SSy, Devyx : SumSquaresType;
                       Var MeanDevyx : extended);
      Var SSx, SXY : SumSquaresType;
          indepSS : extended;
      Begin
         DoAnova(SpSet, NoSets, NoSpp, TipAr[Cx], SSx);
         DoAnova(SpSet, NoSets, NoSpp, TipAr[Cy], SSy);
         GetSXY(SpSet, NoSets, NoSpp, TipAr, SXY);
         DEVyx.t := SSy.t - sqr(SXY.t)/SSx.t;

         {sum of squares of deviations from regression lines (= error)}
         DEVyx.w := SSy.w - sqr(SXY.w)/SSx.w;
         DEVyx.a := DEVyx.t - DEVyx.w;
         MeanDevyx := 0;
         {mean squared deviations from regression for each group}
         For SetNo := 1 to NoSets Do Begin
            indepSS := GetSS(SpSet[SetNo], NoSpp, TipAr[CX]);
            If indepSS <> 0 Then         {avoid divide by zero error }
               MeanDEVyx := MeanDevyx +
                              GetSS(SpSet[SetNo], NoSpp, TipAr[Cy])
                              -sqr(GetSCP(SpSet[SetNo], NoSpp, TipAr[Cy], TipAr[Cx]))
                              /indepSS;
         End; {for setno}
      End; {procedure doancova}

   {************************************************************************}
   {This procedure counts species to arrive at the within, among and total
    degrees of freedom for the ANCOVA.                                      }
   {************************************************************************}
   Procedure GetDF (SpSet : SpSetArType;
                    Nosets, Nospp : Integer;
                    Var Df : DfType);
      Var SetNo, Sp : Integer;
      Begin
         With Df Do Begin
            w := 0;
            a := 0;
            t := 0;
            For SetNo := 1 To NoSets Do Begin
               inc(a); {df among groups}
               For Sp := 1 to NoSpp Do
               If Sp In SpSet[SetNo] Then Begin
                  inc(w);
                  inc(t);
               End; {then}
               dec(w);
            End; {for setno}
            dec(t);
            dec(a);
         End; {with df}
      End;{procedure getdf}

   {************************************************************************}
   {This procedure counts the number of species in each set.  Why does this
    procedure have overlapping functionality with the previous one?         }
   {************************************************************************}
   Procedure GetSN (SpSet : SpSetArType;
                    noSets, noSpp : Integer;
                    Var SN : SNtype);
      Var SetNo, Sp : Integer;
      Begin
         For SetNo := 0 To NoSets Do Begin
            SN[SetNo] := 0;
            For Sp := 1 to noSpp Do
              If Sp In SpSet[SetNo] Then inc(SN[Setno]);
         End; {for setno}
      End;{procedure getsn}

   {************************************************************************}
   { This procedure just lists the species codes.  It should be called      }
   { before the user is prompted for each set of species.                   }
   {************************************************************************}
   Procedure ListSpecies;
      Var sp,j : integer;
      Begin
         For j := 1 To 20 Do writeln;
         For j := 1 To 10 Do Write(j:4);
         Writeln;
         For sp := 1 To NoSpp Do
           If sp Mod 10 = 0 Then Begin
              writeln('  ', Name[Sp]);
              writeln;
              For j := sp + 1 To sp + 10 Do
                If J <= NoSpp Then write(j:4);
              writeln;
           End {then}
           Else write('  ',Name[Sp]);
         Writeln;
      End; {procedure listspecies}

   {************************************************************************}
   { This function looks up the two character name provided by the user and }
   { returns the appropriate species index.  If the name is not found, a    }
   { value greater than the number of species is returned.                                                           }
   {************************************************************************}
   Function nameLookup (nam : String) : integer;
      Var test : integer;
      Begin
         test := 1;
         While ((Name[test] <> nam) And (test <= noSpp)) Do
           inc(test);
         nameLookup := test;
      End; { nameLookup}

   {************************************************************************}
   {This procedure prompts the user through defining sets of species which
    will be compared. }
   {************************************************************************}
   Procedure GetSet (Var SpSet : SpSetArType; Var Nosets : Integer);
      Var SetNo, Sp, J, K : Integer;
          SN : Array[0..MaxNoSets] of integer;
          spName : string[2];
          c1 : char;
          readNames : boolean;
      Begin
         SN[0] := 0;
         SpSet[0] := [];
         setNames[0] := 'Overall';
         Writeln;
         Writeln('How many sets will be defined?');
         Readln(NoSets);
         write('Do you want to specify species by two character code or number? (C/N=Enter): ');
         userChar := ' ';
         Repeat
            readln(userChar);
            If (ord(userChar) = 13) Then userChar := 'N';
            userChar := upCase(userChar);
            If Not(userChar in ['C','N'])
            Then write('Please enter either C or N:');
         Until (userChar in ['C','N']);
         readNames := (userChar = 'C');
         For SetNo := 1 To NoSets Do Begin
            writeln('Enter a seven letter name for set ',setno,' or ENTER for none.');
            readln(setNames[setno]);
            writeln;
            Repeat
               SpSet[SetNo] := [];
               ListSpecies;
               If readNames Then Begin
                  writeln('Enter species for set ',setno,' ("', setNames[setNo],'") by above coding.');
                  writeln('Separate two character names with spaced.  Enter a <return>')
               End {then}
               Else Begin
                  writeln('Enter species for set ',setno,' ("', setNames[SetNo],'") by above numbering.');
                  writeln('Separate numbers with spaces.  Enter a <return>');
               End; {else}
               Writeln('BEFORE you reach the edge of the screen.  After you have');
               Writeln('typed in all the species for the set, type 0 <return> to end.');
               If readNames Then Begin
                  While (spSet[setNo] = []) Do Begin
                     Repeat
                        Repeat
                           read(c1)
                        Until (ord(c1) > 32);
                        spName := '' + c1;
                        read(c1);
                        If (ord(c1) > 32) Then Begin
                           spName := spName + c1;
                           Repeat
                              read(c1)
                           Until (ord(c1) <= 32)
                        End; {then}
                        j := nameLookup(spName);
                        If j > 0
                        Then spSet[setNo] := spSet[setNo] + [j];
                     Until (spName = '0');
                     If spSet[setNo] = [0] Then Begin
                        writeln('Warning: All sets must be nonempty; please reenter!');
                        writeln;
                        spSet[setNo] := []
                     End {then}
                  End {while}
               End {then}
               Else Begin
                  J := 300;
                  While (spSet[setNo] = []) Do Begin
                     Repeat
                        Read(j);
                        If J <= NoSpp Then SpSet[SetNo] := SpSet[SetNo] + [J];
                     Until J = 0;
                     If spSet[setNo] = [0] Then Begin
                        writeln('Warning: All sets must be nonempty; please reenter!');
                        writeln;
                        spSet[setNo] := []
                     End {then}
                  End {while}
               End; {else}
               If SpSet[SetNo] * SpSet[0] <> [] Then Begin
                  writeln;
                  asterisks(output);
                  writeln;
                  writeln('There is overlap between this set and the previously',
                           ' defined sets involving');
                  write('species ');
                  If readNames Then
                    For k := 1 to noSpp Do
                      if k in (SpSet[setNo] * spSet[0]) Then write(name[k])
                  Else
                    For K := 1 to NoSpp Do
                      If K in (SpSet[SetNo] * SpSet[0]) Then write(K:4);
                  writeln;
                  writeln('You will need to re-enter your species list for this set.');
                  writeln;
                  asterisks(output);
                  writeln;
                  writeln('Press enter to proceed');
                  readln; { flush remainder of previous entry line }
                  readln;
               End; {if spset[SetNo] * spset[0] <> []}
            Until SpSet[SetNo] * SpSet[0] = [];
            write('Subset ',setNo,' ("',setNames[SetNo],'") = ');
            SN[SetNo] := 0;
            For J := 1 to NoSpp Do
              If J In SpSet[SetNo] Then Begin
                 Write(Name[J],' ');
                 SN[SetNo] := SN[SetNo] + 1;
                 SpSet[0] := SpSet[0] + [J];
              End; {if j in spset}
            SN[0] := SN[0] + SN[SetNo];
            writeln;
            writeln('N = ',SN[SetNo]);
            writeln;
            If SN[SetNo] <= 1 Then Begin
               writeln;
               asterisks(output);
               writeln;
               If SN[SetNo] = 1
               Then writeln('Warning -- Because set ',setNo,' ("',setNames[SetNo],'") has only 1 member, variances ')
               Else writeln('Warning -- Because set ',setNo,' has no members, variances ');
               writeln('           will not be calculated for this set and no statistics related');
               writeln('           to slopes or residuals for any sets will be calculated.');
               writeln;
               asterisks(output);
               writeln
            End; {then}
            writeln;
            If Not(readNames)
            Then readln       {flush buffer for next set}
            Else Begin        {flush a different way}
               read(c1);
               While (ord(c1) <> 10) Do read(c1)
            End {else}
         End; {for setno}
         writeln(DATfile, 'Note: Set 0 ("Overall") is the union of all subsets.');
         writeln(DATfile);
         For SetNo := 0 To NoSets Do Begin
            write('Subset ',Setno,' ("',setNames[SetNo],'") = ');
            write(DATfile, 'Set ',Setno,' ("',setNames[SetNo],'") = ');
            For J := 1 To NoSpp Do
              If J In SpSet[SetNo] Then Begin
                 write(Name[J],' ');
                 write(DATfile, Name[J],' ');
              End; {if j in spset}
            writeln(DATfile);
            writeln(DATfile);
            writeln;
            writeln;
         End; {for SetNo}
      End; {procedure getset}

   {************************************************************************}
   {This function returns the median for one subset.}
   {************************************************************************}
   Function GetMedian (OneSet : SpSetType;
                       Nospp : Integer;
                       Value : TraitArType)
                                           : extended;
      Var Sp, K : Byte;
          MedianMinus, MedianPlus : extended;
          Ties, NoHigher, NoLower  : Integer;
      Begin
         For Sp := 1 to NoSpp Do
           If Sp In OneSet Then Begin {count the smaller/larger members of the set}
              NoHigher := 0;
              NoLower := 0;
              Ties := 0;
              For K := 1 to NoSpp Do
                If K In OneSet Then Begin
                   If Value[K] > Value[Sp] Then inc(NoHigher);
                   If Value[K] < Value[Sp] Then inc(NoLower);
                   If Value[K] = Value[Sp] Then inc(Ties);
                End; {if k in oneset}
              If Abs(NoHigher - NoLower) < Ties Then Begin
                 MedianMinus := Value[Sp];
                 MedianPlus := Value[Sp];
              End
              Else If Abs(NoHigher - NoLower) = Ties
                Then If NoLower > NoHigher
                     Then MedianMinus := Value[Sp]
                     Else MedianPlus := Value[Sp];
           End; {if Sp in oneset}
         GetMedian := (MedianPlus + MedianMinus)/2;
      End; {procedure getmedian}

   {************************************************************************}
   {This procedure calculates (into AbsDev), for one trait, the absolute
    value of the relative deviation from their within-subgroup central value
    (median or mean)  =   |(x-cent)/cent|     }
   {************************************************************************}
   Procedure GetAbsDev (SpSet : SpSetArType;
                        noSets, noSpp : Integer;
                        Cent : setattributestype;  {= measure of CENTral tendency}
                        Value : TraitArType;
                        Var AbsDev : TraitArType);
      Var Sp, SetNo, Trait : integer;
      Begin
         For Sp := 1 to NoSpp Do
           For SetNo := 1 to NoSets Do
             If Sp In SpSet[SetNo]
             Then AbsDev[Sp] := Abs((Value[Sp] - Cent[SetNo])/Cent[SetNo]);
      End; {procedure getabsdev}

   {************************************************************************}
   {This procedure finds lowest and highest values for each trait in one
    subset.}
   {************************************************************************}
   Procedure GetLoHi (OneSet: SpSetType;
                      Nospp : integer;
                      Value : TraitArType;
                      Var Lo, Hi : extended);
     Var Sp : Integer;
     Begin
        For Sp := 1 to NoSpp Do
          If Sp in OneSet Then Begin
             Hi := Value[Sp];
             Lo := Value[Sp];
          End; {initializing arrays to lie within range of species in oneset}
        For Sp := 1 to NoSpp Do {calculate means and find lo and hi values}
          If Sp In OneSet Then Begin
             If Value[Sp] < Lo Then Lo := Value[Sp];
             If Value[Sp] > Hi Then Hi := Value[Sp];
          End; {if sp in oneset}
     End; {procedure lowhi}

   {************************************************************************}
   {This procedure transforms independent variable to the deviation from
    regression on covariate using within group slopes.}
   {************************************************************************}
   Procedure GetResiduals(SpSet : SpSetArType;
                          NoSets, NoSpp : Integer;
                          Dependent, Covariate : TraitArType;
                          Var NeuValue : TraitArType);
      Var MeanX, MeanY, Slope : extended;
          SS : extended;
          SetNo, Sp : integer;
      Begin
         For SetNo := 1 To Nosets Do Begin
            MeanX := GetMean(SpSet[SetNo], NoSpp, Covariate);
            MeanY := GetMean(SpSet[SetNo], NoSpp, Dependent);
            SS := GetSS(SpSet[SetNo], NoSpp, Covariate);
            If SS <> 0 Then Begin
               Slope := GetSCP(SpSet[SetNo], NoSpp, Dependent, Covariate) / SS;
               For Sp := 1 to NoSpp Do
                 If Sp In SpSet[SetNo]
                 Then NeuValue[Sp]:=Dependent[Sp]-
                                      (MeanY+((Covariate[Sp]-MeanX)*Slope));
            End
            Else
              For Sp := 1 To NoSpp Do
                If Sp In SpSet[SetNo]
                Then NeuValue[Sp] := -9;
         End; {for setno}
      End; {proc getresiduals}

   {************************************************************************}
   {This procedure prompts for and reads constant values to add to sets.    }
   {************************************************************************}
   Procedure getConstants;
      Var constVal : extended;
          valCode : integer;
          i : byte;
          j : byte;
          userChar : char;
          userBuf : string;
      Begin
         write('Add a constant to trait 1 (=',
               T[1],
               ') for one or more sets (Y/N=Enter)? ');
         readln(userChar);
         If (upcase(userChar) in ['Y','y']) Then Begin
            For i := 1 to NoSets Do Begin
               write('Enter a constant for set ',i,' (= ');
               write(setNames[i],') or ENTER for zero: ');
               readln(userBuf);
               If (userBuf <> '') Then Begin
                  val(userBuf,constVal,valCode);
                  If (valCode <> 0)
                  Then writeln('Error - couldn'' read value, no constant will be added.')
                  Else Begin
                     setVals[1,i] := constVal;
                     For j := 1 To NoSpp Do Begin
                        If j in spSet[i]
                        Then setConsts[1,j] := constVal;
                     End; {for}
                  End; {else}
               End; {then}
            End; {for}
         End; {then}
         write('Add a constant to trait 2 (=',T[2]);
         write(') for one or more sets (Y/N=Enter)? ');
         readln(userChar);
         If (upcase(userChar) in ['Y','y']) Then Begin
            For i := 1 to noSets Do Begin
               write('Enter a constant for set ',i,' (= ');
               write(setNames[i],') or ENTER for zero: ');
               readln(userBuf);
               If (userBuf <> '') Then Begin
                  val(userBuf,constVal,valCode);
                  If (valCode <> 0)
                  Then writeln('Error - couldn'' read value, no constant will be added.')
                  Else Begin
                     setVals[2,i] := constVal;
                     For j := 1 To NoSpp Do
                       If j in spSet[i]
                       Then setConsts[2,j] := constVal;
                  End {else}
               End {then}
            End {for}
         End {then}
      End; {Procedure getConstants}

   {************************************************************************}
   Begin { main }
      DoIntro;
      badResiduals := false;
      Writeln('When prompted, enter the full name of your input file (path and extension)');
      writeln('(e.g., .SIM for simulated data from PDSIMUL or .RND for permuted data from');
      writeln('PDRANDOM).  This program will always write results to your current directory');
      writeln('even if the input file is in a different directory.');
      writeln;
      Write('Enter input file name: ');
      Readln(inFilePath);
      While not(checkFile(inFilePath)) Do Begin
         writeln;
         write('Please enter a new file name: ');
         readln(inFilePath);
      End; {while}
      fSplit(inFilePath,inFileDir,inFileName,inFileExt);
      Assign(Infile,inFilePath);
      New(Inbuf);
      SetTextBuf(Infile, Inbuf^);
      Reset(Infile);
      writeln;
      Writeln('Input file = ',infileName+inFileExt);

      Assign(DATfile, InfileName+'.DAT');
      Rewrite(DATfile);

      Assign(VARfile, InfileName+'.VAR');
      New(VarBuf);
      SetTextBuf(Infile, Varbuf^);
      ReWrite(VARfile);

      Assign(AVEfile, InfileName+'.AVE');
      New(AveBuf);
      SetTextBuf(Infile, Avebuf^);
      ReWrite(AVEfile);

      Assign(MEDfile, InfileName+'.MED');
      New(Medbuf);
      SetTextBuf(Infile, Medbuf^);
      ReWrite(MEDfile);

      Assign(MIMfile, InfileName+'.MIM');
      New(Mimbuf);
      SetTextBuf(Infile, Mimbuf^);
      ReWrite(MIMfile);

      Assign(CORfile, InfileName+'.COR');
      New(Corbuf);
      SetTextBuf(Infile, Corbuf^);
      Rewrite(CORfile);

      Assign(Ffile, InfileName+'.F');
      New(Fbuf);
      SetTextBuf(Infile, Fbuf^);
      ReWrite(Ffile);

      Assign(LEVfile, InfileName+'.Lev');
      New(LevBuf);
      SetTextBuf(Infile, Levbuf^);
      ReWrite(LEVfile);


      {Now we need to read the number of species (tips) before writing}
      {the header for the .DAT file, PEM 16 Nov 1999}
      For I := 1 to 3 Do Readln(Infile);
      Readln(Infile, BlankLine, NoSpp);

      DateTime;
      While Pos('____',BlankLine) = 0 Do
        Readln(Infile, BlankLine);
      For I := 1 To NoSpp Do
        Readln(Infile, Name[I]);
      GetSet(SpSet, NoSets);
      Reset(Infile);
      Readln(Infile, BlankLine);
      While Pos('____',BlankLine) = 0 Do
        Readln(Infile, BlankLine);
      Writeln('Type seven letter abbreviation for trait 1, or ENTER for none.');
      Readln(T[1]);
      If T[1] = ''
      Then T[1] := 'Trait 1';
      Writeln('Type seven letter abbreviation for trait 2, or ENTER for none.');
      Readln(T[2]);
      If T[2] = ''
      Then T[2] := 'Trait 2';

      {---* Analysis of Covariance *---}
      CX := 0;
      While Not (CX In [1, 2]) Do Begin
         Writeln('Which variable shall be the covariate? (1 or 2)');
         Readln(CX);
         CY := Cx Mod 2 + 1;
      End; {while not cx in [...]}


      {initialize added constants appropriately}
      For i := 1 to MaxNoSpp Do Begin
         setConsts[1,i] :=0;
         setConsts[2,i] :=0;
      End; {for}
      For i := 1 to MaxNoSets Do Begin
         setVals[1,i] := 0;
         setVals[2,i] := 0;
      End; {for}
      Writeln('Would you like to add a constant to any of the sets? (Y/N=Enter): ');
      readln(userChar);
      If (upCase(userChar) in ['Y','y']) Then getConstants;

      Writeln(DATfile,
        'Each line of the output files corresponds to a single simulated data set.');
      Writeln(DATfile);

      Writeln(Datfile,
        'In each of the following output files, columns of statistics are');
      Writeln(DatFile,'separated by spaces.');
      writeln(DatFile);

      Writeln(DATfile, InfileName,'.AVE contains the subset means in the',
                        ' following order:');
      For I := 0 To NoSets Do
        Writeln(DATfile,i*2+1,' - ',i*2+2,') Set ',I,' ("',setNames[I],'") : ',T[1],', ', T[2]);
      Writeln(DATfile);
      Writeln(DATfile, InfileName,'.MED contains the subset medians in the',
                        ' following order:');
      For I := 0 To NoSets Do
        Writeln(DATfile,i*2+1,' - ',i*2+2,') Set ',I,' ("',setNames[I],'") : ',T[1],', ', T[2]);
      Writeln(DATfile);
      Writeln(DATfile, InfileName,'.VAR contains the subset variances in the',
                        ' following order:');
      For I := 0 To NoSets Do
        Writeln(DATfile,i*2+1,' - ',i*2+2,') Set ',I,' ("',setNames[I],'") : ',T[1],', ', T[2]);
      Writeln(DATfile);
      Writeln(DATfile, InfileName,'.MIM contains the subset minimums and ',
                        'maximums in the following order:');
      For I := 0 To NoSets Do
        Writeln(DATfile,i*4+1,' - ',i*4+4,') Set ',I,' ("',setNames[I],'") : '
                    ,T[1],' (min, max); ', T[2],' (min, max) ');
      Writeln(DATfile);
      Writeln(DATfile, InfileName,'.COR contains the within-set correlations',
                        ' between traits for ');
      For I := 0 To NoSets Do
        Writeln(DatFile, I+1,') Set ',I,' ("',setNames[I],'")');
      Writeln(DATfile);
      Writeln(DATfile,InfileName,
                        '.F contains six F-ratios from ANOVA and ANCOVA in the following order:');
      Writeln(DATfile, '1) ANOVA of ',T[1]);
      Writeln(DATfile, '2)   "   of ',T[2]);
      Writeln(DATfile, '3) ',T[cx],' as covariate in ANCOVA of ',t[cy]);
      Writeln(DATfile,
                  '4) ',T[cy],'(adjusted) in ANCOVA with ',T[cx],' as covariate');
      Writeln(DATfile,
               '5) explained variance in ANCOVA of ',T[cy],' with ',t[cx],' as covariate');
      Writeln(DATfile, '6) difference between slopes');
      Writeln(DATfile);
      Writeln(DATfile,
                 InfileName,'.LEV contains six F-ratios from Levene''s tests:');
      Writeln(DATfile,
                 '1) ANOVA of absolute values of deviations of Trait1 from within-group',
                 ' medians');
      Writeln(DATfile,
                 '2)   "   "      "      "     "      "     "    "     "      "     "   ',
                 'means');
      Writeln(DATfile,
                 '3)   "   "      "      "     "      "     "  Trait2  "      "     "   ',
                 'medians');
      Writeln(DATfile,
                 '4)   "   "      "      "     "      "     "    "     "      "     "   ',
                 'means');
      Writeln(DATfile,
                 '5)   "   "  absolute value of residuals of', T[CY]:8,' from within-group',
                 ' regressions on ',T[cx]);
      Writeln(DATfile,
                 '6)   "   "  squared residuals of', T[CY]:8,' from within-group regressions',
                 ' on ',T[cx]);
      Writeln(DATfile);
      Writeln(DATfile, 'For a description of Levene''s test see ');
      Writeln(DATfile, 'Schultz (1985) Systematic Zoology 34:449-456');
      Writeln(DATfile);

      writeln(DatFile);
      For i := 1 to MaxNoSets Do Begin
         If (setVals[1,i] <> 0) Then Begin
            write(DATfile,'A constant of ');
            extendWrite(DATfile,setVals[1,i]);
            writeln(DatFile,' was added to trait 1 in set ',i,' (',setNames[i],').');
         End;
         If (setVals[2,i] <> 0) Then Begin
            write(DATfile,'A constant of ');
            extendWrite(DATfile,setVals[2,i]);
            writeln(DatFile,' was added to trait 2 in set ',i,' (',setNames[i],').');
         End;
      End;
      writeln(DatFile);


      GetDf(SpSet, NoSets, NoSpp, Df);
      GetSN(SpSet, NoSets, NoSpp, SN);
      Sim := 0;
      While Not Eof(Infile) Do Begin  {analyze each simulated data set}
         inc(Sim);                    {count the sets}
         For Sp := 1 to NoSpp Do      {read the simulated data}
            Readln(Infile, Name[Sp], TipAr[1,Sp], TipAr[2,Sp]);
         Writeln('Analyzing simulation ',sim:4);
         {add set constants (all will be zero if no constants were set)}
         For Sp := 1 to NoSpp Do Begin
            TipAr[1,Sp] := TipAr[1,Sp] + setConsts[1,Sp];
            TipAr[2,Sp] := TipAr[2,Sp] + setConsts[2,Sp];
         End; {for}
         For setno := 0 To nosets Do Begin
            For Trait := 1 to 2 Do Begin
               extendWrite(AVEfile, GetMean(SpSet[SetNo], NoSpp, TipAr[Trait]));
               write(AVEFile,' ');
               extendWrite(MEDfile, GetMedian(SpSet[SetNo], NoSpp, TipAr[Trait]));
               write(MEDfile,' ');
               GetLoHi(SpSet[SetNo], NoSpp, TipAr[Trait], Lo, Hi);
               extendWrite(MIMfile, Lo);
               write(MIMfile,' ');
               extendWrite(MIMfile, Hi);
               write(MIMfile,' ');
               If (SN[SetNo]>1)
               Then Begin
                  extendWrite(VARfile,(GetSS(SpSet[SetNo],NoSpp,TipAr[Trait])/(SN[SetNo]-1)));
                  write(VARfile,' ');
               End {then}
               Else Begin
                  write(VARfile,-9:17,' ');
                  badResiduals := true;
               End; {else}
            End; {for trait}
            SSTr1 := GetSS(SpSet[SetNo], NoSpp, TipAr[1]);
            SSTr2 := GetSS(SpSet[SetNo], NoSpp, TipAr[2]);
            If (SSTr1 <> 0) And (SSTr2 <> 0)
            Then extendWrite(CORfile,GetSCP(SpSet[SetNo],Nospp,TipAr[1],TipAr[2])/sqrt(SSTr1*SSTr2))
            Else Begin
               write(CORfile,-9:17,' ');
               badResiduals := true;
            End; {else}
            write(CORfile,' ');
         End; {for setno}
         writeln(avefile);
         writeln(medfile);
         writeln(MIMfile);
         writeln(varfile);
         writeln(corfile);
         For Trait := 1 to 2 Do Begin
            DoAnova(SpSet, Nosets, Nospp, TipAr[Trait], SS);
            extendWrite(Ffile,(SS.a/Df.a)/(SS.w/DF.w));
            write(Ffile,' ');
         End; {for trait, doing anova}

         DoAnCova(SpSet, Nosets, TipAr, Cy, Cx, SS, Devyx, MeanDev);
         extendWrite(Ffile, (SS.w-DEVyx.w)/(DEVyx.w/(DF.w-1)));
         write(Ffile,' ');
         extendWrite(Ffile, (DEVyx.a/Df.a) / (DEVyx.w/(DF.w-1)));
         write(Ffile,' ');
         extendWrite(Ffile, ((SS.t-DEVyx.w)/NoSets)/(DEVyx.w/(DF.w-1)));
         write(Ffile,' ');

         { F ratio for difference among slopes next}
         If badResiduals
         Then write(Ffile, -9:17)
         Else extendWrite(Ffile, ((DEVyx.w-MeanDEV)/Df.a) / (MeanDev/(Df.w - NoSets)));
         writeln(Ffile);

         {Levene's test follows}
         For Trait := 1 to 2 Do Begin
            {Levene's test requires dividing by median or mean, so must test to be
             sure division by zero is avoided}
            MeanZero := False;
            MedianZero := False;
            For SetNo := 0 to NoSets Do Begin {look for zero medians or means}
               Median[SetNo] := GetMedian(SpSet[SetNo], NoSpp, TipAr[Trait]);
               If abs(Median[SetNo]) < 1E-10
               Then MedianZero := True;
               Mean[SetNo] := GetMean(SpSet[SetNo], NoSpp, TipAr[Trait]);
               If abs(Mean[SetNo]) < 1E-10
               Then MeanZero := True;
            End; {testing for zero medians or means}

            If MedianZero
            Then Write(LEVfile, -9:17)
            Else Begin
               GetAbsDev (SpSet, NoSets, NoSpp, Median, TipAr[Trait], AbsDev[Trait]);
               DoAnova(SpSet, NoSets, NoSpp, AbsDev[Trait], SS);
               extendWrite(LEVfile,(SS.a/Df.a)/(SS.w/DF.w));
               write(LEVfile,' ');
            End; {if not medianzero}

            If MeanZero
            Then Write(LEVfile, -9:17)  {write -9 if any test cannot be performed}
            Else Begin
               GetAbsDev (SpSet, NoSets, NoSpp, Mean, TipAr[Trait], AbsDev[Trait]);
               DoAnova(SpSet, NoSets, NoSpp, AbsDev[Trait], SS);
               extendWrite(LEVfile,(SS.a/Df.a)/(SS.w/DF.w));
            End; {if not meanzero}
            write(LEVfile,' ');
         End; {for trait, doing regular Levene's test}

         If badResiduals Then Begin
            write(LEVFile, -9:17,' '); {bad abs(residuals)}
            write(LEVFile, -9:17); {bad sqr(residuals)}
            writeln(LEVFile);
         End
         Else Begin
            { ANOVA on absolute values of deviations form within-group regressions}
            GetResiduals(SpSet, NoSets, NoSpp, TipAr[Cy], TipAr[Cx], Residuals);
            For Sp := 1 to NoSpp Do
              Residuals[Sp] := Abs(Residuals[Sp]);  {absolute values}
            DoAnova(SpSet, NoSets, NoSpp, Residuals, SS);
            extendWrite(LEVfile,(SS.a/Df.a)/(SS.w/DF.w));
            write(LEVfile,' ');

            { ANOVA on squared deviations }
            For Sp := 1 to NoSpp Do
              If Sp In SpSet[0] {one of the species being considered}
              Then Residuals[Sp] := Sqr(Residuals[Sp]);
            DoAnova(SpSet, NoSets, NoSpp, Residuals, SS);
            extendWrite(LEVfile,(SS.a/Df.a)/(SS.w/DF.w));
            Writeln(LEVfile);
         End; {else}
      End; {while not eof infile}
      If badResiduals Then Begin
         writeln(datFile,'Because a species set in one or more data sets had no variance,');
         writeln(datFile,'some results could not be computed and were set to -9.');
         writeln(datFile);
      End; {then}
      Writeln(DatFile, Sim,' simulations were analyzed.');
      Close(DATfile);
      Close(Infile);
      Close(VARfile);
      Close(AVEfile);
      Close(MEDfile);
      Close(MIMfile);
      Close(CORfile);
      Close(Ffile);
      Close(LEVfile);
      Writeln;
      Writeln('Finished !');
      Writeln(Sim,' simulations were analyzed.');
      Writeln('Input file was ',inFilePath);
      Writeln('Eight output files have been written:');
      Writeln(InfileName,'.DAT = Outline of contents of other files.');
      Writeln(InfileName,'.AVE = Means for each character for each set.');
      Writeln(InfileName,'.VAR = Variances for each character for each set.');
      Writeln(InfileName,'.MIM = Minimum and maximum values for each character ',
               'for each set.');
      Writeln(InfileName,'.MED = Median values for each character for each ',
               'set.');
      Writeln(InfileName,'.COR = Correlations between trait 1 & 2 for each ',
               'set.');
      Writeln(InfileName,'.F   = F-statistics for ANOVA and ANCOVA');
      Writeln(InfileName,'.LEV = F-statistics for Levene''s median-ratio test');
   End. {program PDANOVA}
