{**********************************************************************}
{PDDIST                                                                }
{                                                                      }
{This program reads a tree file and writes two matrices to separate    }
{files.  Each matrix contains the differences in tip values for one of }
{the two traits on the tree.                                           }
{                                                                      }
{Revision History                                                      }
{                                                                      }
{ 11 Oct 1996 - Coding begun.  Peter E. Midford                        }
{  1 Nov 1996 - Finished initial coding.  Peter E. Midford             }
{  1 Dec 1997 - Added two new distance metrics.  Peter E. Midford      }
{ 15 Dec 1997 - Continued addition of new metrics.  Peter E. Midford   }
{ 12 Jan 1998 - Added a fourth "shared/total" metric. Peter E. Midford }
{ 21 Jan 1998 - Completed changes to output format. Peter E. Midford   }
{ 26 Jan 1998 - Fixed small format bug and added checking to self      }
{               distance option for the nodelength metric. PEM         }
{  5 Feb 1998 - Added two new diagonal options, a new metric and an    }
{               option to scale the entire matrix by depth of the tree }
{                                                                      }
{  3 Nov 1999 - Added standard compiler switches to the top of the     }
{               file.  Also updated copies of the modules from PDTREE. }
{               PEM                                                    }
{ 12 Mar 2003 - Updated references for option 5.  PEM                  }
{**********************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


Program PDDIST;

uses t_util1,    {tree and node definitions and tip ordering}
     t_io1,      {for reading all types of tree files}
     crt;         {just used to clear the screen}

   Type
      metric = (tipDiff,pathLength,nodeLength,shared,covariance);
                           {choice of distance functions}
      userOptions =
         Record
            headers        : boolean; {true to write headers and labels}
            scaling        : boolean; {true to scale to a max of 10.0}
            writeColumn    : boolean; {true to dump as column instead of matrix}
            shortForm      : boolean; {true to write in 8:5 format}
            scale          : traits;  {scaling factor for each trait}
            Case choice    : metric of
               pathLength,
               nodeLength,
               shared      : (selfDistance : extended;  {self Distance factor}
                              selfDistConst : boolean); {true for
                                                        constant self
                                                        distance}
               covariance  : (longest : extended;   {height of highest node}
                              scaleBoth : boolean); {true to scale both
                                                      off diagonal elements
                                                      by height of tree}
         End; {type userOptions}

   Var
      rt : nodeptr;       {This is the root of the tree}
      first : nodeptr;    {This is the leftmost tip}
      flname : flstring;  {The name of the tree file}
      ext : typstring;    {The extension of the tree file}
      tipCount : integer; {The number of tips on the tree}
      userChoice : userOptions; {group of option choices}

   {*******************************************************************}
   {This procedure simply displays the introductory screen.            }
   {*******************************************************************}
   Procedure introScreen;
      Begin
         clrscr;
      Writeln('  ********* Phenotypic Diversity Analysis Program - PDDIST Module *********');
      Writeln;
      Writeln(' V. 1.0 Copyright 6 January 2000 by Peter E. Midford and Theodore Garland, Jr.');
      Writeln;
      Writeln('  Direct all questions to T.G. at Department of Zoology, 430 Lincoln Drive,');
      Writeln('          University of Wisconsin, Madison, WI 53706-1381  U.S.A.');
      Writeln('Telephone 608-262-4437; Facsimile 608-262-9083; E-mail tgarland@macc.wisc.edu');
      Writeln;
      Writeln('     This program creates distance matrices from phylogenetic tree files');
      Writeln('produced by the PDTREE program (e.g., in the *.PDI or *.INP format).');
      Writeln('These distance matrices can be of various types, including:');
      Writeln;
      Writeln('(1) the pairwise differences between the character values for each');
      Writeln('    tip species (two matrices will be written);');
      Writeln('(2) the patristic (phylogenetic pathlength) distance;');
      Writeln('(3) the number of internal nodes separating each pair of tips;');
      Writeln('(4) the proportion of total branch length, from root to tip,');
      Writeln('    that represents shared evolutionary history;');
      Writeln('(5) the absolute branch length, from root to tip, that is shared.');
      Writeln;
      Writeln('The distance matrices can be used for various purposes -- see the');
      Writeln('documentation in PDINSTR.DOC for various ideas and examples.');
      Writeln;
      Writeln;
          ret('                   Please press any key to continue.');
      End; {procedure introScreen}

   {*******************************************************************}
   {This procedure should be just a wrapper to a call to inputTree.    }
   {*******************************************************************}
   Procedure readTree(Var rt : nodePtr; Var flname : flstring);
      Var
         succ : boolean;     {throw away flag}
      Begin
         inputTree(rt,flname,ext,succ);
      End;  {procedure readtree}

   {*******************************************************************}
   {This function is a helper for the shared distance metric. It
    calculates the distance from a node to the root.  All it does is
    walk down branches toward the root, adding the length of each
    branch to accum.                                                   }
   {*******************************************************************}
   Function branchHeight(root : nodePtr;
                         node : nodePtr) : extended;
      Var accum : extended;   {accumulates branch lengths}
          curNode : nodePtr;  {used to walk down toward root}
      Begin
         accum := 0;
         curNode := node;
         While (curNode <> root) Do Begin
            accum := accum + curNode^.length;
            curNode := curNode^.ancestor;
         End; {while}
         branchHeight := accum;
      End; {function branchHeight}

   {*******************************************************************}
   {This function is similar to branchHeight (preceeding) except it
    calculates the height of a node in terms of the number of nodes
    from the root.  This may not get used in the current 5 Feb 1998
    version, but it might be useful later on.                          }
   {*******************************************************************}
   Function nodeHeight(root : nodePtr;
                       node : nodePtr) : integer;
      Var accum : integer;   {accumulates branch lengths}
          curNode : nodePtr;  {used to walk down toward root}
      Begin
         accum := 0;
         curNode := node;
         While (curNode <> root) Do Begin
            accum := accum + 1;
            curNode := curNode^.ancestor;
         End; {while}
         nodeHeight := accum;
      End; {function nodeHeight}


   {**********************************************************************}
   {This function calculates the distance between two nodes on the tree.

    There are currently five methods (called metrics) for doing this.

     1. Difference between tip values, optionally scaled to max = 10.0.
        Diagonal elements default to 0.

     2. Distance between nodes; this is calculated by searching back
        from the source and dest nodes until a common ancestor is found
        then adding the distance from source and dest to the common
        ancestor.  Diagonal elements default to 0.

     3. number of nodes on path between source and dest; this is
        calculated by searching back from source and dest to a common
        ancestor, counting 1 for each branch.  The distance is one less
        than the sum of the counts.  Diagonal elements default to 0.

     4. the ratio of the shared branch length between nodes (from the
        root) and the average of the total branch lengths (from the root)
        to each node.  This is calculated by finding the common
        ancestor, calculating its height from the root, multiplying it
        by two and dividing by the sum of the root distances of the
        two nodes. Diagonal elements default to 1.

     5. the shared branch length between nodes (calculated as in number
        4).  Diagonal elements default to the height of the node.

    If the two nodes are the same, the function returns the product of
    the node's distance to its ancestor and the selfDistance factor.
   {**********************************************************************}
   Function calcDistance(root       : nodePtr;
                         sourceNode : nodePtr;
                         destNode   : nodePtr;
                         trait      : integer; { tipdiff trait number 1 or 2}
                         userChoice : userOptions) : extended;
      Var
         sourceAncestor : nodePtr;   {used in pathlength distance}
         destAncestor   : nodePtr;   {used in pathlength distance}
         sourceDist     : extended;  {distance from source to shared ancestor}
         destDist       : extended;  {distance from dest to shared ancestor}
         totalDist      : extended;  {sum of sourceDist and destDist}
         scale          : extended;  {tipdiff scaling factor}
         rootSource     : extended;  {distance from root to source node}
         rootDest       : extended;  {distance from root to dest node}
         rootShared     : extended;  {distance from root to shared ancestor}
         foundShared    : boolean;   {true when shared ancestor found}

      Begin
         scale := userChoice.scale[trait];
         Case userChoice.choice Of

            tipDiff :
               Begin
                 calcDistance := scale*abs(sourceNode^.dat[trait]-
                                              destNode^.dat[trait]);
               End; {case tipDiff}

            pathlength :                    {this code lifted from PDRANDOM}
               Begin
                  If (sourceNode = destNode) {self distance?}
                  Then Begin
                     If userChoice.selfDistConst
                     Then calcDistance := userChoice.selfDistance*scale
                     Else calcDistance := userChoice.selfDistance*
                                             sourceNode^.length*scale
                  End {then}
                  Else Begin
                     totalDist := -1.0;   {code for "common ancestor not found"}
                     sourceAncestor := sourceNode^.ancestor;
                     If sourceNode^.length>0
                     Then sourceDist := sourceNode^.length;
                     If (sourceAncestor <> root)
                     Then Begin
                        While ((sourceAncestor <> root) And
                               (totalDist < 0.0)) Do Begin
                           destAncestor := destNode^.ancestor;
                           If destNode^.length > 0
                           Then destDist := destNode^.length;
                           While ((destAncestor <> root) And (totalDist < 0.0)) Do Begin
                              If (destAncestor = sourceAncestor)
                              Then totalDist := sourceDist + destDist
                              Else Begin
                                 If destAncestor^.length > 0
                                 Then destDist := destDist +
                                                    destAncestor^.length;
                                 destAncestor := destAncestor^.ancestor
                              End {else}
                           End; {while}
                           If (totalDist < 0.0)
                           Then Begin
                              If sourceAncestor^.length > 0
                              Then sourceDist := sourceDist +
                                                   sourceAncestor^.length;
                              sourceAncestor := sourceAncestor^.ancestor;
                              If sourceAncestor = root
                              Then totalDist := sourceDist + destDist
                           End {then}
                        End {while}
                     End {then}
                     Else Begin
                        destAncestor := destNode^.ancestor;
                        If destNode^.length > 0
                        Then destDist := destNode^.length;
                        While ((destAncestor <> root) And
                               (totalDist < 0.0)) Do Begin
                           If (destAncestor = sourceAncestor)
                           Then totalDist := sourceDist + destDist
                           Else Begin
                              If destAncestor^.length>0
                              Then destDist := destDist + destAncestor^.length;
                              destAncestor := destAncestor^.ancestor
                           End {else}
                        End; {while}
                        If (totalDist < 0.0)
                        Then Begin
                           If destAncestor^.length > 0
                           Then destDist := destDist + destAncestor^.length;
                           totalDist := sourceDist + destDist
                        End {then}
                     End; {else}
                     calcDistance := totalDist*scale;
                  End {else}
               End; {case pathlength}

            nodeLength :   {mostly copied from the pathlength code, but all branches = 1}
               Begin
                  If (sourceNode = destNode) Then Begin
                     If userChoice.selfDistConst
                     Then calcDistance := userChoice.selfDistance
                     Else calcDistance := 0
                  End {then}
                  Else Begin
                     totalDist := -1.0;
                     sourceAncestor := sourceNode^.ancestor;
                     If sourceNode^.length>0 Then sourceDist := 1;
                     If (sourceAncestor <> root)
                     Then Begin
                        While ((sourceAncestor <> root) And
                               (totalDist < 0.0)) Do Begin
                           destAncestor := destNode^.ancestor;
                           If destNode^.length > 0
                           Then destDist := 1;
                           While ((destAncestor <> root) and (totalDist < 0.0)) Do Begin
                              If (destAncestor = sourceAncestor)
                              Then totalDist := sourceDist + destDist
                              Else Begin
                                 If destAncestor^.length > 0
                                 Then destDist := destDist + 1;
                                 destAncestor := destAncestor^.ancestor
                              End {else}
                           End; {while}
                           If (totalDist < 0.0)
                           Then Begin
                              If sourceAncestor^.length > 0
                              Then sourceDist := sourceDist + 1;
                              sourceAncestor := sourceAncestor^.ancestor;
                              If sourceAncestor = root
                              Then totalDist := sourceDist + destDist
                           End {then}
                        End {while}
                     End {then}
                     Else Begin
                        destAncestor := destNode^.ancestor;
                        If destNode^.length > 0
                        Then destDist := 1;
                        While ((destAncestor <> root) And
                               (totalDist < 0.0)) Do Begin
                           If (destAncestor = sourceAncestor)
                           Then totalDist := sourceDist + destDist
                           Else Begin
                              If destAncestor^.length>0
                              Then destDist := destDist + 1;
                              destAncestor := destAncestor^.ancestor
                           End {else}
                        End; {while}
                        If (totalDist < 0.0)
                        Then Begin
                           If destAncestor^.length > 0
                           Then destDist := destDist + 1;
                           totalDist := sourceDist + destDist
                        End {then}
                     End; {else}
                     calcDistance := totalDist-1;
                  End {else}
               End; {case nodelength}

            shared :
               Begin
                  If (sourceNode = destNode) 
                  Then calcDistance := userChoice.selfDistance*scale
                  Else Begin
                     foundShared := false;
                     sourceAncestor := sourceNode^.ancestor;
                     If (sourceAncestor <> root)
                     Then Begin
                        While ((sourceAncestor <> root) And
                               not(foundShared)) Do Begin
                           destAncestor := destNode^.ancestor;
                           While ((destAncestor <> root) And
                                  Not(foundShared)) Do Begin
                              If (destAncestor = sourceAncestor)
                              Then foundShared := true
                              Else destAncestor := destAncestor^.ancestor;
                           End; {while}
                           If Not(foundShared)
                           Then Begin
                              sourceAncestor := sourceAncestor^.ancestor;
                              If sourceAncestor = root
                              Then foundShared := true
                           End {then}
                        End {while}
                     End {then}
                     Else Begin
                        destAncestor := destNode^.ancestor;
                        While ((destAncestor <> root) And
                               Not(foundShared)) Do Begin
                           If (destAncestor = sourceAncestor)
                           Then foundShared := true
                           Else destAncestor := destAncestor^.ancestor;
                        End; {while}
                     End; {else}
                     rootSource := branchHeight(root,sourceNode);
                     rootDest := branchHeight(root,destNode);
                     rootShared := branchHeight(root,destAncestor);
                     calcDistance := 2*rootShared/(rootDest+rootSource)*scale;
                  End {else}
               End; {case shared}

            covariance :
               Begin
                  If (sourceNode = destNode)
                  Then calcDistance := branchHeight(rt,sourceNode)*scale
                  Else Begin
                     foundShared := false;
                     sourceAncestor := sourceNode^.ancestor;
                     If (sourceAncestor <> root)
                     Then Begin
                        While ((sourceAncestor <> root) And
                               not(foundShared)) Do Begin
                           destAncestor := destNode^.ancestor;
                           While ((destAncestor <> root) And
                                  Not(foundShared)) Do Begin
                              If (destAncestor = sourceAncestor)
                              Then foundShared := true
                              Else destAncestor := destAncestor^.ancestor;
                           End; {while}
                           If Not(foundShared)
                           Then Begin
                              sourceAncestor := sourceAncestor^.ancestor;
                              If sourceAncestor = root
                              Then foundShared := true
                           End {then}
                        End {while}
                     End {then}
                     Else Begin
                        destAncestor := destNode^.ancestor;
                        While ((destAncestor <> root) And
                               Not(foundShared)) Do Begin
                           If (destAncestor = sourceAncestor)
                           Then foundShared := true
                           Else destAncestor := destAncestor^.ancestor;
                        End; {while}
                     End; {else}
                     rootShared := branchHeight(root,destAncestor);
                     calcDistance := rootShared*scale;
                  End {else}
               End; {case covariance}
         End {case}
      End; { function calcDistance }

   {*******************************************************************}
   {This function finds the maximum height of any node from the root.
    This is refered to as the height of the tree.  All it does is loop
    through the list of tips, calculate the height of each and save the
    maximum.                                                           }
   {*******************************************************************}
   Function treeHeight(rt,first : nodePtr) : extended;
      Var
         max : extended;
         temp : extended;
         curNode : nodePtr;
      Begin
        curNode := first;
        max := 0;
        While (curNode <> nil) Do Begin
           temp := branchHeight(rt,curNode);
           If temp>max Then max := temp;
           curNode := curNode^.tipR;
        End; {while}
        treeHeight := max;
      End; {function treeHeight}


   {*******************************************************************}
   {This procedure runs the scaling options screen and calculates the
    scaling factor.  The scaling factors are simply the desired maximum
    distance (currently 10), divided by the largest distance found for
    each trait.
   {*******************************************************************}
   Procedure doScaling(rt             :nodeptr;
                       first          :nodeptr;
                       Var userChoice :userOptions);
      Var
         outer    : nodePtr;
         inner    : nodePtr;
         maxDist1 : extended;
         maxDist2 : extended;
         temp     : extended;
         userChar : char;
         noScale  : boolean;
      Begin
         If (userChoice.choice = covariance)
         Then Begin
            userChoice.scale[1] := 1.0/treeHeight(rt,first);
            userChoice.scale[2] := userChoice.scale[1]
         End
         Else Begin
            maxDist1 := 0;
            maxDist2 := 0;
            outer := first;                {nested loops through the tips}
            While (outer <> nil) Do Begin
               inner := first;
               While (inner <> nil) Do Begin
                  {these require userChoice.scale[*] to be initialized to 1.0}
                  temp := calcDistance(rt,outer,inner,1,userChoice);
                  If (maxDist1 < temp)
                  Then maxDist1 := temp;
                  temp := calcDistance(rt,outer,inner,2,userChoice);
                  If (maxDist2 < temp)
                  Then maxDist2 := temp;
                  inner := inner^.tipR;
               End; {inner loop}
               outer := outer^.tipR;
            End; {outer loop}
            userChoice.scale[1] := 10.0/maxDist1;
            userChoice.scale[2] := 10.0/maxDist2;
         End; {else}
      End; {procedure doScaling}

   {*******************************************************************}
   {This procedure prompts the user for a number of choices:

     1. which of the four methods (metrics) shall be used to calculate
        distances

     2. whether to write results in a matrix or a column

     3. whether to include header information; for a matrix this means
        row and column labels, for a column, headers add two columns
        that specify source and destination.  Headers also contain
        the number of "tips."

   {*******************************************************************}
   Procedure getMetric(Var options : userOptions; first : nodePtr);
      Var metricChoice : integer;  {holds code entered by the user}
          selfChoice   : integer;  {menu choice for self distance option}
          userChar     : char;     {holds y/n character from user}

      {****************************************************************}
      {This local procedure handles the scaling questions.             }
      {****************************************************************}
      Procedure scalingOptions;
         Begin
            write('Do you want to scale the values to a maximum of 10.0 (Y/N)? ');
            readln(userChar);
            While Not(userChar In ['Y','y','N','n']) Do Begin
               write('Please enter Y or N: ');
               readln(userChar);
            End; {while}
            options.scaling := userChar In ['Y','y'];
            writeln;
            writeln('Do you want to write the distance values in a compact format');
            write('without exponents (Y/N)? ');
            readln(userChar);
            While Not(userChar In ['Y','y','N','n']) Do Begin
               write('Please enter Y or N: ');
               readln(userChar);
            End; {while}
            options.shortForm := userChar In ['Y','y'];
            If (options.shortForm And Not(options.scaling)) Then Begin
               writeln;
               writeln('Checking largest distance...');
               writeln;
               userChoice.scale[1] := 1.0; {scaling is always initialized}
               userChoice.scale[2] := 1.0;
               doScaling(rt,first,options);
               If ((userChoice.scale[1]<=0.1) Or
                   (userChoice.scale[2]<=0.1))
               Then Begin
                  writeln('Warning -- One or more distance values was greater than 99.9999');
                  writeln('           The values will be scaled to a maximum of 10.0');
                  writeln;
                  If (options.choice = tipDiff) Then Begin
                     writeln('           If this is not what you want, then you need to rescale the tip');
                     writeln('           data, e.g., by use of (T)ip data manipulations in the');
                     writeln('           Editing Module of PDTREE.');
                  End
                  Else Begin
                     writeln('           If this is not what you want, then you need to rescale the branch');
                     writeln('           length data, e.g., by use of (B)ranch length manipulations');
                     writeln('           in the Editing Module of PDTREE.');
                  End; {else}
                  options.scaling := true;
                  writeln;
                  writeln('Press enter to proceed');
                  readln;
               End; {then}
               userChoice.scale[1] := 1.0; {scaling is always initialized}
               userChoice.scale[2] := 1.0;
            End; {then}
         End; {local procedure scalingOptions}
      {********************************************************************}
      Begin {procedure getMetric}
         writeln;
         writeln('What kind of matrix do you want to calculate?');
         writeln;
         writeln(' (1) Pairwise difference of tip-data values');
         writeln('          Two matrices will be written (*.DS1, *.DS2).');
         writeln;
         writeln(' (2) Phylogenetic pathlength (patristic) distance (as in PDRANDOM)');
         writeln('          One matrix will be written (*.DST).');
         writeln;
         writeln(' (3) # of nodes along path between tips (e.g., Miles and Dunham, 1992)');
         writeln('          (diagonals will be zeros)');
         writeln('          One matrix will be written (*.DSN).');
         writeln;
         writeln(' (4) Shared branch length (from root to last common ancestor) divided by');
         writeln('     average of heights of the two tips (diagonals will be ones)');
         writeln('          One matrix will be written (*.DSH).');
         writeln;
         writeln(' (5) Shared branch length (as in matrix C of Garland and Ives (2000) and');
         writeln('          Blomberg, Garland, & Ives (2003)).');
         writeln('          (diagonals will be total branch length from root to tip)');
         writeln('          One matrix will be written (*.DSC).');
         writeln;
         metricChoice := 0;
         While (not (metricChoice in [1,2,3,4,5])) Do Begin
            write('Enter your choice: ');
            readln(metricChoice);
         End; {while}
         Case metricChoice Of
            1: options.choice := tipDiff;
            2: options.choice := pathLength;
            3: options.choice := nodeLength;
            4: options.choice := shared;
            5: options.choice := covariance;
         End; {case metricChoice}
         writeln;
         write('Would you like to write results in a (M)atrix or a (C)olumn (M/C)? ');
         readln(userChar);
         While Not(userChar In ['M','m','C','c']) Do Begin
            write('Please enter M or C: ');
            readln(userChar);
         End; {while}
         options.writeColumn := userChar In ['C','c'];
         writeln;
         If options.writeColumn
         Then write('Would you like a header in the file (Y/N)? ')
         Else write('Would you like a header and row and column labels in the file (Y/N)? ');
         readln(userChar);
         While Not(userChar in ['Y','y','N','n']) Do Begin
            write('Please enter Y or N: ');
            readln(userChar);
         End; {while}
         options.headers := userChar In ['Y','y'];
         writeln;
         userChoice.scale[1] := 1.0; {scaling is always initialized}
         userChoice.scale[2] := 1.0;
         Case options.choice Of
            tipDiff:
                Begin
                   writeln;
                   writeln('By default all self distances are set to zero.');
                   write('Would you like to change this (Y/N)? ');
                   readln(userChar);
                   If (userChar in ['Y','y']) Then Begin
                      write('What value shall the distance to self be? ');
                      readln(userChoice.selfDistance);
                   End {then}
                   Else userChoice.selfDistance := 0;
                   writeln;
                   scalingOptions
                End; {case tipDiff}

            pathLength :
                Begin
                   writeln;
                   writeln('By default all self distances are set to zero.');
                   write('Would you like to change this (Y/N)? ');
                   readln(userChar);
                   If (userChar in ['Y','y']) Then Begin
                      writeln;
                      writeln('There are two ways to assign self distances:');
                      writeln;
                      writeln('  1 Multiply the distance to the nearest ancestor by a constant');
                      writeln('  2 Set the self distance to a constant value');
                      selfChoice := 0;
                      While Not(selfChoice in [1,2]) Do Begin
                         writeln;
                         write('Enter a choice: ');
                         readln(selfChoice);
                      End; {while}
                      If (selfChoice = 1)
                      Then userChoice.selfDistConst := false
                      Else userChoice.selfDistConst := true;
                      writeln;
                      If (userChoice.selfDistConst)
                      Then write('Enter the constant self distance value: ')
                      Else write('Enter value to multiply the nearest ancestor distance by: ');
                      readln(userChoice.selfDistance)
                   End {then}
                   Else userChoice.selfDistance := 0;
                   writeln;
                   scalingOptions
                End; {case pathlength}

            nodeLength:
                Begin
                   writeln;
                   writeln('By default all self distances are set to zero.');
                   write('Would you like to change this (Y/N)? ');
                   readln(userChar);
                   If (userChar in ['Y','y']) Then Begin
                      write('What value shall the distance to self be (<1000)? ');
                      readln(userChoice.selfDistance);
                      While ((userChoice.selfDistance<0) Or
                             (userChoice.selfDistance>999)) Do Begin
                         write('Value out of range, please re-enter: ');
                         readln(userChoice.selfDistance)
                      End {while}
                   End {then}
                   Else userChoice.selfDistance := 0;
                   writeln;
                End; {case nodeLength}

            shared:
                Begin
                   writeln;
                   writeln('By default all self distances are set to one.');
                   write('Would you like to change this (Y/N)? ');
                   readln(userChar);
                   If (userChar in ['Y','y']) Then Begin
                      write('What value shall the distance to self be? ');
                      readln(userChoice.selfDistance);
                   End {then}
                   Else userChoice.selfDistance := 1;
                   writeln;
                   scalingOptions
                End; {case shared}

            covariance:
                Begin
                   userChoice.longest := treeHeight(rt,first);
                   writeln('Do you want to scale all values (diagonal and off diagonal) by');
                   write('dividing them by the maximum height of the tree (Y/N)? ');
                   readln(userChar);
                   While Not(userChar In ['Y','y','N','n']) Do Begin
                      write('Please enter Y or N: ');
                      readln(userChar);
                   End; {while}
                   options.scaling := userChar In ['Y','y'];
                   writeln;
                   writeln('Do you want to write the distance values in a compact format');
                   write('without exponents (Y/N)? ');
                   readln(userChar);
                   While Not(userChar In ['Y','y','N','n']) Do Begin
                      write('Please enter Y or N: ');
                      readln(userChar);
                   End; {while}
                   options.shortForm := userChar In ['Y','y'];
                   If (options.shortForm And
                       Not(options.scaling) And
                       (userChoice.longest > 99))
                   Then Begin
                      writeln;
                      writeln;
                      writeln('Warning -- One or more distance values was greater than 99.9999');
                      writeln('           The values will be scaled by the height of the tree.');
                      writeln;
                      writeln('           If this is not what you want, then you need to rescale the branch');
                      writeln('           length data, e.g., by use of (B)ranch length manipulations');
                      writeln('           in the Editing Module of PDTREE.');
                      options.scaling := true;
                      writeln;
                      writeln('Press enter to proceed');
                      readln;
                   End; {then}
                   If options.scaling Then Begin
                      userChoice.scale[1] := 1/userChoice.longest;
                      userChoice.scale[2] := userChoice.scale[1];
                   End; {then}
                End; {case covariance}
         End; {case options.choice}
      End; {procedure getmetric}

   {*******************************************************************}
   {This procedure is responsible for opening the output file(s),
    writing the results, and closing the file(s).                                         }
   {*******************************************************************}
   Procedure writeMatrix(flname : flstring;
                         rt     : nodePtr;
                         first  : nodePtr;
                         count  : integer;
                         choice : userOptions);

      {These constants are used for formatting tables and columns:
       short gaps appear in tables of :8:5 formatted values
       medium gaps appear in tables written with shortWrite (longer than :8:5)
       long gaps appear in tables written with extendWrite
       T gaps are in matrix titles
       T starts are the spaces between names and the first value in matrices
       D gaps are the spaces between values in matrices
       column T Gaps are used for formatting column titles}

      Const shortTGap = '       ';
            intTGap = '  ';
            longTGap = '                ';
            shortTStart = '       ';
            intTStart = '    ';
            longTStart = '           ';
            shortDGap = ' ';
            intDGap = ' ';
            longDGap = ' ';
            columnTGap = '   ';
      Var
         outFile1  : text;        {file for first trait}
         outFile2  : text;        {file for second trait}
         outer     : nodePtr;     {pointer for outer loop}
         inner     : nodePtr;     {pointer for inner loop}
         dist1     : extended;    {temp for holding value}
         dist2     : extended;    {temp for holding second value}
         dotPos    : integer;     {used for stripping extensions}
         tmpInt    : integer;
         outName1  : flstring;    {name for outFile1}
         outName2  : flstring;    {name for outFile2}
      Begin
         clrscr;
         If choice.choice = tipDiff
         Then writeln('Writing distance matrices to files:')
         Else writeln('Writing distance matrix to file:');
         {now generate the new file name(s)}
         dotpos := pos('.',flname);
         Case choice.choice Of     {replace the extensions}
            tipDiff    : Begin
                            outName1 := copy(flname,1,dotpos) + 'DS1';
                            outName2 := copy(flname,1,dotpos) + 'DS2';
                            writeln;
                            writeln('Trait1: ',outName1); {write the new names to the screen}
                            writeln('Trait2: ',outName2);
                         End; {case tipDiff}
            pathLength : Begin
                            outName1 := copy(flname,1,dotpos) + 'DST';
                            writeln(outName1);  {write the new names to the screen}
                         End; {case pathLength}
            nodeLength : Begin
                            outName1 := copy(flname,1,dotpos) + 'DSN';
                            writeln(outName1);  {write the new names to the screen}
                         End; {case nodeLength}
            shared     : Begin
                            outName1 := copy(flname,1,dotpos) + 'DSH';
                            writeln(outName1); {write the new name to the screen}
                         End; {case shared}
            covariance : Begin
                            outName1 := copy(flname,1,dotpos) + 'DSC';
                            writeln(outName1); {write the new name to the screen}
                         End; {case shared}
         End; {case choice.choice}
         assign(outFile1,outName1);       {Assign the file names}
         If choice.choice = tipDiff
         Then assign(outFile2,outName2);
         rewrite(outFile1);              {open the files for writing}
         If choice.choice = tipDiff
         Then rewrite(outFile2);
         If choice.headers Then Begin
            writeln(outFile1,count);     {Headers include the number of tips}
            If choice.choice = tipDiff
            Then writeln(outFile2,count);
            If choice.writeColumn Then Begin
               inner := first;          {list of tip names}
               While (inner <> nil) Do Begin
                  writeln(outFile1,inner^.name);
                  If choice.choice = tipDiff
                  Then writeln(outFile2,inner^.name);
                  inner := inner^.tipR;
               End {while}
            End
            Else Begin
               inner := first;     {write top row of tip names}
               If choice.choice = nodeLength
               Then write(outFile1,intTStart)
               Else Begin
                  If choice.shortForm Then Begin
                     write(outFile1,shortTStart);
                     If choice.choice = tipDiff
                     Then write(outFile2,shortTStart);
                  End
                  Else Begin
                     write(outFile1,longTStart);
                     If choice.choice = tipDiff
                     Then write(outFile2,longTStart);
                  End; {else}
               End; {else}
               While (inner <> nil) Do Begin
                  If choice.choice = nodeLength
                  Then write(outFile1,inner^.name,intTGap)
                  Else Begin
                     If choice.shortForm Then Begin
                        write(outFile1,inner^.name,shortTGap);
                        If choice.choice = tipDiff
                        Then write(outFile2,inner^.name,shortTGap);
                     End
                     Else Begin
                        write(outFile1,inner^.name,longTGap);
                        If choice.choice = tipDiff
                        Then write(outFile2,inner^.name,longTGap);
                     End; {else}
                  End; {else}
                  inner := inner^.tipR;
               End; {while}
               writeln(outFile1);
               If choice.choice = tipDiff Then writeln(outFile2);
            End {else}
         End; {then}
         outer := first;                {nested loops through the tips}
         If choice.writeColumn Then While (outer <> nil) Do Begin
            inner := first;
            While (inner <> nil) Do Begin
               dist1 := calcDistance(rt,outer,inner,1,choice);
               If choice.choice = tipDiff
               Then dist2 := calcDistance(rt,outer,inner,2,choice);
               If choice.choice = nodeLength Then Begin
                  tmpInt := round(dist1);
                  write(outFile1,tmpInt:3,intDGap)
               End {then}
               Else Begin
                  If choice.shortForm Then Begin
                     write(outFile1,dist1:8:5,shortDGap);
                     If choice.choice = tipDiff
                     Then write(outFile2,dist2:8:5,shortDGap);
                  End
                  Else Begin
                     extendWrite(outFile1,dist1);
                     write(outFile1,longDGap);
                     If choice.choice = tipDiff Then Begin
                        extendWrite(outFile2,dist2);
                        write(outFile2,longDGap);
                     End {then}
                  End; {else}
               End; {else}
               writeln(outFile1);
               If choice.choice = tipDiff Then writeln(outFile2);
               inner := inner^.tipR;
            End; {while inner <> nil}
            outer := outer^.tipR;
         End {then/while outer <> nil}
         Else While (outer <> nil) Do Begin {write a matrix}
            If choice.headers Then Begin                {write tip names}
               If choice.shortForm Then Begin
                  write(outFile1,outer^.name,shortDGap);
                  If choice.choice = tipDiff
                  Then write(outFile2,outer^.name,shortDGap);
               End {then}
               Else Begin
                  write(outFile1,outer^.name,longDGap);
                  If choice.choice = tipDiff
                  Then write(outFile2,outer^.name,longDGap);
               End; {else}
            End; {then}
            inner := first;
            While (inner <> nil) Do Begin                    {write data}
               dist1 := calcDistance(rt,outer,inner,1,choice);
               If choice.choice = tipDiff
               Then dist2 := calcDistance(rt,outer,inner,2,choice);
               If choice.choice = nodeLength
               Then Begin
                  tmpInt := round(dist1);
                  write(outFile1,tmpInt:3,intDGap);
               End {then}
               Else Begin
                  If choice.shortForm Then Begin
                     write(outFile1,dist1:8:5,shortDGap);
                     If choice.choice = tipDiff
                     Then write(outFile2,dist2:8:5,shortDGap);
                  End {then}
                  Else Begin
                     extendWrite(outFile1,dist1);
                     write(outFile1,longDGap);
                     If choice.choice = tipDiff Then Begin
                        extendWrite(outFile2,dist2);
                        write(outFile2,longDGap);
                     End {then}
                  End; {else}
               End; {else}
               inner := inner^.tipR;
            End; {while inner <> nil}
            writeln(outFile1);              {start the next line}
            If choice.choice = tipDiff
            Then writeln(outFile2);
            outer := outer^.tipR;
         End; {while outer <> nil}
         close(outFile1);                   {close both files}
         If choice.choice = tipDiff
         Then close(outFile2);
         ret('Done, press any key to exit');
      End; {function writeMatrix}

{**********************************************************************}
Begin
   introScreen;                            {show the screen}
   readTree(rt,flname);                    {read the tree}
   If (rt = nil)
   Then Begin
      clrscr;
      writeln('No tree loaded, exiting');
   End {then}
   Else Begin
      clrscr;
      preparetips(rt,tipCount,first);                    {sort the tips}
      getMetric(userChoice,first);                       {select the metric}
      If userChoice.scaling
      Then doScaling(rt,first,userChoice);
      writeMatrix(flname,rt,first,tipCount,userChoice);  {write the matrices}
   End; {else}
End. {program pddist}
