{********************************************************************}
{PDError.pas 
 This program copyright 1996,1998,1999.
 Written by Peter E. Midford for Theodore Garland, Jr., with
 extensive copying from PDTREE.PAS.
 Last modified 10 December 1999.}
{********************************************************************}

{This program explores the effect of changing branch lengths on
 correlations in simulated evolution data sets (e.g., output from 
 PDSIMUL).}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


{Important note in code: earlier versions of the program referred to adding
 errors to branch lengths as tweaking.  The user dialog now uses the
 terminology of Diaz-Uriarte and Garland 1996, but the term 'tweak' occurs
 throughout the code. PEM 10-Dec-1999}



Program PDError(input,output);
uses e_utils,                 {Utility functions}
     e_io,                    {For reading trees}
     e_stats,                 {statistics functions}
     dos,                     {file name spliting}
     crt;                     {Screen control functions}

Const minLength = 0.000001;   {Minimum tweaked branch length}

      stepSize = 0.1;         {used in stepping through parameters}
                              {Change the step size here!}
                              {If you make the size smaller, you may also}
                              {want to change the real number formatting}
                              {in writeParameters, in E_IO.PAS}

      loopTol = 0.001;        {used to make sure the upper limit is}
                              {included when appropriate.  Should always}
                              {be 1-2 orders of magnitude smaller than}
                              {stepSize.  Brought to you by the limitations}
                              {of floating point arithmetic!}

Var inputTree :nodePtr;                  {The tree read from the .SIM file}
    simData :nodeDataPtr;                {A data set read from the .SIM file}
    analysisTree :nodePtr;               {tree for analysis (may <> inputTree)}
    xForms : xFormPtr;                   {set of transform options}
    simFile :text;                       {The simulation (.SIM) file}
    resultFile :text;                    {The output file}
    tweaksFile :text;                    {The file for saving tweaked trees}
    paramFile :text;                     {The file that saves parameter choices}
    theTweak :tweakPtr;                  {How to modify the tree}
    tweakedtree :nodePtr;                {New tree resulting from tweaking branch lengths}
    tree1,tree2 :nodePtr;                {Trees resulting from best tweak for each trait}
    bestBLContrasts :bLConArray;         {Array of best contrasts for each trait}
    savedSeed :longint;                  {State of the random number generator}
    simStats :statpak;                   {Statistical results and intermediates}
    correlations :corrArray;             {A set of correlations (columns 12-20)}
    curTree :nodePtr;                    {Tree generated to test a transform value}
    curXform :extended;                  {Code for branch length transform (-0.1 - 2.0)}
    curRho :extended;                    {Rho value for branch length transform 0-2.0)}
    bestCorr1,bestCorr2 :extended;       {Intermediates in transform search}
    bestRhoCorr1,bestRhoCorr2 :extended; {Intermediates in rho transform search}
    treeHeight :extended;                {Height of the tree for rho transform}
    corrT1,corrT2 :extended;             {Correlation scores in transform search}
    rho1,rho2 :extended;                 {Best rho transform for each trait}
    t1Param, t2Param :extended;          {Best transforms for each trait}
    t1Display, t2Display :extended;      {Numbers written for transforms}
    tmpFloat :extended;                  {used for reading and swapping transform endpoints}
    rhoStart,rhoEnd :extended;           {Bracket values for the Rho transform}
    expStart,expEnd :extended;           {Bracket values for the exponential transform}
    globalMin : traits;                  {minimum values for traits for all sets}
    globalMax : traits;                  {maximum values for traits for all sets}
    doCategories : traitFlags;
    doRanks : traitFlags;
    userLimit :integer;                  {Optional limit to number of data sets to process}
    contrastCount :integer;              {Number of contrasts in the array}
    tweaksToSave :integer;               {Number tweaked trees to save (if any) }
    tipCount : integer;                  {Holds the number of tips}
    binCount : itraits;                  {Number of categories for categorization}
    binBase  : itraits;                  {Value assigned to tips in the smallest bin (O or 1) }
    simNumber :integer;                  {Index of current data set in input file}
    tmpInt :integer;                     {Used for swapping things}
    tipStat : statpak;                   {Holds descriptive tip statistics }
    succ :boolean;                       {Indicates a successful file operation}
    doTweak :boolean;                    {Whether to tweak the tree}
    showCategories :boolean;             {Whether to display categorization results}
    blLTE1 :boolean;                     {Disables checking log10 transforms if branches shorter than 1 are present}
    tipsOnly : boolean;                  {using a second tree, so only add data to tips}
    zeroLenTip : boolean;                {flag for adjacent zero length tip branches}
    t1type,t2type :char;                 {Character codes for best transform type}
    userChar : char;                     {For reading user responses}
    userStr : string;                    {For reading for multiple character responses}
    tmpStr : string;                     {For formatting error messages}
    inputFlName :flString;               {The name of the input file}
    rootName :flString;                  {The name component of the result file}
    secondFileName :flString;
    resultFileName :flString;            {The complete file name of the result file}
    tweakFileName :flString;             {The name of the file for saving tweaked trees}
    paramFileName :flString;             {The name of the file that saves user parameters}

{*********************************************************************}
{This procedure undoes the reassignment of input and output to the
{crt device.  See page 199 of the Turbo Pascal programmer's manual.   }
{*********************************************************************}
Procedure resetCrtAssignments;
   Begin
      Assign(input,'');
      reset(input);
      assign(output,'');
      rewrite(output);
   End;

{*********************************************************************}
{This procedure returns the assignment of input and output to the CRT }
{device.  Compare with the preceeding procedure.
{*********************************************************************}
Procedure restoreCrtAssignments;
   Begin
      assignCrt(input);
      reset(input);
      assignCrt(output);
      rewrite(output);
   End;

{**************************************************************************}
{This procedure checks that there are no adjacent tip nodes with zero
{branch length.}
{**************************************************************************}
Procedure checkZeroTips(nd:nodePtr; Var zeroTip:boolean);
   Begin
      If nd^.sibL<>nil Then Begin
         checkZeroTips(nd^.sibL,zeroTip);
         checkZeroTips(nd^.sibR,zeroTip);
      End {then}
      Else Begin
         If (((nd^.ancestor^.sibL^.sibL=nil)
              And (nd^.ancestor^.sibR^.sibL=nil))
              And ((nd^.ancestor^.sibL^.length=0) And
                  (nd^.ancestor^.sibR^.length=0)))
         Then zeroTip:=true;
      End; {else}
   End; {procedure checkzerotips}

{**************************************************************************}
{This function reads a second tree and validates it.  If the names and
 numbers of tips do not match, a error message is printed and the user
 can either read another file or give up and use the simTree, which this
 procedure will return.                                                    }
{**************************************************************************}
Procedure treeSubst(Var analysisTree : nodePtr;
                    Var analysisTreeName : flString;
                    simTree : nodePtr;
                    simSize : integer);
   Var newTree : nodePtr;
       secondFile : text;
       secondCount : integer;
       newName : pathStr;
       nameDir : dirStr;
       nameName : nameStr;
       nameExt : extStr;
       treeOK  : boolean;
       zeroLenTips : boolean;
       success : boolean;
   {***********************************************************************}
   {This function checks that the tip names match between the orginal tree }
   {and the optional extra tree.                                           }
   {***********************************************************************}
   Function checkTipNames(checkTree : nodePtr;
                          templateTree : nodePtr) : boolean;
      Var result : boolean;
          dummy : nodePtr;
      Begin
         If (checkTree^.sibL = NIL) Then Begin
            checkName(templateTree,checkTree^.name,result,dummy);
            result := not(result);
         End
         Else Begin
            result := checkTipNames(checkTree^.sibL,templateTree);
            If result
            Then result := checkTipNames(checkTree^.sibR,templateTree);
         End; {else}
         checkTipNames := result;
      End; {function checktipnames}
   {***********************************************************************}
   Begin {procedure treeSubst}

      treeOK := false;
      newTree := NIL;
      Repeat
         Repeat
            write('Enter the file name (including extension) for the second tree: ');
            readln(newName);
            fsplit(newName,nameDir,nameName,nameExt);
            If ((upCase(nameExt[2]) = 'I') And
                (upCase(nameExt[3]) = 'N') And
                (upCase(nameExt[4]) = 'P')) Then Begin
               assign(secondFile,newName);
               readSecondTree(newTree,secondFile,secondCount);
               treeOK := (newTree<>NIL);
               If treeOK Then close(secondFile);
            End
            Else Begin
               writeln('Tree file must be in .INP format');
               treeOK := false;
            End
         Until treeOK;
         If newTree <> simTree Then Begin
            zeroLenTips := false;
            checkZeroTips(newTree,zeroLenTips);
            If zeroLenTips Then Begin
               writeln('The new tree has two adjacent tips with branch lengths of zero.');
               writeln('PDERROR does not allow this.  Would you like to try another tree (Y/N=Enter)?');
               readln(userChar);
               If userChar in ['Y','y']
               Then treeOK := false
               Else Begin
                  writeln('PDERROR will exit.');
                  halt;
               End; {else}
            End;
            If treeOK Then Begin
               success := (simSize = secondCount);

               If success
               Then Begin
                  success := checkTipNames(newTree,simTree);
                  If not(success) Then Begin
                     writeln('The tip names of the new tree do not match.');
                     write('Do you want to try another tree (Y/N=Enter)? ');
                     readln(userChar);
                     If userChar in ['Y','y'] Then treeOK := false;
                  End
               End
               Else Begin
                  write('The number of tips of the new tree ',secondCount);
                  writeln(' do not match the number of tips from the SIM file ',simSize);
                  writeln('Do you want to try another tree (Y/N=Enter)? ');
                  readln(userChar);
                  If userChar in ['Y','y'] Then treeOK := false;
               End {else}
            End {then}
         End; {then}
      Until treeOK;
      If success
      Then Begin
         analysisTree := newTree;
         analysisTreeName := newName;
         writeln;
         writeln('Second tree successfully loaded from ',newName);
         writeln;
      End
      Else analysisTree := simTree
   End; {procedure substTree}

{**************************************************************************}
{This procedure prompts the user for a tweak (added error) type and
 value(s). Prior to doing so, it displays some simple statistics about the t
 ree's branch lengths.  The results are returned in a freshly allocated
 tweak structure.}
{**************************************************************************}
Procedure getTweak(Var newTweak :tweakPtr; theTree :nodePtr);

   Var tweakChar :char;           {Character for user choice of tweak type}
       minBranch,maxBranch,sumBranch :extended;
       countBranch :integer;
       fileStr :flString;

   {***********************************************************************}
   {This local procedure calculates descriptive statistics (min,max) for
    the branches in the tree. These may be useful to the user in selecting
    a tweaking method.                                                     }
   {***********************************************************************}
   Procedure branchDescriptives(theNode :nodePtr;
                                theRoot :nodePtr;
                                Var minBranch,maxBranch,sumBranch :extended;
                                Var countBranch :integer);
      Begin
         If (theNode^.sibL <> NIL) Then
            branchDescriptives(theNode^.sibL,
                               theRoot,
                               minBranch,
                               maxBranch,
                               sumBranch,
                               countBranch);
         If (theNode^.sibR <> NIL) Then
            branchDescriptives(theNode^.sibR,
                               theRoot,
                               minBranch,
                               maxBranch,
                               sumBranch,
                               countBranch);
         If (theNode <> theRoot) Then Begin
            inc(countBranch);
            sumBranch := sumBranch + theNode^.length;
            If (theNode^.length < minBranch) Then
               minBranch := theNode^.length;
            If (theNode^.length > maxBranch) Then
               maxBranch := theNode^.length;
         End {if}
      End; {procedure branchDescriptives}
   {*************************************************************************}
   Begin
      writeln;
      writeln;
      minBranch := 1E500;
      maxBranch := 0;
      sumBranch := 0;
      countBranch := 0;
      branchDescriptives(theTree,
                         theTree,
                         minBranch,
                         maxBranch,
                         sumBranch,
                         countBranch);
      write('Minimum Branch Length: ');
      extendWrite(output,minBranch);
      writeln;
      write('Maximum Branch Length: ');
      extendWrite(output,maxBranch);
      writeln;
      write('Mean Branch Length: ');
      extendWrite(output,sumBranch/countBranch);
      writeln;
      writeln;
      new(newTweak);
      newTweak^.tipLength := minLength;
      tweakChar := ' ';
      writeln('Do you want add a (F)ixed error or a relative (N)ormal error?');
      writeln;
      While (not (tweakChar in ['F','N'])) Do Begin
         write('Enter either F or N: ');
         readln(tweakChar);
         writeln;
         tweakChar := upCase(tweakChar);
         If (tweakChar = 'N') Then newTweak^.myType := NORMAL;
         If (tweakChar = 'F') Then newTweak^.myType := FIXED;
      End; {while}

      Case (newTweak^.myType) of

         FIXED:
            Begin
               write('Enter the fixed error multiplier you want applied to each branch length: ');
               readln(newTweak^.deviation);
               writeln;
               While (newTweak^.deviation <= 0) Do Begin
                 writeln('The fixed error multiplier must be positive.');
                 write('Enter the error multiplier: ');
                 readln(newTweak^.deviation);
                 writeln;
               End; {while}
            End; {case FIXED}

         NORMAL:
            Begin
               writeln('Enter the coefficient of variation of the relative normal distribution ');
               writeln('      as a percent.');
               writeln('Warning -- Values larger than 100 may cause problems for some transforms.');
               write('Enter % CV: ');
               readln(newTweak^.width);
               writeln;
               newTweak^.width := newTweak^.width/100.0;
            End; {case NORMAL}

      End; {case newTweak^.mytype}
      If (newTweak^.myType = NORMAL) Or (newTweak^.deviation > 1) Then Begin
         tweakChar := ' ';
         While ((tweakChar > '2') Or (tweakChar < '0')) Do Begin
            writeln('If a branch length is set to a value less than zero,');
            writeln('do you want to set the length to: ');
            writeln;
            writeln('0.000001     Press 0 (disallows log10 transform because of - branch lengths)');
            writeln('1.1          Press 1');
            writeln('Other value  Press 2');
            write('Enter choice (0, 1 or 2): ');
            readln(tweakChar);
            writeln;
         End; {while}
         If (tweakChar = '0') Then newTweak^.zeroLength := minLength;
         If (tweakChar = '1') Then newTweak^.zeroLength := 1.1;
         If (tweakChar = '2') Then Begin
            write('Enter the value you want negative branch changed to: ');
            readln(newTweak^.zeroLength);
            writeln;
         End; {then}
         {This should fix a problem with zero length adjacent branches PEM 12-22-1999}
         If newTweak^.zeroLength > 0
         Then newTweak^.tipLength := newTweak^.zeroLength;
         If (newTweak^.zeroLength < 1.0)
         Then Begin
            writeln('Because the value you selected is less than one, log10 branch length');
            writeln('transformations will not be allowed.');
            writeln;
         End; {then}
      End; {then}
   End; {procedure gettweak}

{**************************************************************************}
{This recursive procedure creates a copy of an input tree with each
 branch stochastically tweaked according to the tweak.}
{**************************************************************************}
Function tweakTree(oldTree :nodePtr; how :tweakPtr) :nodePtr;
   Var newTree :nodePtr;
   {**************************************************************************}
   {Calculates a new stochastic branch length given the old value and a tweak
    type.}
   {**************************************************************************}
   Procedure tweakNode(Var newLength :extended;
                           oldLength :extended;
                           how :tweakPtr);
      Var normal1,normal2 :extended;
          normalDev :extended;
      Begin
         Case how^.myType of
            NORMAL:
               Begin
                  If (oldLength > 0) Then Begin
                     getUnivariates(normal1,normal2);
                     normalDev := how^.width*oldlength*normal1;
                     newLength := oldLength+normalDev;
                     If (newLength < 0) Then newLength := how^.zeroLength
                  End
                  Else newLength := 0
               End;

            FIXED:
               Begin
                  If (random<0.5) Then
                     newLength := (1 - how^.deviation) * oldLength
                  Else newlength := (1 + how^.deviation) * oldlength;
                  If (newLength < 0) Then newLength := how^.zeroLength;
               End;
         End;{case how^.myType}
      End; {procedure tweaknode}
   {*************************************************************************}
   Begin {procedure tweaktree}
      new(newTree);
      initMe(newTree);
      newTree^.name := oldTree^.name;
      tweakNode(newTree^.length,oldTree^.length,how);
      If oldTree^.sibL = NIL  {it's a tip!}
      Then Begin
         If newTree^.length <= 0
         Then newTree^.length := how^.tipLength;
      End
      Else Begin                {internal node}
         newTree^.sibL := tweakTree(oldtree^.sibL,how);
         newTree^.sibL^.ancestor := newTree;
         newTree^.sibR := tweakTree(oldtree^.sibR,how);
         newTree^.sibR^.ancestor := newTree
      End; {else}
      tweaktree := newtree
   End; {procedure tweakTree}

{*********************************************************************}
{This function checks the tree for branch lengths less than or equal to
 one and returns true if it finds any.  This function is used to avoid
 problems that occur if the user wants to do a log transform on the
 tree.  Note that a log transform of a negative branch length is an
 error, while the log transform of a branch length less than one yields
 a negative branch length which will cause more subtle problems
 resulting from bogus contrast calculations.  PEM 13 May 1996.
 This function now allows internal branches of length = 1.  The main
 purpose of this function is to "guard" the log10 transform from
 creating impossible branch lengths. PEM 17 Dec 1999}
{*********************************************************************}
Function branchesLTE1(tree :nodePtr; rt :nodePtr) :boolean;
   Var result :boolean;
   Begin
      result := false;
      If (tree <> rt)
      Then If (tree^.length < 1.0) Then result := true;
      If ((tree^.sibl^.sibl = nil) And
          (tree^.sibr^.sibl = nil)) Then Begin   {check for adjecent tips}
          If ((tree^.sibl^.length = 1) And
              (tree^.sibr^.length = 1)) Then result := true;
      End;
      If (not(result) and (tree^.sibl <> nil))
      Then result := branchesLTE1(tree^.sibL,rt);
      If (not(result) and (tree^.sibR <> nil))
      Then result := branchesLTE1(tree^.sibR,rt);
      branchesLTE1 := result;
   End; {function branchesLessThan1}

{*********************************************************************}
{This procedure categorizes a list of data according to the values of
 binCount and binBase.                                                }
{*********************************************************************}
Procedure categorizeData(data       : nodeDataPtr;
                         tree       : nodePtr;
                         binCount   : itraits;
                         binBase    : itraits;
                         globalMin  : traits;
                         globalMax  : traits);
   Var temp :nodeDataPtr;
       want :nodePtr;
       notFound :boolean;
   {********************************************************************}
   {This function determines which bin the value should be assigned to. }
   {********************************************************************}
   Function computeBin(value    :extended;
                       min      :extended;
                       max      :extended;
                       binCount :integer) : integer;
      Begin
         computeBin := round((binCount-1)*(value-min)/(max-min));
      End; { computeBin }
   {**********************************************************************}
   Begin { procedure categorizeData }
      If binCount[1] > 0 { if binCount is zero then do nothing}
      Then Begin
         temp := data;
         While (temp <> nil) Do Begin
            notFound := true;
            checkName(tree,temp^.name,notFound,want); {search for node with temp's name}
            temp^.data[1] := binBase[1] + computeBin(temp^.data[1], {do the transform}
                                                     globalMin[1],
                                                     globalMax[1],
                                                     binCount[1]);
            temp := temp^.next;
         End; { while }
      End; {then}
      If binCount[2] > 0 { if binCount is zero then do nothing}
      Then Begin
         temp := data;
         While (temp <> nil) Do Begin
            notFound := true;
            checkName(tree,temp^.name,notFound,want);
            temp^.data[2] := binBase[2] + computeBin(temp^.data[2],
                                                     globalMin[2],
                                                     globalMax[2],
                                                     binCount[2]);
            temp := temp^.next;
         End; { while }
      End; { then }
   End; { procedure categorizeData}

{*********************************************************************}
{ This procedure ranks a list of data in ascending order.  The
  algorithm is o(n^2) (so it could be improved), but it does what is
  required: the sum of the assigned ranks is always equal to the
  sum of the integers from 1 to tipCount.  This procedure differs from
  the similar procedure in PDTIPS because PDERROR keeps all nodes,
  interior nodes as well as tips in the node data list (PDTIPS
  collapses interior node data into the tip node that preceeds it in
  the line of simulated data from the sim file).  Because PDERROR
  cares about simulated data from interior nodes, it keeps this data
  in its own node data structures.  However, data from interior nodes
  shouldn't be included in the ranking (or the categorization), so
  PDERROR skips alternate list enteries which will always be interior
  nodes, or the bogus node called '**'.
{*********************************************************************}
Procedure rankData(data :nodeDataPtr;
                   tipCount : integer;
                   doRanks : traitFlags);
   Var orderArray : array [1..maxTips] of nodeDataPtr;
       maxData : extended;
       minData : extended;
       rank    : extended;
       index   : integer;
       start   : integer;
   {********************************************************************}
   {This local function resets the marks of all the nodes on the list.  }
   {These marks are used to pull each node out of the list only once,   }
   {when the node has the lowest trait value among the unmarked nodes   }
   {on the list.  Although the marks of interior nodes don't need to be }
   {reset, because they are never used, it doesn't hurt (much) to set   }
   {them to a known value.                                              }
   {********************************************************************}
   Procedure resetMarks(data : nodeDataPtr);
      Var current : nodeDataPtr;
      Begin
         current := data;
         While (current <> nil) Do Begin
            current^.mark := false;
            current := current^.next;
         End; {while}
      End; {procedure resetMarks}
   {********************************************************************}
   {This function returns the node with the lowest trait value among the}
   {unmarked tip nodes in the list.  This is the first function that    }
   {actually skips interior nodes.  Surprisingly, it is the only one    }
   {that really needs to skip the interior nodes.  [This may be an      }
   {advantage to the otherwise inefficient O(n^2) sort that this        }
   {function implements.                                                }
   {********************************************************************}
   Function smallestData(data :nodeDataPtr;
                         index :integer)   {selects trait 1 or 2}
                                        : nodeDataPtr;
      Const largeValue = 9.999E99;
      Var temp :nodeDataPtr;
          min : extended;
          winner : nodeDataPtr;
      Begin
         temp := data;
         min := largeValue;
         While (temp <> nil) Do Begin
            If (Not(temp^.mark) And (temp^.data[index] < min))
            Then Begin
               min := temp^.data[index];
               winner := temp;
            End; {then}
            temp := temp^.next^.next; {skip interior node}
         End; { while}
         winner^.mark := true;
         smallestData := winner;
      End; { smallestData }
   {********************************************************************}
   {This local procedure assigns rank values to the index-th node in the}
   {orderArray.  Since the orderArray contains sorted tip values, there }
   {is no need to skip.  The only messiness here is calculating the     }
   {ranks of tied values (which is why a simple sort isn't good enough. }
   {********************************************************************}
   Procedure assignRank(trait : integer);
      Var rank : extended;
          i : integer;
      Begin
         If (start = 0)
         Then orderArray[index]^.rank[trait] := index
         Else Begin
            rank := (start+index)/2.0;
            For i := start To index Do orderArray[i]^.rank[trait] := rank
         End; {else}
         start := 0;
      End; {procedure assignRank}
   {********************************************************************}
   Begin { procedure rankData }
      resetMarks(data);
      If (1 in doRanks) Then Begin
         For index := 1 To tipCount Do orderArray[index] := smallestData(data,1);
         start := 0;
         For index := 1 To tipCount Do Begin
            If ((index < tipCount) And
                (orderArray[index+1]^.data[1] = orderArray[index]^.data[1]))
            Then Begin
               If (start = 0) Then start := index;
            End
            Else assignRank(1);
         End; {for i}
         resetMarks(data);
      End; {then}
      If (2 in doRanks) Then Begin
         For index := 1 To tipCount Do orderArray[index] := smallestData(data,2);
         start := 0;
         For index := 1 To tipCount Do Begin
            If ((index < tipCount) And
                (orderArray[index+1]^.data[2] = orderArray[index]^.data[2]))
            Then Begin
               If (start = 0) Then start := index;
            End
            Else assignRank(2);
         End; {for i}
      End; {then}
   End; { procedure rankData}

{**************************************************************************}
{This function assigns each node rank values to their data fields.  As in  }
{the above code that generates the ranks, this procedure exploits the      }
{alternating structure of the node data lists to only update the data      }
{fields of tip nodes.                                                      }
{**************************************************************************}
Procedure assignRanks(data : nodeDataPtr; doRanks : traitFlags);
   Var temp : nodeDataPtr;
   Begin
      temp := data;
      While (temp <> nil) Do Begin
         If (1 in doRanks) Then temp^.data[1] := temp^.rank[1];
         If (2 in doRanks) Then temp^.data[2] := temp^.rank[2];
         temp := temp^.next^.next;       {skip next node, which isn't a tip}
      End; {while}
   End; {assignRanks}

{**************************************************************************}
{Assigns each data value and assigns it to the node within the tree with
 the same name.                                                            }
{**************************************************************************}
Procedure addData(node :nodePtr; data :nodeDataPtr; tipsOnly :boolean);
  Var myData :nodeDataPtr;
  {************************************************************************}
  {Find the data element with the given name.                              }
  {************************************************************************}
  Function lookupData(name :nameString; dataList :nodeDataPtr) :nodeDataPtr;
     Var temp   :nodeDataPtr;
         done   :boolean;
     Begin
        temp := dataList;
        done := false;
        While ((temp <> nil) And Not(done)) Do Begin
           If (temp^.name = name)
           Then done := true
           Else temp := temp^.next
        End; {while}
        lookupData := temp;
        If (done = false)
        Then writeln('Lookup Error in AddData!: ',name);
     End; { function lookupData }
  {***********************************************************************}
  Begin {procedure adddata}
     If (node <> nil)
     Then Begin
        If ((node^.sibL = NIL) Or (Not(tipsOnly))) Then Begin
           myData := lookupData(node^.name,data);
           If (data <> nil) Then Begin
              node^.dat[1] := myData^.data[1];
              node^.dat[2] := myData^.data[2];
           End; {then}
        End; {then}
        addData(node^.sibL,data,tipsOnly);
        addData(node^.sibR,data,tipsOnly);
     End; {then}
  End; {procedure adddata}


{*******************************************************************}
{This procedure transforms the branch lengths of a tree by the
 exponent passed in xForm (e.g., 0.5 is a square root).
{*******************************************************************}
Function xFormTree(tree :nodePtr; xForm :extended) :nodePtr;
   Var newTree :nodePtr;
   Begin {procedure xformtree}
      new(newTree);
      initMe(newTree);
      newTree^.name := Tree^.name;
      If (tree^.sibL<>nil) Then newTree^.sibL := xFormTree(tree^.sibL,Xform);
      If (tree^.sibR<>nil) Then newTree^.sibR := xFormTree(tree^.sibR,Xform);
      If ((xForm = 1.0) Or (tree^.length = 0))
      Then newTree^.length := tree^.length
      Else newTree^.length := exp(ln(Tree^.length)*xform);
      xFormTree := newTree
   End; {function xFormTree}

{*********************************************************************}
{ This procedure copies the tree, including ancestor links}
{*********************************************************************}
Function copyTree(tree :nodePtr) :nodePtr;
   Var newTree :nodePtr;
   Begin
      new(newTree);
      initMe(newTree);
      newTree^.name := tree^.name;
      newTree^.length := tree^.length;
      If (tree^.sibL<>nil) Then Begin
         newTree^.sibL := copyTree(tree^.sibL);
         newTree^.sibL^.ancestor := newtree
      End; {then}
      If (tree^.sibR<>nil) Then Begin
         newTree^.sibR := copyTree(tree^.sibR);
         newTree^.sibR^.ancestor := newtree
      End; {then}
      copyTree := newTree;
   End; {function copytree}

{*********************************************************************}
{This recursive procedure finds the (uncorrected) height of the tree
when passed the root.  It calls nothing but itself.}
{*********************************************************************}
Function findHeight(tree :nodeptr) :extended;
   Var heightL,heightR :extended;
   Begin
      If (tree^.sibL<>nil) Then Begin
         heightL := findheight(tree^.sibL);
         heightR := findheight(tree^.sibR);
         If (heightR>heightL) Then findheight := tree^.length + heightR
                              Else findHeight := tree^.length + heightL
      End {then}
      Else findHeight := tree^.length
   End; {function findheight}

{*******************************************************************}
{This procedure transforms the branch lengths of a tree according
to Grafen's Rho parameter.
{*******************************************************************}
Function rhoXFormTree(tree :nodePtr;
                      rho :extended;
                      height :extended) :nodePtr;
   Var newTree :nodePtr;
       multiplier :extended;
     {*********************************************************************}
      {This recursive procedure does the work.  It is given the scaling
      factor, and then traverses the tree, multiplying all the branch
      lengths by the correct amount. (from TE_BRNCH1.PAS}
      {*********************************************************************}
      Procedure scl(rt,nd :nodePtr; mult :extended);
         Begin
            If (nd<>rt) Then nd^.length := nd^.length*mult;
            If (nd^.sibL<>nil) Then Begin
               scl(rt,nd^.sibL,mult);
               scl(rt,nd^.sibR,mult)
            End; {then}
         End; {procedure scl}
      {*********************************************************************}
      {This recursive procedure traverses the tree and raises all the branch
      lengths to a power.}
      {*********************************************************************}
      Procedure expMe(rt,nd :nodePtr; power :extended);
         Begin
            If ((nd<>rt) and (nd^.length <> 0)) Then
              nd^.length := exp(power*ln(nd^.length))
            Else If (nd<>rt) Then nd^.length := 0;   {preserve polytomies!}
            If nd^.sibL<>nil Then Begin
                  expme(rt,nd^.sibL,power);
                  expme(rt,nd^.sibR,power)
            End {then}
         End; {procedure expMe}
      {**********************************************************************}
      {This recursive procedure transforms the tree so that the branch
      length field no longer holds the branch length, but holds the distance
      from the highest tip to itself.  It assumes that the height
      of the tree is one to start with.}
      {**********************************************************************}
      Procedure t1(rt,nd :nodePtr; ht :extended);
         Begin
            If (nd<>rt) Then ht := ht+nd^.length;
            If (nd^.sibL<>nil) Then Begin
                  t1(rt,nd^.sibL,ht);
                  t1(rt,nd^.sibR,ht);
            End; {then}
            If (nd<>rt) Then nd^.length := 1-(ht);
            If (nd^.length<1e-18) Then nd^.length := 0; {corrected 12-27-99 PEM}
         End; {procedure t1}
      {*********************************************************************}
      {This recursive procedure transforms the tree so that the branch
      length field holds the branch length assuming that it previously held
      the distance from the highest tip to the node below previously.
      It also assumes that the height of the tree is one to start with.}
      {*********************************************************************}
      Procedure t2(rt,nd :nodePtr);
         Begin
            If (nd^.sibL<>nil) Then Begin
               t2(rt,nd^.sibL);
               t2(rt,nd^.sibR)
            End; {then}
            If ((nd<>rt) and (nd<>rt^.sibR) and (nd<>rt^.sibL)) Then Begin
               nd^.length:=nd^.ancestor^.length-nd^.length
            End {then}
            Else If (nd<>rt) Then nd^.length := 1-nd^.length;
         End; {procedure t2}
   {***********************************************************************}
   Begin {function rhoXFormTree}
      newTree := copyTree(tree);
      If (rho <> 1) Then Begin
         multiplier := 1/height;
         scl(newTree,newTree,multiplier);
         t1(newTree,newTree,0);
         expme(newTree,newTree,rho);
         t2(newTree,newTree);
         scl(newTree,newTree,height)
      End; {then begin}
      rhoXFormTree := newTree
   End; {function rhoXFormTree}

{**********************************************************************}
{This procedure performs the log transform on the branch lengths of
 a tree.  It ignores zero length branches (polytomies), but can cause
 problems with fractional length branches, so there are checks
 scattered elsewhere throughout this application to avoid this function
 getting a tree with fractional length branches.
{**********************************************************************}
Function logXFormTree(tree :nodePtr) :nodePtr;
   Var newTree :nodePtr;
   Begin
      new(newTree);
      initMe(newTree);
      newTree^.name := Tree^.name;
      If (tree^.sibL<>nil) Then newTree^.sibL := logXFormTree(tree^.sibL);
      If (tree^.sibR<>nil) Then newTree^.sibR := logXFormTree(tree^.sibR);
      If (tree^.length = 0)       {check for polytomy}
      Then newTree^.length := 0
      Else newTree^.length := ln(Tree^.length)/ln(10.0);
      logXFormTree := newTree
   End; {function xFormTree}

{**********************************************************************}
{This function tries Grafen's arbitrary branch length transform, not   }
{to be confused with Grafen's rho, implemented in the preceeding       }
{procedure.                                                            }
{**********************************************************************}
Function grafenXFormTree(tree : nodeptr) : nodePtr;
   Var
      me : nodePtr;
      dummy : integer;
   {******************************************************************}
   {This procedure computes the depths recursively.  The original
    Grafenme procedure was split into 2 pieces to support polytomy.
    PEM 9-May-95.                                                     }
   {******************************************************************}
   Procedure grafenMe1(rt,nd: nodePtr; Var totalTips :integer);
      Var
         Rtips, Ltips :integer;
      Begin
         If (nd^.sibL=NIL) Then Begin
            totaltips:=1; {this is a tip, so there is 1 in the subtree}
            nd^.temp:=0 {set depth to 0}
         End {then}
         Else Begin
            grafenMe1(rt,nd^.sibL, Ltips);
            grafenMe1(rt,nd^.sibR, Rtips);
            totalTips:=Ltips+Rtips;
            nd^.temp := totalTips-1 {save number of tips-1 as depth}
         End {else}
      End; {Grafenme1}
   {******************************************************************}
   {This procedure finishes the transformation (taking polytomies into
    account.  It is also recursive. Added PEM 9-May-95.               }
   {******************************************************************}
   Procedure grafenMe2(rt,nd: nodePtr);
      Begin
         If (nd^.sibL <> NIL) Then Begin
            If (nd^.sibL^.length <> 0)    {test for polytomy}
            Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
            Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
            grafenMe2(rt,nd^.sibL);
            If (nd^.sibR^.length <> 0)    {test for polytomy}
            Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
            Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
            grafenme2(rt,nd^.sibR)
         End; {then}
      End; {procedure grafenMe2}
   {******************************************************************}
   Begin
      me := copyTree(tree);
      grafenMe1(me,me,dummy);
      grafenMe2(me,me);
      grafenXFormTree := me;
   End; {function grafenXFormTree}

{**********************************************************************}
{This function tries Nee's arbitrary branch length transform, which is
 similar to Grafen's arbitrary transform, implemented above, except
 that the depth is set to the log10 of the number of tips, rather than
 the "decrement" of the number of tips.}
{**********************************************************************}
Function NeeXFormTree(tree : nodeptr) : nodePtr;
   Var
      me : nodePtr;
      dummy : integer;
   {******************************************************************}
   {This procedure computes the depths recursively. }
   {******************************************************************}
   Procedure neeMe1(rt,nd: nodePtr; Var totalTips :integer);
      Var
         Rtips, Ltips :integer;
      Begin
         If (nd^.sibL=NIL) Then Begin
            totaltips:=1; {this is a tip, so there is 1 in the subtree}
            nd^.temp:=0 {set depth to 0}
         End {then}
         Else Begin
            neeMe1(rt,nd^.sibL, Ltips);
            neeMe1(rt,nd^.sibR, Rtips);
            totalTips:=Ltips+Rtips;
            nd^.temp := ln(totalTips)/ln(10) {save log(number of tips) as depth}
         End {else}
      End; {neeme1}
   {******************************************************************}
   {This procedure finishes the transformation (taking polytomies into
    account.  It is also recursive. Added PEM 9-May-95.               }
   {******************************************************************}
   Procedure neeMe2(rt,nd: nodePtr);
      Begin
         If (nd^.sibL <> NIL) Then Begin
            If (nd^.sibL^.length <> 0)        {test for polytomy}
            Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
            Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
            neeMe2(rt,nd^.sibL);
            If (nd^.sibR^.length <> 0)        {test for polytomy}
            Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
            Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
            neeMe2(rt,nd^.sibR)
         End; {then}
      End; {procedure neeMe2}
   {******************************************************************}
   Begin
      me := copyTree(tree);
      neeMe1(me,me,dummy);
      neeMe2(me,me);
      neeXFormTree := me;
   End; {function neeXFormTree}

{**********************************************************************}
{This function implements Pagel's arbitrary branch length transform.   }
{**********************************************************************}
Function pagelXFormTree(tree : nodePtr) : nodePtr;
   Var
      me : nodePtr;
      dummy : integer;
   {******************************************************************}
   {This procedure computes the depths recursively.  The original
    Pagelme procedure was split into 2 pieces to support polytomy.
    PEM 9-May-95.                                                     }
   {******************************************************************}
   Procedure pagelMe1(rt,nd: nodePtr;Var bifurcations :integer);
      Var
         Rbif, Lbif :integer;
      Begin
         If (nd^.sibL = NIL) Then Begin
            bifurcations:=0;
            nd^.temp := 0 {PEM 9-May-95}
         End {then}
         Else Begin
            pagelme1(rt,nd^.sibL, Lbif);
            pagelme1(rt,nd^.sibR, Rbif);
            If (Lbif > Rbif) Then bifurcations:=Lbif+1
            Else bifurcations:=Rbif +1;
            If ((nd^.length = 0) And  { test for polytomy }
                (nd <> rt))           { avoid unintended root polytomy }
            Then bifurcations := bifurcations - 1; { don't add depth }
                                                   { for polytomies. PEM }
            nd^.temp:=bifurcations;
            End; {else}
         End; {procedure pagelme1}
   {*******************************************************************}
   {This recursive procedure finishes the transformation (taking
    polytomies into account.  Note that it does its work while going
    back down the tree (toward the root), while pagelme2 does its work
    going up the tree (toward the tips).  Added PEM 9-May-95.          }
   {*******************************************************************}
   Procedure pagelMe2(rt,nd: nodePtr);
      Var
         Rbif, Lbif :integer;
         Begin
            If (nd^.sibL <> NIL) Then Begin
               If (nd^.sibL^.length <> 0) {test for polytomy}
               Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
               Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
               pagelMe2(rt,nd^.sibL);
               If (nd^.sibR^.length <> 0) {test for polytomy}
               Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
               Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
               pagelMe2(rt,nd^.sibR)
            End {then}
         End; {procedure pagelme2}
   {******************************************************************}
   Begin
      me := copyTree(tree);
      pagelMe1(me,me,dummy);
      pagelMe2(me,me);
      pagelXFormTree := me;
   End; {function pagelXFormTree}

{**********************************************************************}
{This function sets all branches equal to one.  This is copied from    }
{copyTree, except for the length assignment.                           }
{**********************************************************************}
Function oneXFormTree(tree : nodePtr) : nodePtr;
   Var newTree :nodePtr;
   Begin
      new(newTree);
      initMe(newTree);
      newTree^.name := tree^.name;
      If tree^.length = 0.0
      Then newTree^.length := 0.0  {don't destroy existing polytomies!}
      Else newTree^.length := 1.0;
      If (tree^.sibL<>nil) Then Begin
         newTree^.sibL := oneXFormTree(tree^.sibL);
         newTree^.sibL^.ancestor := newtree
      End; {then}
      If (tree^.sibR<>nil) Then Begin
         newTree^.sibR := oneXFormTree(tree^.sibR);
         newTree^.sibR^.ancestor := newtree
      End; {then}
      oneXFormTree := newTree;
   End; {function oneXFormTree}

{**********************************************************************}
{ This procedure displays a summary screen explaining the columns of   }
{ the file containing the best transforms for each trait, but without  }
{ any statistics.                                                      }
{**********************************************************************}
Procedure displayXFormSummary(Var where :text);
   Begin
      writeln(where,'Summary of transform results file');
      writeln(where,'Column  Description');
      writeln(where);
      writeln(where,'1       Type of best transform for trait 1 (L,E,R,G,N,P,O, or S)');
      writeln(where,'2       Best transform parameter for trait 1');
      writeln(where,'3       Type of best transform for trait 2 (L,E,R,G,N,P,O, or S)');
      writeln(where,'4       Best transform parameter for trait 2');
      writeln(where);
   End; {procedure displayXFormsummary}

{**********************************************************************}
{ This procedure displays a summary screen explaining the columns of   }
{ the statistics file when branch length transforms are not applied.   }
{ As this output does not refer to separate optimizations for each     }
{ trait, it refers to only one tree.                                   }
{**********************************************************************}
Procedure display1TreeSummary(Var where :text);
   Begin
      writeln(where,'Summary of statistics file for tree without transforms applied.');
      writeln(where);
      writeln(where,'Column  Description');
      writeln(where);
      writeln(where,'1       Simulation Number');
      writeln(where,'2       Correlation of Standardized Contrasts (SDC)');
      writeln(where,'3       Ordinary least squares slope of SDC');
      writeln(where,'4       Reduced major axis slope of SDC');
      writeln(where,'5       Major axis slope of SDC');
      writeln(where,'6       Variance of SDC for trait 1');
      writeln(where,'7       Variance of SDC for trait 2');
      writeln(where,'8       F statistic');
      writeln(where,'9       # of Y SDC >0');
      writeln(where,'10      # of Y SDC <0');
      writeln(where,'11      # of Y SDC =0');
      writeln(where,'12      Corr of ABS of residuals of SDC Y on SDC X vs. corrected base node ht.');
      writeln(where,'13,14   Corr of ABS of contrasts of (trait 1,2) vs SQRT sum of branch lengths');
      writeln(where,'15,16   Corr of ABS of SDC (trait 1,2) vs. node values at base of contrasts');
      writeln(where,'17,18   Corr of ABS of SDC (trait 1,2) vs. corrected height of base nodes');
      writeln(where,'19,20   Corr of est. nodal values (trait 1,2) vs. corrected node ht. (w/o tips)');
      writeln(where,'21,22   Root values (trait 1,2) after any tip operations');
      writeln(where,'23,24   Simple means of tips (trait 1,2) after any tip operations');
      writeln(where);
   End; {procedure display1treesummary}

{**********************************************************************}
{ This procedure displays a summary screen explaining the columns of   }
{ the statistics file when branch length transforms are applied.       }
{ "2tree" refers to the fact that the statistics are based on the trees}
{ with optimum branch length for each trait.                           }
{**********************************************************************}
Procedure display2TreeSummary(Var where :text);
   Begin
      writeln(where,'Summary of statistics file for transformed tree(s).');
      writeln(where);
      writeln(where,'Column  Description');
      writeln(where);
      writeln(where,'1       Simulation Number');
      writeln(where,'2       Correlation of Standardized Contrasts (SDC)');
      writeln(where,'3       Ordinary least squares slope of SDC');
      writeln(where,'4       Reduced major axis slope of SDC');
      writeln(where,'5       Major axis slope of SDC');
      writeln(where,'6,7     Variance of SDC for traits 1,2');
      writeln(where,'8       F statistic');
      writeln(where,'9       # of Y SDC >0');
      writeln(where,'10      # of Y SDC <0');
      writeln(where,'11      # of Y SDC =0');
      writeln(where,'12      Corr of ABS of residuals of SDC Y on SDC X vs. corrected base node ht.');
      writeln(where,'13,14   Corr of ABS of contrasts (trait 1,2) vs. SQRT sum of branch lengths');
      writeln(where,'15,16   Corr of ABS of SDC (trait 1,2) vs. node values at base of contrasts');
      writeln(where,'17,18   Corr of ABS of SDC (trait 1,2) vs. corrected height of base node');
      writeln(where,'19,20   Corr of est. nodal values (trait 1,2) vs. corrected node ht (w/o tips)');
      writeln(where,'21,22   Best transform type (L,E,R,G,N,P,O, or S) and value: trait 1');
      writeln(where,'23,24   Best transform type (L,E,R,G,N,P,O, or S) and value: trait 2');
      writeln(where,'25,26   Root values (trait 1,2) after any tip operations and best transform');
      writeln(where,'27,28   Simple means of tips (trait 1,2) after any tip operations');
   End; {procedure display2treesummary}

{**********************************************************************}
{This procedure shows a page of notes pertaining to the available
 transforms.  }
{**********************************************************************}
Procedure displayXFormNotes;
   Var userChar : char;
   Begin
      clrscr;
      writeln('PDERROR will search for the transform that minimizes the absolute value of');
      writeln('a correlation you will select in a subsequent screen.  PDERROR will try');
      writeln('one or more of the following transforms (that you will select in the next');
      writeln('screen).  If two transforms produce the same correlation, PDERROR will always');
      writeln('select the first one.  PDERROR tries transforms in the following order:');
      writeln;
      writeln('Self: this "transform" leaves the tree unaltered.');
      writeln('Grafen: Grafen''s (1989) arbitrary branch lengths');
      writeln('Nee: Nee''s arbitrary branch lengths (see Purvis, 1995, p. 416)');
      writeln('Pagel: Pagel''s (1992) arbitrary branch lengths');
      writeln('One: all non-zero length branches are set to a length of one.');
      writeln('Log: branches of length greater than one are set to their base 10 log.');
      writeln('     Note: This transform will not be tried if the tree contains branches');
      writeln('     of length less than one in the tree.  This avoids non-positive branches.');
      writeln('Exponential: branch lengths are raised to a power.  You will be prompted for');
      writeln('     a range of values to try, which are stepped through with a fixed upwards');
      writeln('     increment of 0.1.  Note that an exponent of 0 is equivalent to the');
      writeln('     "One" transform.  Round-off errors may cause PDERROR to report either');
      writeln('     the One transform or the exponent of 0, if both are tried.  Likewise');
      writeln('     an exponent of 1 is equivalent to the "Self" transformation.');
      writeln('Rho: Grafen''s (1992) rho transform.  Like the exponential transform, you');
      writeln('     will be prompted for a range of values which will be stepped through in');
      writeln('     increments of 0.1.  A rho of 1 is equivalent to the "Self" transformation.');
      writeln;
      write('Press any key to continue');
      userChar := readKey;
   End; {procedure displayXFormNotes}

{**********************************************************************}
{This procedure draws a screen display for choosing among the 8(!)
 available transform options.  The display includes the "square root
 checkmark" for selected options.  The following procedure handles the
 user's response.}
{**********************************************************************}
Procedure drawXFormMenu;
   Var i : integer;
   Begin
      With xForms^ Do Begin
         clrscr;
         gotoxy(17,1);
         write('Transformation Choice Menu');
         gotoxy(1,3);
         write('Enter one or more letters (spaces optional) to select or unselect');
         gotoxy(1,4);
         write('   transform choice.  Enter blank line when done.');
         gotoxy(1,6);
         For i:= 1 To 79 Do write(chr(196));
         gotoxy(1,7);
         If doSelf
         Then write(chr(251))
         Else clreol;
         gotoxy(3,7);
         write('(S)elf, include original branch lengths from input as option;');
         gotoxy(1,9);
         If doGrafen
         Then write(chr(251))
         Else clreol;
         gotoxy(3,9);
         write('(G)rafen''s (1989) arbitrary branch lengths');
         gotoxy(1,10);
         If doNee
         Then write(char(251))
         Else clreol;
         gotoxy(3,10);
         write('(N)ee''s arbitrary branch lengths (see Purvis, 1995, p. 416)');
         gotoxy(1,11);
         If doPagel
         Then write(chr(251))
         Else clreol;
         gotoxy(3,11);
         write('(P)agel''s (1992) arbitrary branch lengths');
         gotoxy(1,12);
         If doOne
         Then write(chr(251))
         Else clreol;
         gotoxy(3,12);
         write('(O)ne, all branches set to unity (polytomies are preserved)');
         gotoxy(1,14);
         If doLog
         Then write(chr(251))
         Else clreol;
         gotoxy(3,14);
         write('(L)og10 transformation');
         gotoxy(1,15);
         If doExps
         Then write(chr(251))
         Else clreol;
         gotoxy(3,15);
         write('(E)xponential transformation (user-specified range, step size=0.1)');
         gotoxy(1,16);
         If doRho
         Then write(chr(251))
         Else clreol;
         gotoxy(3,16);
         write('(R)ho (see Grafen, 1989) transformation (user-specified range, step size=0.1)');
         gotoxy(1,18);
         For i:= 1 To 79 Do write(chr(196));
         gotoxy(1,20);
         write('Enter your choice(s) {E,L,R,G,N,P,O,S}: ');
      End; {with}
   End; {procedure drawXFormMenu}

{**********************************************************************}
{This procedure reads the user's choice from the menu in the previous
 procedure.  Note that if the user selects nothing from this list, the
 self (no change) option is automatically selected. }
{**********************************************************************}
Procedure processXFormChoices(var done : boolean);
   Var
      i : integer;
      userChar : char;
      userStr : string;
   Begin
      With xForms^ Do Begin
         readln(userStr);
         If ((length(userStr) = 0) And
             (doExps Or doLog Or doRho Or doGrafen Or doNee Or
                   doPagel Or doOne))
         Then done := true
         Else Begin
            done := false;
            For i := 1 to length(userStr) Do Begin
               Case userStr[i] of
                  'E','e' : doExps := not(doExps);
                  'L','l' : doLog := not(doLog);
                  'R','r' : doRho := not(doRho);
                  'G','g' : doGrafen := not(doGrafen);
                  'N','n' : doNee := not(doNee);
                  'P','p' : doPagel := not(doPagel);
                  'O','o' : doOne := not(doOne);
                  'S','s' : doSelf := not(doSelf);
               End {case}
            End; {for i}
            {sanity check...}
            If not(doExps Or doLog Or doRho Or doGrafen Or doNee Or
                   doPagel Or doOne)
            Then Begin
               gotoxy(1,23);
               If doSelf Then Begin 
                  write('You need to specify a transformation in additon to (S)elf');
                  gotoxy(1,7);
                  write(chr(251))
               End
               Else Begin
                  write('You need to specify a transformation');
                  gotoxy(1,7);
                  write(chr(32))
               End; {else}
               gotoxy(1,24);
               write('Press a key to return to the menu');
               userChar := readKey;
               gotoxy(1,23);
               clreol;
               gotoxy(1,24);
               clreol;
            End; {then}
         End {else}
      End {with}
   End; {procedure processXFormChoices}

{**********************************************************************}
{This procedure gets user choices for which transforms to try (by
 calling the previous two procedures) and, if the user selects either
 the exponential or log transforms, prompts for ranges of these values.}
{**********************************************************************}
Procedure getXFormOptions(var xForms : xFormPtr);
   Var
      userInt : integer;
      done : boolean;
      ok : boolean;
   Begin
      new(xForms);
      With xForms^ do Begin
         doExps := false;    {Initialize everything else to false}
         doLog := false;
         doRho := false;
         doGrafen := false;
         doNee := false;
         doPagel := false;
         doOne := false;
         doSelf := true;     {except self}
         done := false;
         displayXFormNotes;
         Repeat
            drawXFormMenu;
            processXFormChoices(done)
         Until done;
         writeln;
         If (doTweak And
             doLog And
             ((theTweak^.myType = NORMAL) And
              (theTweak^.zeroLength <= 1.0)) Or
              ((theTweak^.myType = FIXED) And
               (theTweak^.deviation > 1) And
               (theTweak^.zeroLength <= 1.0))) Then Begin
            writeln;
            writeln('*****   Selected added errors may yield branches of non-        *****');
            writeln('*****   positive length with the log10 transform.  Therefore:   *****');
            writeln('*****                                                           *****');
            writeln('*****   Log10 transform will not be tested.                     *****');
            writeln;
            writeln('Press any key to continue.');
            userChar := readkey;
            doLog := false;
         End; {then}

         {Now ask the user what "best" means, in other words, what's the}
         {correlation to minimize?}
         clrscr;
         writeln('PDERROR will search for the transform which minimizes a user-specified');
         writeln('correlation independently for each trait.');
         writeln('Which correlation would you like to minimize?');
         writeln;
         writeln('(B)ranch  minimizes the correlation between the absolute value of');
         writeln('          standardized contrasts and their standard deviations.');
         writeln('          These are the correlations on PDTREE diagnostic screens 1 and 2.');
         writeln('          This diagnostic was suggested by Garland et al. (1992) and');
         writeln('          studied by Diaz-Uriarte and Garland (1996, 1998) and');
         writeln('          Garland and Diaz-Uriarte (1999).');
         writeln;
         writeln('(N)ode    minimizes the correlation between the absolute value of');
         writeln('          standardized contrasts and the estimated value of the base node.');
         writeln('          These are the correlations on PDTREE diagnostic screens 3 and 4.');
         writeln;
         writeln('(H)eight  minimizes the correlation between the absolute value of');
         writeln('          standardized contrasts and the corrected height of the base node.');
         writeln('          These are the correlations on PDTREE diagnostic screens 5 and 6.');
         writeln;
         writeln('(V)alue   minimizes the correlation between the estimated value of the');
         writeln('          base node and its corrected height.');
         writeln('          These are the correlations on PDTREE diagnostic screens 7 and 8.');
         writeln;
         Repeat
            write('Which correlation shall be minimized? (B=Enter,N,H,V): ');
            readln(userChar);
            If userChar = chr(13) Then userChar := 'B'
            Else userChar := upCase(userChar);
            writeln;
         Until (userChar in ['B','N','H','V']);
         Case userChar of
             'B' : xForms^.cost := SdcBL;  {ABS(SDC) vs sqrt(BL1+BL2); #1&2}
             'N' : xForms^.cost := SdcNV;  {ABS(SDC) vs est Nodal value; #3&4}
             'H' : xForms^.cost := SdcHt;  {ABS(SDC) vs base node height; #5&6}
             'V' : xForms^.cost := NvHt;   {Node value vs base node height; #7&8}
         End; {case userChar}
         writeln;

         write('Do you want to calculate contrast statistics using the best');
         write(' transform you find? (N/Y=Enter): ');
         readln(userChar);
         writeln;
         If (upcase(userChar)='N')
         Then doBLStats := false
         Else doBLStats := true;

         {Select parameters for Exponential and Rho transforms}

         If doExps Then Begin
            writeln('**************************************************************************');
            writeln('PDERROR will search a range of exponents for a value that minimizes');
            writeln('the correlation you selected above.  The program uses a FIXED step-size');
            writeln('of 0.1.');
            writeln;
            writeln('Select a range to search for the best exponential parameter:');
            writeln;
            writeln('(1) 0.0 to 1.0');
            writeln('(2) 1.0 to 2.0');
            writeln('(3) 0.0 to 2.0');
            writeln('(4) user-specified range');
            writeln;
            Repeat
               writeln('Enter 1, 2, 3 or 4: ');
               readln(userInt);
            Until ((userInt>0) and (userInt<5));
            writeln;
            Case userInt Of
                 1: Begin
                       expStart := 0.0;
                       expEnd := 1.0
                    End; {case 1}
                 2: Begin
                       expStart := 1.0;
                       expEnd := 2.0
                    End; {case 2}
                 3: Begin
                       expStart := 0.0;
                       expEnd := 2.0
                    End; {case 3}
                 4: Begin
                       write('Enter the lower value for the exponential parameter: ');
                       readln(expStart);
                       writeln;
                       write('Enter the upper value for the exponential parameter: ');
                       readln(expEnd);
                       writeln;
                       If (expStart > expEnd)
                       Then Begin
                          writeln('Swapping endpoints...');
                          writeln;
                          tmpFloat := expStart;
                          expStart := expEnd;
                          expEnd := tmpFloat;
                       End; {then}
                    End; {case 4}
            End; {case}
         End; {then}

         If doRho Then Begin
            writeln('**************************************************************************');
            writeln('PDERROR will search a range of values for one that specifies a Rho');
            writeln('parameter (Grafen, 1989) that minimizes the correlation you selected');
            writeln('above.  The program uses a FIXED step-size of 0.1.');
            writeln;
            writeln('Select a range to search for the best Grafen''s (1989) Rho parameter:');
            writeln;
            writeln('(1) 0.1 to 1.0');
            writeln('(2) 1.0 to 5.0');
            writeln('(3) 0.1 to 5.0');
            writeln('(4) user-specified range (lower limit can not be less than 0.05)');
            writeln;
            Repeat
               writeln('Enter 1, 2, 3, or 4: ');
               readln(userInt);
            Until ((userInt>0) and (userInt<5));
            writeln;
            Case userInt Of
                 1: Begin
                       rhoStart := 0.1;
                       rhoEnd := 1.0
                    End; {case 1}
                 2: Begin
                       rhoStart := 1.0;
                       rhoEnd := 5.0
                    End; {case 2}
                 3: Begin
                       rhoStart := 0.1;
                       rhoEnd := 5.0
                    End; {case 3}
                 4: Begin
                       writeln;
                       writeln('Note: values of rho < 0.05 are not allowed, to avoid problems with');
                       writeln('      trees with noncontemporaneous tips.');
                       ok := false;
                       Repeat
                          write('Enter the lower value for the rho parameter: ');
                          readln(rhoStart);
                          writeln;
                          write('Enter the upper value for the rho parameter: ');
                          readln(rhoEnd);
                          writeln;
                          If (rhoStart > rhoEnd)
                          Then Begin
                             writeln('Swapping endpoints...');
                             writeln;
                             tmpFloat := rhoStart;
                             rhoStart := rhoEnd;
                             rhoEnd := tmpFloat;
                          End; {then}
                          If (rhoStart < 0.05) Then Begin
                             writeln;
                             writeln('The lower bound is too small, it must be at least 0.05.');
                             writeln;
                          End
                          Else ok := true;
                       Until ok;
                    End; {case 4}
            End; {case}
         End; {then}
      End; {with}
  End; {procedure getXFormOptions}


  {*************************************************************************}
  {This procedure reports a message that tweaking has created a tree with   }
  {adjacent zero length tip branches.  After reporting the error and        }
  {suggesting some fixes, it will cause PDERROR to exit to DOS.             }
  {*************************************************************************}
  Procedure tweakZLTipError(tweak : tweakPtr);
     Begin
        clrscr;
        writeln('Adding random branch length errors produced a tree with ');
        writeln('adjacent zero-length tip branches.  Contrasts cannot be ');
        writeln('calculated; therefore, the program will exit to DOS.');
        writeln;
        If (((tweak^.myType = NORMAL) Or (tweak^.deviation > 1)) And
             (tweak^.zeroLength <= 1 )) Then Begin
           writeln('You may want to either change the value you assign to');
           writeln('branches that are modified to a non-positive value.');
           writeln('(You chose to set these branches to ',tweak^.zeroLength,')');
           If (tweak^.myType = NORMAL)
           Then Begin
              writeln('Or, you might change the %CV for adding normal values.');
              writeln('You chose ',tweak^.width,'.');
           End
           Else Begin
              writeln('Or, you might change your fixed deviation to something');
              writeln('less than one.');
              End; {else}
        End; {then}
        halt;
     End; {procedure tweakZLTipError}

  {*************************************************************************}
  {This procedure reports a message that a transform has created a tree with}
  {adjacent zero length tip branches.  After reporting the error and        }
  {suggesting some fixes, it will cause PDERROR to exit to DOS.             }
  {*************************************************************************}
  Procedure xFormZLTipError(xForm:string);
     Begin
        clrscr;
        writeln('Applying ' + xForm + ' produced a tree with ');
        writeln('adjacent zero-length tip branches.  Contrasts cannot be ');
        writeln('calculated; therefore, the program will exit to DOS.');
        writeln;
        writeln('If this transform continues to cause errors, then you might');
        write('consider not including it');
        If ((copy(xForm,1,15) = 'an exponent of ') Or
            (copy(xForm,1,23) = 'Grafen''s (1992) Rho of '))
        Then Begin
           writeln(', or narrowing the range of values explored to be');
           writeln('closer to 1.0.')
        End {then}
        Else writeln('.');
        halt;
     End; {procedure xFormZLTipError}

{**********************************************************************}
   Begin { main program }
      clrscr;
      writeln('********* Phenotypic Diversity Analysis Program - PDERROR Module *********');
      writeln;
      writeln('V. 2.0 Copyright 27 Dec 1999 by Peter E. Midford and Theodore Garland, Jr.');
      writeln;
      writeln('Direct all questions to T.G. at Department of Zoology, 430 Lincoln Drive,');
      writeln('University of Wisconsin, Madison, WI 53706 U.S.A.');
      writeln('Phone 608-262-4437; Facsimile 608-265-6320; Email tgarland@facstaff.wisc.edu');
      writeln;
      writeln('This program is used to analyze simulated data (e.g., from PDSIMUL of');
      writeln('Garland et al., 1993, Syst. Biol. 42:265-292) by computing a bivariate');
      writeln('correlation with Felsenstein''s (1985) independent contrasts.  It allows');
      writeln('diagnostic tests and transformations of branch lengths.  It has been used to');
      writeln('study effects of deviations from Brownian motion, errors in branch lengths,');
      writeln('and polytomies (Diaz-Uriarte and Garland, 1996, Syst. Biol. 45:27-47, 1998,');
      writeln('Syst. Biol. 47:654-672; Garland and Diaz-Uriarte, 1999, Syst. Biol. 48:547-558).');
      writeln('PDERROR can manipulate branch lengths in two ways:');
      writeln(' (1) adding random errors to branch lengths (fixed or relative normal)');
      writeln(' (2) transforming using log10/exponential or Grafen''s (1989) rho, or arbitrary');
      writeln('     branch lengths (unity; Pagel, 1992; Grafen, 1989; Nee, see Purvis 1995)');
      writeln('The program can also categorize or rank the tip data prior to all analyses.');
      writeln;

      succ := false;
      Repeat
         openSimFile(inputFlName,simFile,succ);
         readTree(inputTree,tipCount,SimFile); {read the tree early}
         zeroLenTip := false;
         checkZeroTips(inputTree,zeroLenTip);
         If zeroLenTip Then Begin
            writeln('Your tree has two adjacent tips with branch lengths of zero.');
            writeln('Contrasts can not be calculated with such a tree, so PDERROR will');
            writeln('not accept it.');
            writeln;
            writeln('You can either exit PDERROR, fix the problem with PDTREE and');
            writeln('recreate your sim file, or you can try analyzing your data on a');
            writeln('different tree.');
            writeln;
            write('Would you like to analyze the simulated data on a different tree (Y/N=Enter)? ');
            readln(userChar);
            writeln;
            If (upcase(userChar) = 'Y')
            Then Begin
               treeSubst(analysisTree,secondFileName,inputTree,tipCount);
            End; {then}
            If ((upcase(userChar) <> 'Y') Or
                (analysisTree = inputTree)) Then Begin
               writeln('Exiting PDERROR...');
               halt;
            End; {then}
         End; {then}
      Until succ;
      rootName := copy(inputFlName,1,pos('.',inputFlName)-1);

      writeln('You can use a different tree to analyze data.  The tree is read from a');
      writeln('file that must be in .INP format.  If the tree is in a .PDI file, use');
      writeln('PDTREE to convert it to .INP.');
      writeln;

      write('Would you like to analyze the simulated data on a different tree (Y/N=Enter)? ');
      readln(userChar);
      writeln;
      If (upcase(userChar) = 'Y')
      Then treeSubst(analysisTree,secondFileName,inputTree,tipCount)
      Else Begin
         analysisTree := inputTree;
         secondFileName := ''
      End; {else}

      {Add data values to interior nodes only if using the tree from the
       simulation file.  Other we're likely to get a name problem. }
      tipsOnly := (inputTree <> analysisTree); 

      write('Do you want to set a limit on the # of simulations to analyze? (Y/N=Enter): ');
      readln(userChar);
      writeln;
      If (upcase(userChar)='Y') Then Begin
         write('Enter the maximum number of simulations to analyze: ');
         readln(userLimit);
         writeln
      End {then}
      Else userLimit := maxInt;

      binCount[1] := 0;
      binCount[2] := 0;
      clrscr;
      writeln('**************************************************************************');
      writeln('Do you want to process tip data by categorizing one or both traits?');
      writeln;
      writeln('Catagorizing places simulated tip values into equal-width bins starting');
      writeln('at the minimum value and ending at the maximum value for that trait');
      writeln('among all simulated data sets.  Tip values will be replaced by their');
      writeln('their bin numbers.');
      Repeat
         write('Please enter (N=Enter, 1=trait 1, 2=trait 2, 3 = both traits): ');
         readln(userChar);
         If (userChar = chr(13)) Then userChar := 'N';
         writeln
      Until (userChar in ['1','2','3','N','n']);
      writeln;

      Case userChar of
         '1' : doCategories := [1];
         '2' : doCategories := [2];
         '3' : doCategories := [1,2];
         'N',
         'n' : doCategories := [];
      End; {case userChar}

      If (doCategories <> []) Then Begin
         writeln('This option categorizes simulated tip values into equal-width bins');
         writeln('starting at the minimum value and ending at the maximum value.');
         writeln('The tip values are then replaced by their bin numbers.');
         writeln;
         restoreCrtAssignments;
         clrscr;
         gotoxy(1,1);
         textColor(LightGray+Blink);
         write('Scanning ',inputFLName,' for minimum and maximum values');
         scanForLimits(inputFLName,globalMin,globalMax,tipCount,userLimit);
         textColor(LightGray);
         clrscr;
         resetCrtAssignments;
         If 1 in doCategories Then Begin
         End; {then}
         If 2 in doCategories Then Begin
         End; {then}

         writeln;
         If 1 in doCategories Then Begin
            write('The maximum value of trait 1 is ');
            extendWrite(output,globalMax[1]);
            writeln;
            write('The minimum value of trait 1 is ');
            extendWrite(output,globalMin[1]);
            writeln;
            writeln;
            writeln('Do you want to specify limits for trait 1 (M)anually or');
            write('(A)utomatically (M/A=Enter)? ');
            readln(userChar);
            If (userChar in ['M','m']) Then Begin
               write('Enter the lower bound value for trait 1: ');
               readln(globalMin[1]);
               writeln;
               write('Enter the upper bound value for trait 1: ');
               readln(globalMax[1]);
               writeln;
            End
            Else Begin
               writeln('Bins for trait 1 will be set from the minimum and maximum');
               writeln('values reported above.');
               writeln;
            End; {else}
            write('How many bins do you want? (0 cancels option): ');
            readln(binCount[1]);
            writeln;
            If (binCount[1] > 0) Then Begin
               writeln('Category numbers can be assigned starting from 0 or from 1.');
               writeln;
               writeln('Would you like to number the bins (categories) from 0 to ',binCount[1]-1:2);
               write('or 1 to ',binCount[1]:2,'? (0/1 = enter): ');
               readln(userStr);
               writeln;
               If (pos('0',userStr) > 0)
               Then binBase[1] := 0
               Else binBase[1] := 1;
            End; {then}
         End; {then}

         writeln;
         If 2 in doCategories Then Begin
            clrscr;
            write('The maximum value of trait 2 is ');
            extendWrite(output,globalMax[2]);
            writeln;
            write('The minimum value of trait 2 is ');
            extendWrite(output,globalMin[2]);
            writeln;
            writeln;
            writeln('Do you want to specify limits for trait 2 (M)anually or');
            write('(A)utomatically (M/A=Enter)? ');
            readln(userChar);
            If (userChar in ['M','m']) Then Begin
               write('Enter the lower bound value for trait 2: ');
               readln(globalMin[2]);
               writeln;
               write('Enter the upper bound value for trait 2: ');
               readln(globalMax[2]);
               writeln;
            End
            Else Begin
               writeln('Bins for trait 2 will be set from the minimum and maximum');
               writeln('values reported above.');
               writeln;
            End; {else}
            write('How many bins do you want? (0 cancels option): ');
            readln(binCount[2]);
            writeln;
            If (binCount[2] > 0) Then Begin
               writeln('Category numbers can be assigned starting from 0 or from 1.');
               writeln;
               writeln('Would you like to number the bins (categories) from 0 to ',binCount[2]-1:2);
               write('or 1 to ',binCount[2]:2,'? (0/1 = enter): ');
               readln(userStr);
               writeln;
               If (pos('0',userStr) > 0)
               Then binBase[2] := 0
               Else binBase[2] := 1;
               showCategories := false;
            End; {then}
         End; { then begin }
      End; {then begin}

      clrscr;
      writeln('**************************************************************************');
      writeln('Do you want to process tip data by ranking one or both traits?');
      writeln;
      writeln('Ranking places tip values in ascending order, with ties assigned');
      writeln('an average value so that the sum of the ranks is always the same;');
      writeln('Tip values will be replaced by their ranks.');
      If (doCategories <> []) Then Begin
         writeln;
         If 1 in doCategories
         Then writeln('Note: you have already categorized trait 1.');
         If 2 in doCategories
         Then writeln('Note: you have already categorized trait 2.');
         writeln
      End; {then}
      Repeat
         write('Please enter (N=Enter, 1=trait 1, 2=trait 2, 3 = both traits): ');
         readln(userChar);
         If (userChar = chr(13)) Then userChar := 'N';
         writeln
      Until (userChar in ['1','2','3','N','n']);
      writeln;

      Case userChar of
         '1' : doRanks := [1];
         '2' : doRanks := [2];
         '3' : doRanks := [1,2];
         'N',
         'n' : doRanks := [];
      End; {case userChar}


      doTweak := false;
      theTweak := nil;
      tweaksToSave := 0;
      writeln('**************************************************************************');
      write('Do you want to add random branch errors? (Y/N=Enter): ');
      readln(userChar);
      writeln;
      If (upCase(userChar) = 'Y') Then Begin
         getTweak(theTweak,analysisTree);
         doTweak := true;
         write('Do you want to save any trees with branch errors to a file? (Y/N=Enter): ');
         readln(userChar);
         writeln;
         If (upCase(userChar) = 'Y') Then Begin
            write('How many trees with branch errors do you want to save?: ');
            readln(tweaksToSave);
            If (tweaksToSave > userLimit) Then Begin
               tweaksToSave := userLimit;
               writeln('You requested more saved trees than simulation sets to process.');
            End
            Else writeln('Depending on the number of simulations in the data file, up to');
            writeln(tweaksToSave,' trees with branch errors will be saved.');
            writeln;
            write('Trees with errors will be saved to: ');
            tweakFileName := rootName + '.TKS';
            writeln(tweakFileName);
         End;
         writeln;
         write('Do you want to initialize the random number generator (Y/N=Enter): ');
         readln(userChar);
         writeln;
         userChar := upCase(userChar);
         If (userChar = 'Y') Then Begin
            write('Enter the (integer) seed you wish to use: ');
            readln(savedSeed);
            writeln;
            writeln;
            randomize;
            randSeed := savedSeed
         End {then}
         Else Begin
            randomize;
            savedSeed := randSeed;
            writeln('New seed is ',savedSeed);
            writeln;
         End {else}
      End; {then}

      writeln('**************************************************************************');
      write('Do you want to explore transforms for branch lengths (Y/N=Enter)? ');
      readln(userChar);
      writeln;
      If upcase(userChar) = 'Y'
      Then getXFormOptions(xForms)
      Else xForms := NIL;


      {Determine what sort of output file to create.}
      openResultFile(rootName,resultFile,resultFileName,'TXB');

      {next, open the parameters file to generate its name.}
      openParametersFile(rootName,paramFile,paramFileName);

      {Now write the parameters file}
      writeParameters(paramFile,
                      resultFileName,
                      inputFlName,
                      secondFileName,
                      tweaksToSave,
                      tweakFileName,
                      binCount,
                      binBase,
                      doRanks,
                      theTweak,
                      xForms,
                      savedSeed);
      close(paramFile);          {make sure everything is saved to this}
                                 {point in case of a crash.}

      If (tweaksToSave > 0) Then
         openTweaksFile(tweaksFile,tweakFileName);

      {Finally, open the simulation file and do some work.}
      skipTree(SimFile);

      readDataSet(simData,tipCount,simFile);
      If (simData <> nil) Then Begin
         If (doCategories <> [])
         Then categorizeData(simData,
                             analysisTree,
                             binCount,
                             binBase,
                             globalMin,
                             globalMax)
         Else If (doRanks <> []) Then Begin
                 rankData(simData,tipCount,doRanks);
                 assignRanks(simData,doRanks);
              End; {then}
         computeTipStats(simData,tipStat);
      End; {then}

      simNumber := 0;
      blLTE1 := false;
      Repeat
         inc(simNumber);
         writeln('Processing Set ',simNumber);
         If doTweak Then Begin
            tweakedTree := tweakTree(analysisTree,theTweak);
            If (simNumber <= tweaksToSave) Then Begin
               addData(tweakedTree,simData,tipsOnly);   {add some simulated data to tweaked tree file}
               writeTweakedTree(tweakedTree,tweaksFile)
            End
         End {then}
         Else tweakedTree := analysisTree;

         blLTE1 := branchesLTE1(tweakedTree,tweakedTree); {always check for short branches!}

         zeroLenTip := false;
         checkZeroTips(tweakedTree,zeroLenTip);
         If zeroLenTip Then tweakZLTipError(theTweak);

         If (xForms = NIL) Then Begin
            addData(tweakedTree,simData,tipsOnly);
            computeStatistics(tweakedTree,simStats,correlations);
            write1TreeStats(resultFile,
                            simStats,
                            simNumber,
                            correlations,
                            tweakedTree,
                            tipStat);
         End {then}
         Else Begin  {explore transforms}
            bestCorr1 := -9.0; {Less than any possible correlation}
            bestCorr2 := -9.0; {Likewise...}

            If xForms^.doSelf Then Begin
               curTree := copyTree(tweakedTree);
               addData(curTree,simData,tipsOnly);
               correlationScore(curTree,xForms^.cost,corrT1,corrT2);
               If (abs(corrT1)<abs(bestCorr1)) Then Begin
                  t1Param := -9.0;
                  t1Display := -9.0;
                  t1type := 'S';
                  bestCorr1 := corrT1
               End; {then}
               If (abs(corrT2)<abs(bestCorr2)) Then Begin
                  t2Param := -9.0;
                  t2Display := -9.0;
                  t2type := 'S';
                  bestCorr2 := corrT2
               End; {then}
               killTree(curTree);
            End; {then}

            If xForms^.doGrafen Then Begin
               curTree := grafenXFormTree(tweakedTree);
               zeroLenTip := false;
               checkZeroTips(curTree,zeroLenTip);
               If zeroLenTip Then xFormZLTipError('Grafen''s arbitrary');
               addData(curTree,simData,tipsOnly);
               correlationScore(curTree,xForms^.cost,corrT1,corrT2);
               If (abs(corrT1)<abs(bestCorr1)) Then Begin
                  t1Param := -9.0;
                  t1Display := -9.0;
                  t1type := 'G';
                  bestCorr1 := corrT1
               End; {then}
               If (abs(corrT2)<abs(bestCorr2)) Then Begin
                  t2Param := -9.0;
                  t2Display := -9.0;
                  t2type := 'G';
                  bestCorr2 := corrT2
               End; {then}
               killTree(curTree);
            End; {then}

            If xForms^.doNee Then Begin
               curTree := neeXFormTree(tweakedTree);
               zeroLenTip := false;
               checkZeroTips(curTree,zeroLenTip);
               If zeroLenTip Then xFormZLTipError('Nee''s arbitrary');
               addData(curTree,simData,tipsOnly);
               correlationScore(curTree,xForms^.cost,corrT1,corrT2);
               If (abs(corrT1)<abs(bestCorr1)) Then Begin
                  t1Param := -9.0;
                  t1Display := -9.0;
                  t1type := 'N';
                  bestCorr1 := corrT1
               End; {then}
               If (abs(corrT2)<abs(bestCorr2)) Then Begin
                  t2Param := -9.0;
                  t2Display := -9.0;
                  t2type := 'N';
                  bestCorr2 := corrT2
               End; {then}
               killTree(curTree);
            End; {then}

            If xForms^.doPagel Then Begin
               curTree := pagelXFormTree(tweakedTree);
               zeroLenTip := false;
               checkZeroTips(curTree,zeroLenTip);
               If zeroLenTip Then xFormZLTipError('Pagel''s arbitrary');
               addData(curTree,simData,tipsOnly);
               correlationScore(curTree,xForms^.cost,corrT1,corrT2);
               If (abs(corrT1)<abs(bestCorr1)) Then Begin
                  t1Param := -9.0;
                  t1Display := -9.0;
                  t1type := 'P';
                  bestCorr1 := corrT1
               End; {then}
               If (abs(corrT2)<abs(bestCorr2)) Then Begin
                  t2Param := -9.0;
                  t2Display := -9.0;
                  t2type := 'P';
                  bestCorr2 := corrT2
               End; {then}
               killTree(curTree);
            End; {then}

            If xForms^.doOne Then Begin
               curTree := oneXFormTree(tweakedTree);
               addData(curTree,simData,tipsOnly);
               correlationScore(curTree,xForms^.cost,corrT1,corrT2);
               If (abs(corrT1)<abs(bestCorr1)) Then Begin
                  t1Param := -9.0;
                  t1Display := -9.0;
                  t1type := 'O';
                  bestCorr1 := corrT1
               End; {then}
               If (abs(corrT2)<abs(bestCorr2)) Then Begin
                  t2Param := -9.0;
                  t2Display := -9.0;
                  t2type := 'O';
                  bestCorr2 := corrT2
               End; {then}
               killTree(curTree);
            End; {then}

            If xForms^.doLog Then  Begin
               If blLTE1 Then Begin
                  If doTweak Then Begin
                     writeln('After adding errors, tree for simulation set ', simNumber,' had branch lengths < 1.0');
                     writeln(' skipping log10 transform');
                  End
                  Else Begin
                     writeln('The tree for simulation set ', simNumber,' had branch lengths < 1.0');
                     writeln(' skipping log10 transform');
                  End; {else}
               End {then}
               Else Begin
                  curTree := logXFormTree(tweakedTree);
                  zeroLenTip := false;
                  checkZeroTips(curTree,zeroLenTip);
                  If zeroLenTip Then xFormZLTipError('a Log10');
                  addData(curTree,simData,tipsOnly);
                  correlationScore(curTree,xForms^.cost,corrT1,corrT2);
                  t1Param := -0.1;     {provisional assignments}
                  bestCorr1 := corrT1;
                  t1Type := 'L';
                  t2Param := -0.1;
                  bestCorr2 := corrT2;
                  t2Type := 'L';
                  killTree(curTree);
               End {else}
            End; {then}

            If xForms^.doExps Then With xForms^ Do Begin
               treeHeight := findHeight(tweakedTree);
               curXForm := expStart;
               Repeat
                  If curXForm = 1 Then
                    curTree := copyTree(tweakedTree)
                  Else
                    curTree := xformTree(tweakedTree,curXForm);
                  zeroLenTip := false;
                  checkZeroTips(curTree,zeroLenTip);
                  If zeroLenTip
                  Then Begin
                     str(curXForm,tmpStr);
                     xFormZLTipError('an exponent of '+tmpStr);
                  End; {then}
                  addData(curTree,simData,tipsOnly);
                  correlationScore(curTree,xForms^.cost,corrT1,corrT2);
                  If abs(bestCorr1)>abs(corrT1) Then Begin
                     t1Type := 'E';
                     t1Param := curXForm;
                     t1display := t1Param;
                     bestCorr1 := corrT1
                  End; {then}
                  If abs(bestCorr2)>abs(corrT2) Then Begin
                     t2Type := 'E';
                     t2Param := curXForm;
                     t2display := t2Param;
                     bestCorr2 := corrT2
                  End; {then}
                  killTree(curTree);
                  curXForm := curXForm + stepSize;
               Until (curXForm > expEnd+loopTol);
            End; {then}

            If xForms^.doRho Then With xForms^ Do Begin
               treeHeight := findHeight(tweakedTree);
               bestrhoCorr1 := -2.0; {Less than any possible correlation}
               bestrhoCorr2 := -2.0; {Likewise...}
               curRho := rhoStart;
               Repeat
                  curTree := rhoXformTree(tweakedTree,curRho,treeHeight);
                  zeroLenTip := false;
                  checkZeroTips(curTree,zeroLenTip);
                  If zeroLenTip Then Begin
                     str(curXForm,tmpStr);
                     xFormZLTipError('Grafen''s (1992) Rho of '+tmpStr);
                  End; {then}
                  addData(curTree,simData,tipsOnly);
                  correlationScore(curTree,xForms^.cost,corrT1,corrT2);
                  If (abs(bestRhoCorr1)>abs(corrT1)) Then Begin
                     rho1 := curRho;
                     bestRhoCorr1 := corrT1
                  End; {then}
                  If (abs(bestRhoCorr2)>abs(corrT2)) Then Begin
                     rho2 := curRho;
                     bestRhocorr2 := corrT2
                  End; {then}
                  killTree(curTree);
                  curRho := curRho + stepSize;
               Until (curRho > rhoEnd+loopTol);
               If (abs(bestRhocorr1)<abs(bestCorr1)) Then Begin
                  t1Param := rho1;
                  t1Display := rho1;
                  t1type := 'R';
                  bestCorr1 := bestRhocorr1;
               End; {then}
               If (abs(bestRhoCorr2)<abs(bestCorr2)) Then Begin
                  t2Param := rho2;
                  t2Display := rho2;
                  t2type := 'R';
                  bestCorr2 := bestRhoCorr2;
               End; {then}
            End; {then}


            With xForms^ Do Begin
               If doBLStats Then Begin
                  contrastCount := 0;
                  Case t1Type of
                      'R' : tree1 := rhoXFormTree(tweakedTree,t1Param,treeHeight);
                      'L' : tree1 := logXFormTree(tweakedTree);
                      'E' : tree1 := xFormTree(tweakedTree,t1Param);
                      'G' : tree1 := grafenXFormTree(tweakedTree);
                      'N' : tree1 := neeXFormTree(tweakedTree);
                      'P' : tree1 := pagelXFormTree(tweakedTree);
                      'O' : tree1 := oneXFormTree(tweakedTree);
                      'S' : tree1 := copyTree(tweakedTree);
                  End; {case t1Type}
                  Case t2Type of
                      'R' : tree2 := rhoXFormTree(tweakedTree,t2Param,treeHeight);
                      'L' : tree2 := logXFormTree(tweakedTree);
                      'E' : tree2 := xFormTree(tweakedTree,t2Param);
                      'G' : tree2 := grafenXFormTree(tweakedTree);
                      'N' : tree2 := neeXFormTree(tweakedTree);
                      'P' : tree2 := pagelXFormTree(tweakedTree);
                      'O' : tree2 := oneXFormTree(tweakedTree);
                      'S' : tree2 := copyTree(tweakedTree);
                  End; {case t2Type}
                  addData(tree1,simData,tipsOnly);
                  addData(tree2,simData,tipsOnly);
                  calcBLContrasts(tree1,tree2,bestBLContrasts,contrastCount);
                  bLContrastStats(bestBLContrasts,
                                  contrastCount,
                                  tree1,
                                  tree2,
                                  t1Param,
                                  t2Param,
                                  t1type,
                                  t2type,
                                  simStats,
                                  correlations);
                  write2TreeStats(resultFile,
                                  simStats,
                                  simNumber,
                                  correlations,
                                  t1Display,
                                  t2Display,
                                  t1type,
                                  t2type,
                                  tree1,
                                  tree2,
                                  tipStat);
                  killTree(tree1);
                  killTree(tree2)
               End
               Else writeXFormResults(t1Param,t2Param,t1type,t2type,resultFile)
            End; {else}

         End; {else}
         If doTweak Then killTree(tweakedTree);
         { otherwise tweakedTree=analysisTree }

         killData(SimData);
         readDataSet(simData,tipCount,simFile);

         If (simData <> nil) Then Begin
            If (doCategories  <> [])
            Then categorizeData(simData,
                                analysisTree,
                                binCount,
                                binBase,
                                globalMin,
                                globalMax)
            Else If (doRanks <> [])
                 Then Begin
                    rankData(simData,tipCount,doRanks);
                    assignRanks(simData,doRanks);
                 End; {then}
            computeTipStats(simData,tipStat);
         End; {then}

      Until((simData = nil) or (simNumber >= userLimit));

      sound(880);
      delay(750);
      nosound;

      If (analysisTree <> inputTree) Then killTree(analysisTree);
      killTree(inputTree);

      close(simFile);
      close(resultFile);
      If (tweaksToSave > 0) Then close(tweaksFile);

      append(paramFile);
      writeln(paramFile);
      writeln(paramFile,'PDERROR processed ',simNumber,' simulated data sets');
      writeln(paramFile);
      writeln(paramFile,
               '***************************************************************************');
      writeln(paramFile);

      writeln('The name of the output file is ',resultFileName,'.');
      writeln('The file ',paramFileName,' contains a summary of your choices.');
      write('Press any key to see a list of output file columns.');
      userChar := readKey;
      clrscr;
      If (xForms <> nil) Then
         If xForms^.doBLStats
         Then Begin
            display2TreeSummary(output);
            display2TreeSummary(paramFile)
         End
         Else Begin
            displayXFormSummary(output);
            displayXFormSummary(paramFile);
         End
      Else Begin
         display1TreeSummary(output);
         display1TreeSummary(paramFile);
      End; {Else}
      write('Press any key to exit to DOS');
      userChar := readkey;
      close(paramFile);

   End. {program PDERROR}
