{***************************************************************************}
{This is program PDSIMUL.  This program simulates evolution up phylogentetic
trees.  It can implement five basic models of evolutionary change:
        Gradual Brownian Motion, Speciational Brownian Motion,
        Gradual Ornstein-Uhlenbeck, Speciational Ornstein-Uhlenbeck,
        Punctuated Equilibrium (change allowed in only one daughter).
These models are described in greater detail in the accompanying
documentation and in the following manuscripts:
Garland, T., Jr., A. W. Dickerman, C. M. Janis, and J. A. Jones. 1993.
         Phylogenetic analysis of covariance by computer simulation.
         Systematic Biology 42(3):265-292.
Martins, E. P., and T. Garland, Jr. 1991. Phylogenetic analyses of the
         correlated evolution of continuous characters: a simulation study.
         Evolution 45:534-557.
The program further allows for bounds on the values obtainable by the traits,
and allows the user to set numerous paramters such as the starting values and
variances of the traits.
       This program requires the file PDSHELP.TXT to be in the current
directory or in a directory specified in the DOS PATH parameter.  It was
compiled from the following Turbo Pascal 6.0 units:
         s_menu1.pas
         s_util1.pas
         s_io1.pas
         s_sim1.pas
         s_simul1.pas
         s_stat2.pas
         t_util1.pas
         te_util1.pas
         t_io1.pas
and of course pdsimul.pas.  The units starting with 't' were originally
used by PDTREE, but many of their procedures were usefull for this program
as well so they are used here too.  This program is organized as follows:

pdsimul - reads in the help information, draws the introductory screen, and
initializes the paramters.

s_menu - runs the main menu screen.  Allows the user to read in trees, to
get help information, and to change simulation parameters.

s_io1 - uses t_io1 to read in a tree from disk.

s_sim1 - checks the parameters for compatability and prints a simulation
summary to the screen.  If the parameters were bad, it returns to pdsimul.
This unit is also responsible for writing the tree structure to the disk.

s_simul1 - actually does the simulations.  It is invoked by s_sim1 only after
all the parameters are checked.

s_stat2 - does the statistical analysis after the simulations if the user
selects it to.

s_util1 - hold some procedures and all the type definitions unique to this
program.

t_util1 and te_util1 - hold some procedures and type definitions used by
this program and by pdtree.}
{***************************************************************************}
Program PDSIMUL;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Uses crt,
     dos,
     s_util1,
     t_util1,
     s_menu1;
Var
  ch :char;
  parm : parameter;

Function heapfunc(size:word): integer; far;
Begin
  heapfunc:=1;
End;

{***************************************************************************}
{This simply writes introductory text to the screen.}
{***************************************************************************}
   Procedure DoIntro;
   Begin
      clrscr;
      Writeln('        ********* Phenotypic Diversity Analysis Program *********');
      writeln;
      Writeln('                             PDSIMUL Module');
      writeln;
      Writeln('V. 2.0 Copyright 6 Jan 2000 by Jason A. Jones ',
                     'and Theodore Garland, Jr.');
      Writeln('                   Department of Zoology, 430 Lincoln Drive,');
      Writeln('              University of Wisconsin, Madison, WI 53706  U.S.A.');
      Writeln('              Telephone:  608-262-4437; Facsimile:  608-265-6320');
      Writeln('                      Email:  tgarland@facstaff.wisc.edu');
      writeln;
      Writeln('This program simulates the (correlated) evolution of',
              ' two continuous characters');
      Writeln('along a known phylogenetic tree.  The input file must be',
                    ' in one of 3 formats,');
      Writeln('all of which can be created or interconverted using PDTREE:');
      Writeln('    .INP (see Martins and Garland [1991] ',
                               'Evolution 45:534-557)');
      Writeln('    .PDI (T. Garland, Jr., A. W. Dickerman,',
              ' C. M. Janis, and J. A. Jones,');
      Writeln('                1993, Systematic Biology 42:265-292.');
      Writeln('    .BRK (requires either a .TIP file or estimates of means and variances)');
      writeln;
      Writeln('Evolution can be simulated under various models',
              ' of character change, and');
      Writeln('by default maintaining means and variances similar to',
              ' the original tip data.');
      Writeln('Additionally, character evolution can be either',
              ' Bounded or Unbounded');
      Writeln('         (the former will affect variances; see Garland et al., 1993).');
      writeln;
      ret('                     Please press any key to continue.');
   End;  {procedure dointro}
{***************************************************************************}
{This procedure reads in data from a file called pdshelp.txt.  It searches
the current directories as well as all the directories in the DOS PATH
parameter.  The information read in is the information sent to the help
box on the main menu screen.}
{***************************************************************************}
   Procedure readhelp;
      Var
         i,j :integer;
         fl : text;
         flename:string;
         dum :string;
         exists :boolean;
      Begin
         flename:=FExpand(Fsearch('pdshelp.txt',GetEnv('PATH')));
         {check to see if the file exists}
         {$I-}   {turn off i/o checking}
         Assign(fl,flename);
         Reset(Fl);
         Close(Fl);
         {$I+}   {turn i/o checking back on}
         exists:= (IOResult=0) and (flename<>'');

         If Not exists Then Begin
            clrscr;
            writeln('Could not find "PDSHELP.TXT" in the current directory or');
            writeln('  in any of the directories in the DOS PATH command.');
            writeln('Please indicate where this file can be found.');
            writeln('Example:  enter "C:\PD\" if it is in the C:\PD subdirectory.');
            readln(flename);
            flename:=flename+'pdshelp.txt';
            {new directory, check again}

            {$I-}                {repeat the whole i/o checking business}
            Assign(fl,flename);
            Reset(Fl);
            Close(Fl);
            {$I+}
            exists:= (IOResult=0) and (flename<>'');
            If Not exists Then Begin
               writeln;
               writeln('Still could not find pdshelp.txt.');
               writeln('  Because of this, all the help information will be');
               writeln('  absent or distorted.');
               ret('Please Press any key to continue.');
            End; {then}
         End; {then}
         If exists Then Begin
            Reset(fl);
            For i:=1 To 10 Do {skip first 10 lines}
              readln(fl,dum);
            For i:=1 To 30 Do Begin
               readln(fl,dum);  {skip two lines}
               readln(fl,dum);
               for j:=1 to 8 do {copy eight lines into the help array}
               readln(fl,helparray[i,j]);
            End; {for i}
         End; {then}
      End; {procedure readhelp}

{***************************************************************************}
{This procedure initializes the parameters that are used.  They are all
kept in 1 data structure and are heavily used by this program and its units.}
{***************************************************************************}
   Procedure initialize(Var parm :parameter);
      Begin
         parm.filename:='';
         parm.root:=nil;
         parm.tipnumber:=0;
         parm.height:=0;
         parm.mean[1]:=0;
         parm.mean[2]:=0;
         parm.variance[1]:=0;
         parm.variance[2]:=0;
         parm.correlation:=0;
         parm.gradual:=0;
         parm.starting[1]:=0;
         parm.starting[2]:=0;
         parm.selvariance[1]:=0;
         parm.selvariance[2]:=0;
         parm.brownian:=0;
         parm.selcorr:=0;
         parm.decay[1]:=0;
         parm.decay[2]:=0;
         parm.peak[1]:=0;
         parm.peak[2]:=0;
         parm.bounded:=0;
         parm.upperb[1]:=0;
         parm.upperb[2]:=0;
         parm.lowerb[1]:=0;
         parm.lowerb[2]:=0;
         parm.seed:=0;
         parm.simnum:=0;
         parm.finalmean[1]:=0;
         parm.finalmean[2]:=0;
         parm.nordheim:=0;
         parm.contemp:=false;
         parm.peakvelocity[1]:=0;
         parm.peakvelocity[2]:=0;
         parm.meanvariance[1]:=0;
         parm.meanvariance[2]:=0;
         parm.peakvar[1]:=0;
         parm.peakvar[2]:=0;
               
         parm.min[1]:=0;
         parm.max[1]:=0;
         parm.min[2]:=0;
         parm.max[2]:=0;
      End; {initialize}

{***************************************************************************}
{Here is the body of the program.  It simply initializes the paramters,
reads in the help menu information, then calls a procedure to run a menu
to input the parameters.}
{***************************************************************************}
   Begin
     heaperror:=@heapfunc;
     readhelp;
     dointro;
     initialize(parm);
     runmenu(parm);
     If parm.root<> nil Then
       killtree(parm.root);                 {Dispose all the pointers.}
   End. {program pdsimul}
