{****************************************************************************}
{This menu-driven program loads in trees of the .PDI, .INP, and .BRK
formats from disk to a dynamically linked list.  (The .INP format is
compatible with the Comparative Method Analysis Program of Martins and
Garland [1991] Evolution 45:534-557.)  The program PDDIST can be used to
convert trees in .PDI or .INP format to the .DIS format.

A pointer, root, points to the root of this tree.

While graphically editing the tree, the user may choose to only deal with a
smaller subset tree within the larger tree.  The pointer root or rt or r
points to this smaller tree.  This is handled completely by the editing unit.
By default rt = root.  Variables generally are not used globally.  All the
global variables that a procedure uses are passed to it.  However, the
constants are used globally.  They can be found in the utility unit.  The
layout of the program is as follows.  Several general purpose procedures are
in the t_util unit with the type definitions and constants.  The other three
units visible from here support the main menu, one procedure for each item
on the Main Menu - 4.  Each of these 4 procedures has many local procedures
within it.  Just before the body of the main program is a short procedure
to print the title screen.  A more detailed description of the unit structure
follows.  All the units are necessary to compile the Turbo Pascal 6 code.

PDTREE: The unit you are in now.  It uses the next 5 + T_Util1.
        (Title screen)

T_IO1: The input and output unit.  It does all the tree I/O as well as the
       new tree creation.  It uses only T_Util1.
       (all Input/Output screens)
          

T_FELS_M: This unit provides the menu to allow the user to do analyses
          using Felsenstein's (1985, Am. Nat. 125:1-15) phylogenetically
          independent contrasts.  It uses the following two units +
          T_UTIL1. (Felsenstein menu)

T_FELS_D: The unit that does the analysis and displays the 8 X-Y diagnostic
          plots for Felsenstein's phylogenetically independent contrasts.
          It uses TE_UTIL1 and T_UTIL1. (Display screens 1-8)

T_FELS_R: The unit that does the analysis and display the 3 regression
          related plots and an additional screen of root node related
          statistics. It uses TE_UTIL1 and T_UTIL1. (Display screens 9-11)

T_EDIT: The unit that allows the user to edit the tree in memory.  It further
        divides the task into three parts, which are implemented in the
        following three units.  It uses the following three units + TE_UTIL1
        and T_UTIL1.  (Edit Module Menu screen)

TE_SEE: This unit provides for the graphic editing of the tree.  Most of the
        user interaction is done through the graphics screen.  It uses
        TE_UTIL1 and T_UTIL1.

TE_BRCH: This unit provides for the branch editing features.  It uses
         TE_UTIL1 and T_UTIL1.  (Branch Length manipulation menu)

TE_TIP1: This unit provides the tip data editing features on the text screen.
         It uses only T_UTIL1.  (Tip data editing screen)

TE_UTIL1: This utility unit has among other things a procedure that turns on
          the graphics screen.  It uses T_UTIL1.

T_UTIL1: This unit provides general purpose procedures as well as any
         procedures that need to be shared between units.  It uses nothing.
         The actual display procedures used by T_FELSEN were moved here in
         October 1996.  This unit also contains the code for calculating t
         statistics. ('Which trait along X axis' text screen and Felsenstein
         plot screens)
{****************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


Program PDTree;
   uses
      crt,
      graph,
      t_util1,      {This unit has constants, type definitions, and utility
                      procedures.}
      t_io1,        {This unit has the input/output tree from disk procedures
                      as well as the create tree procedure.}
      t_edit1,      {This unit has one externally visible procedure.  It
                      supports all of the tree editing features.}
      t_fels_m;     {This unit computes, displays, and saves the Felsenstein
                      contrasts.}
   Const asterisks =
                '********************************************************';
   Var
      filename : flstring;
      extension: typstring;
      root : nodeptr;
      ch : char;
      liveFile :boolean;

{***************************************************************************}
{Here the program just writes a title page for the program.  It is only
called by the main program and only once.}
{***************************************************************************}
Procedure doIntro;
   Begin
      clrscr;
      Writeln(' ********* Phenotypic Diversity Analysis Program - PDTREE Module *********');
      Writeln('       Version 5.0 Copyright 19 March 2003 by Theodore Garland, Jr.,');
      Writeln('                   Peter E. Midford, and Jason A. Jones');
      Writeln('Questions to T.G. at Dept. of Biology, Univ. of Calif., Riverside, CA 92521');
      Writeln('    Phone 909-787-3524; Fax 909-787-4286; E-mail tgarland@citrus.ucr.edu');
      Writeln;
      Writeln('This program creates, edits, and converts trees in 1 of 3 formats:');
      Writeln('    .INP (see Martins and Garland [1991] Evolution 45:534-557)');
      Writeln('    .PDI (Garland, Dickerman, Janis, & Jones [1993] Syst. Biol. 42:265-292)');
      Writeln('    .BRK, which can be used to exchange trees with MacClade, PAUP or PHYLIP.');
      Writeln('Use the accompanying PDDIST module to create various distance matrices.');
      Writeln('The .INP, .PDI, and .BRK files are in plain ASCII format,');
      Writeln('     so the tip data can also be edited with any word processor.');
      Writeln('The program also allows computation of phylogenetically independent contrasts');
      Writeln('     as described in Felsenstein (1985) Am. Nat. 125:1-15.  Polytomies are');
      Writeln('     handled as described in Purvis and Garland (1993) Syst. Biol. 42:569-575.');
      Writeln('The program allows transformations of branch lengths as described in:');
      Writeln('     Grafen (1989) Phil. Trans. R. Soc. London Ser. B 326:119-157 and');
      Writeln('     Garland, Harvey, and Ives (1992) Syst. Biol. 41:18-32,');
      Writeln('     as well as editing and transformations of the tip data.');
      writeln('Finally, ancestral reconstructions and regression confidence and prediction');
      writeln('     intervals are computed (Am. Zool. 39:374-388; Am. Nat. 155:346-364.)');
          ret('                    Please press any key to continue.');
   End; {procedure doIntro}
{***************************************************************************}
{            The main program is basically a menu and little more.          }
{***************************************************************************}
   Begin                 { THE MAIN PROGRAM BEGINS HERE!!! }
      loadGraphics;
      root:=nil;
      doIntro;
      liveFile:=false;            {No file in memory}
      Repeat                      {until user chooses to quit from the program}
         Repeat                   {until user makes valid choice from main menu}
            clrscr;
            writeln('*********************** Main Menu **********************');
            writeln;
            If liveFile                    {If a file is in memory, tell us}
            Then writeln('The tree in memory is: ',filename)
            Else writeln('No tree is in memory.');
            writeln('********************************************************');
            writeln('Which of the following would you like to do?');
            writeln('     (I)nput a tree (.INP, .PDI, or .BRK format).');
            writeln('     (C)reate a tree from scratch.');
            writeln('     (E)dit the tree currently in memory.');
            writeln('     (A)nalyze data with Felsenstein''s (1985) contrasts.');
            writeln('     (S)ave the tree currently in memory to disk.');
            writeln('     (Q)uit this program and exit to DOS.');
            writeln('********************************************************');
            writeln('Please choose (I/C/E/A/S/Q)');
            ch:=UpCase(readkey);
         Until ch In ['I','C','E','S','Q','A'];
         Case ch Of
            'I' : InputTree(root,filename,extension,livefile);
            'C' : CreateTree(root,livefile,filename,extension);
            'E' : EditTree(root,livefile,filename);
            'S' : SaveTree(root,filename,extension,livefile);
            'A' : Felsenstein(root,livefile,filename);
         End; {case ch}
         If (ch='Q') And liveFile Then Begin
            Repeat
               clrscr;
               Writeln('You have the tree:  ',filename,' in memory.');
               Writeln('If you quit the program now and have not saved the');
               writeln('tree, then all the changes you have made will be',
                                       ' lost.');
               writeln;
               writeln('Do you want to quit the program (Y/N)?');
               ch:=upCase(readKey);
            Until ch in ['Y','N'];
            If ch='Y' Then ch:='Q';
         End; {then}
      Until ch='Q';
      clrscr;
      If root<> nil
      Then killTree(root);   {Dispose all the pointers.}
   End.  {program PDTREE}
