{**************************************************************************}
{This is unit s_simul1.  This unit is used by PDSIMUL and actually does the
simulations.  It relies on the parameters passed to it to be consistent,
the tree valid, and the nodes linked in the order they are output.  It is
still possible though for the user to input parameters that make the
simulations take an arbitrarily large amount of time.  It is especially
easy with the "Throw Out" bounding algorithm for the user to make the
simulations take much longer than anticipated.}
{**************************************************************************}
unit s_simul1;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

interface

uses s_util1;

procedure simulate(var outfile: text; p :parameter);

implementation

uses   crt,
       t_util1;


{***************************************************************************}
{provides two independent variables from normal distribution}
{***************************************************************************}
   Procedure GetNorChar (var X, Y : real);
   var
      A, B, R, Theta : real;
   begin
      a := Random;
      b := Random;
      Theta := 2*PI*A;
      R := sqrt(-2*(ln(B)));
      X := R*(cos(Theta));
      Y := R*(sin(Theta));
   end;  {getnorchar}
{***************************************************************************}
{gets two normal variates from getnorchars
 and output two normal variates with mean = zero, variance = 1}
{***************************************************************************}
   Procedure CorChars (var X, Y : real; R : real);
   var
      c1, c2 : real;
   begin
      getnorchar(c1, c2);
      X := c1;
      Y := (R*c1 + sqrt(1-R*R) * c2);
   end; {corchars}

{**************************************************************************}
{This uses CorChar to give the velocity of change.  This is in units of
change per unit branch length.}
{**************************************************************************}
procedure getchangevelocity(time,corr :real;variance,vel:traits;
                              var velocity:traits);
   Var a,b : real;
begin
  CorChars(a,b,corr);
  velocity[1] := a;
  velocity[2] := b;
  if time<>0 then
    begin
      velocity[1]:=velocity[1]*variance[1]/sqrt(time) + vel[1];
      velocity[2]:=velocity[2]*variance[2]/sqrt(time) + vel[2];
    end
  else
    begin
      velocity[1]:=0;
      velocity[2]:=0;
    end;
end;

{**************************************************************************}
{Here are the bounding procedures.  The one directly below is for the
"Truncate Change" method of bounding.  When a trait goes out of bounds and
this method of bounding is selected, then this procedure is called and
it sets the trait that is out of bounds to have the value of the nearest
boundary.}
{**************************************************************************}
procedure truncatechange(nd: nodeptr; p:parameter; whichone:integer);
begin
  if nd^.dat[whichone]>p.upperb[whichone] then
    nd^.dat[whichone] := p.upperb[whichone]
  else
    if nd^.dat[whichone]<p.lowerb[whichone] then
      nd^.dat[whichone] := p.lowerb[whichone];
end;

{**************************************************************************}
{This procedure is called whenever a trait goes out of bounds and the "Flip"
method of bounding is selected.  The trait is out of bounds before this
procedure was called so 2 times the size of the step must be subtracted from
the trait.  The effect is that the trait steps in the other direction.  If
it is still out of bounds, truncate change is used to pull it back.}
{**************************************************************************}
procedure flip(nd: nodeptr; p :parameter; whichone:integer);
begin
  nd^.dat[whichone]:=nd^.dat[whichone] - 2*(nd^.dat[whichone] -
          nd^.ancestor^.dat[whichone]);
  if (nd^.dat[whichone]<p.lowerb[whichone]) or
                      (nd^.dat[whichone]>p.upperb[whichone]) then
    truncatechange(nd,p,whichone);
end;

{**************************************************************************}
{This procedure is called when a trait is out of bounds and the "Hard Bounce"
method of bounding is selected.  Because the trait is already out of bounds,
2 times the distance it is out must be taken away from the value (or added,
depending on the sign of the step).  The effect is that the trait started
at the value its ancestor had.  It then moves toward a bound, hits it, and
bounces off moving away from the bound.}
{**************************************************************************}
procedure hardbounce(nd: nodeptr; p :parameter; whichone:integer);
begin
  if nd^.dat[whichone]>p.upperb[whichone] then
    nd^.dat[whichone]:= nd^.dat[whichone] - 2*(nd^.dat[whichone] -
          p.upperb[whichone])
  else
    nd^.dat[whichone]:= nd^.dat[whichone] - 2*(nd^.dat[whichone] -
          p.lowerb[whichone]);
  if (nd^.dat[whichone]<p.lowerb[whichone]) or
                      (nd^.dat[whichone]>p.upperb[whichone]) then
    hardbounce(nd,p,whichone);
end;

{**************************************************************************}
{This procedure is like the last except that any time the ancestors value is
near a bound (i.e., closer than 1/5 of the region in bounds) a modifier is
multiplied into the step if the step is near a bound.  The nearer the bound,
the smaller the multiplier.  The effect is that when the trait is near a
bound, it is hard for the trait to step toward the bound, but easy for it to
step away.  Note: a trait does not have to be out of bounds to have its step
modified.  In addition, no trait is allowed to decrease its distance
from a bound by more than one half.}
{**************************************************************************}
procedure softbounce(nd :nodeptr; p:parameter);
var
  step, top, bottom : traits;
begin
  step[1]:= nd^.dat[1] - nd^.ancestor^.dat[1];
  step[2]:= nd^.dat[2] - nd^.ancestor^.dat[2];
  top[1]:=p.upperb[1]-nd^.ancestor^.dat[1];
  top[2]:=p.upperb[2]-nd^.ancestor^.dat[2];
  bottom[1]:=nd^.ancestor^.dat[1] - p.lowerb[1];
  bottom[2]:=nd^.ancestor^.dat[2] - p.lowerb[2];
  if step[1] > 0 then
    begin
      if nd^.ancestor^.dat[1] > (p.upperb[1] - p.region[1]) then
        step[1]:=step[1]*sin((pi/2.0)*(top[1]/p.region[1]));
      if step[1] > top[1]/2 then
        step[1]:=top[1]/2;
    end
  else
    begin
      if nd^.ancestor^.dat[1] < (p.lowerb[1] + p.region[1]) then
        step[1]:=step[1]*sin((pi/2.0)*(bottom[1]/p.region[1]));
      if abs(step[1]) > bottom[1]/2 then
        step[1]:= -1*bottom[1]/2;
    end;
  if step[2] > 0 then
    begin
      if nd^.ancestor^.dat[2] > (p.upperb[2] - p.region[2]) then
        step[2]:=step[2]*sin((pi/2.0)*(top[2]/p.region[2]));
      if step[2] > top[2]/2 then
        step[2]:=top[2]/2;
    end
  else
    begin
      if nd^.ancestor^.dat[2] < (p.lowerb[2] + p.region[2]) then
        step[2]:=step[2]*sin((pi/2.0)*(bottom[2]/p.region[2]));
      if abs(step[2]) > bottom[2]/2 then
        step[2]:= -1*bottom[2]/2;
    end;
  nd^.dat[1]:=nd^.ancestor^.dat[1]+step[1];
  nd^.dat[2]:=nd^.ancestor^.dat[2]+step[2];
end;

{**************************************************************************}
{Here we check to see if the traits just stepped out of bounds.  If one or
both did, then we use the appropriate routine to deal with them.  Note that
if "Soft Bounce" is selected then we must have it check to see if any
modifying must be done as the trait need not be out of bounds.}
{**************************************************************************}
procedure bounds(nd:nodeptr; p:parameter; var redo,abort:boolean);
var
  out1, out2:boolean;
  number : integer;
begin
  if (nd^.dat[1]<p.lowerB[1]) or (nd^.dat[1]>p.upperB[1]) then
    out1:=true
  else
    out1:=false;
  if (nd^.dat[2]<p.lowerB[2]) or (nd^.dat[2]>p.upperB[2]) then
    out2:=true
  else
    out2:=false;
  if p.bounded=7 then
    begin
      number:=trunc(random(5));
      number:=number+2;
    end
  else
    number:=p.bounded;
  if out1 or out2 then
    case number of
      1 : abort:=true;
      2 : redo:=true;
    end;
  if out1 then
    case number of
      3 : flip(nd,p,1);
      4 : hardbounce(nd,p,1);
      6 : truncatechange(nd,p,1);
    end;
  if out2 then
    case number of
      3 : flip(nd,p,2);
      4 : hardbounce(nd,p,2);
      6 : truncatechange(nd,p,2);
    end;
  if number=5 then
    softbounce(nd,p);
end;

{**************************************************************************}
{This Procedure simulates evolution up the tree.  It is only used when
gradual OU simulations are selected.  It only does 1 simulation.  The
results are in the tree when this procedure is done.  This is a recursive
procedure.  It first checks that it should change the traits with the
first 2 "if" statements.  It then uses getchangevelocity to get the change
velocity of the peak.  It then gets the change velocity of the trait.  Then
it implements an equation explained in the documentation to find the value
of the traits at the next node.  If there is bounding it is checked with
a call to bounds.  If bounds does not pass back "redo" true then we recurse.
Else we pull another change for the traits.}
{**************************************************************************}
procedure gradualOU(p :parameter; nd:nodeptr; p0: traits;var abort:boolean);
var
  E,mp,vel,D,x0 :traits;
  redo :boolean;
  ch :char;
begin
  vel[1]:=0;
  vel[2]:=0;
  if (nd<>p.root) then
    begin
      if nd^.length<>0 then
        begin
          getchangevelocity(nd^.length,0,p.peakvar,p.peakvelocity,mp);
          x0[1]:=nd^.ancestor^.dat[1];
          x0[2]:=nd^.ancestor^.dat[2];
          repeat
            redo:=false;
            getchangevelocity(nd^.length,p.selcorr,p.nordvar,vel,E);
            D[1]:=x0[1]-p0[1]-(E[1]+mp[1])/p.decay[1];
            D[2]:=x0[2]-p0[2]-(E[2]+mp[2])/p.decay[2];
            if p.decay[1]*nd^.length>80 then
              nd^.dat[1]:=mp[1]*nd^.length + x0[1] - D[1]
            else
              nd^.dat[1]:=D[1]*exp(-1*p.decay[1]*nd^.length) + mp[1]*nd^.length +
                   x0[1] - D[1];
            if p.decay[2]*nd^.length>80 then
              nd^.dat[2]:=mp[2]*nd^.length + x0[2] - D[2]
            else
              nd^.dat[2]:=D[2]*exp(-1*p.decay[2]*nd^.length) + mp[2]*nd^.length +
                   x0[2] - D[2];
            if p.bounded<>0 then
              bounds(nd,p,redo,abort);
          until not redo;
          p0[1]:= p0[1] + mp[1]*nd^.length;
          p0[2]:= p0[2] + mp[2]*nd^.length;
        end
      else
        nd^.dat:=nd^.ancestor^.dat;
    end;
  if (nd^.sibL<>NIL) and (not abort) then
    begin
      gradualOU(p,nd^.sibL,p0,abort);
      if not abort then
        gradualOU(p,nd^.sibR,p0,abort);
    end;
end;

{**************************************************************************}
{This procedure is similar to the one before, except that it implements
a different equation and does not take branch lengths into account.}
{**************************************************************************}
procedure speciationalOU(p:parameter; nd:nodeptr; p0:traits;
                                     var abort:boolean);
var
  redo : boolean;
  vel,D,mp,E :traits;
begin
  vel[1]:=0;
  vel[2]:=0;
  if nd<>p.root then
    begin
      if nd^.length<>0 then
        begin
          getchangevelocity(1,0,p.peakvar,p.peakvelocity,mp);
          repeat
            redo:=false;
            getchangevelocity(1,p.selcorr,p.nordvar,vel,E);
            D[1]:=(nd^.ancestor^.dat[1] + E[1]) - p0[1];
            D[2]:=(nd^.ancestor^.dat[2] + E[2]) - p0[2];
            if p.decay[1]>80 then
              nd^.dat[1]:=p0[1]
            else
              nd^.dat[1]:=D[1]*exp(-1*p.decay[1]) + p0[1];
            if p.decay[2]>80 then
              nd^.dat[2]:=p0[2]
            else
              nd^.dat[2]:=D[2]*exp(-1*p.decay[2]) + p0[2];
            if p.bounded<>0 then
              bounds(nd,p,redo,abort);
          until not redo;
          p0[1]:=p0[1] + mp[1];
          p0[2]:=p0[2] + mp[2];
        end
      else
        nd^.dat:=nd^.ancestor^.dat;
    end;
  if (nd^.sibL<>NIL) and (not abort) then
    begin
      speciationalOU(p,nd^.sibL,p0,abort);
      if not abort then
        speciationalOU(p,nd^.sibR,p0,abort);
    end;
end;

{**************************************************************************}
{This is again similar to the one before except that it does not implement
an OU equation.  Instead, it just adds on the changes due to the mean
movement and the Brownian trait movement.}
{**************************************************************************}
procedure gradualbrownian(p: parameter; nd:nodeptr; var abort :boolean);
var
  vel,mv,v : traits;
  redo :boolean;
begin
  vel[1]:=0;
  vel[2]:=0;
  if nd<>p.root then
    begin
      if nd^.length<>0 then
        begin
          getchangevelocity(nd^.length,0,p.meanvariance,p.meanvelocity,mv);
          repeat
            redo:=false;
            getchangevelocity(nd^.length,p.selcorr,p.nordvar,vel,v);
            nd^.dat[1]:=nd^.ancestor^.dat[1] + (mv[1] + v[1])*nd^.length;
            nd^.dat[2]:=nd^.ancestor^.dat[2] + (mv[2] + v[2])*nd^.length;
            if p.bounded<>0 then
              bounds(nd,p,redo,abort);
          until not redo;
        end
      else
        nd^.dat:=nd^.ancestor^.dat;
    end;
  if (nd^.sibL<>NIL) and (not abort) then
    begin
      gradualbrownian(p,nd^.sibL,abort);
      if not abort then
        gradualbrownian(p,nd^.sibR,abort);
    end;
end;

{**************************************************************************}
{This is just like the one before except that it disregards the branch
lengths.}
{**************************************************************************}
procedure speciationalbrownian(p:parameter; nd:nodeptr; var abort:boolean);
var
  vel, mv, v :traits;
  redo :boolean;
begin
  vel[1]:=0;
  vel[2]:=0;
  if nd<>p.root then
    begin
      if nd^.length<>0 then
        begin
          getchangevelocity(1,0,p.meanvariance,p.meanvelocity,mv);
          repeat
            redo:=false;
            getchangevelocity(1,p.selcorr,p.nordvar,vel,v);
            nd^.dat[1]:=nd^.ancestor^.dat[1] + mv[1] + v[1];
            nd^.dat[2]:=nd^.ancestor^.dat[2] + mv[2] + v[2];
            if p.bounded <> 0 then
              bounds(nd,p,redo,abort);
            until not redo;
        end
      else
        nd^.dat:=nd^.ancestor^.dat;
    end;
  if (nd^.sibL<>NIL) and (not abort) then
    begin
      speciationalbrownian(p,nd^.sibL,abort);
      if not abort then
        speciationalbrownian(p,nd^.sibR,abort);
    end;
end;

{**************************************************************************}
{This is just like the one before except that it only changes one sister
node.  It implements this with an additional variable called changeme.
Note: we multiply the nordheim coefficient by the square root of two to
account for changes occuring in only one sibling.  We don't know why the
square root of two works, but it seems to work.}
{**************************************************************************}
procedure punctuatedbrownian(p:parameter; nd:nodeptr;
                        var abort:boolean; changeme:boolean);
var
  vel, mv, v,dumvar :traits;
  redo :boolean;
  dumseed : longint;
begin
  vel[1]:=0;
  vel[2]:=0;
  if (nd<>p.root) then
    begin
      if changeme and (nd^.length<>0) then
        begin
          dumvar[1]:=p.nordvar[1]*sqrt(2);        {We don't know why this}
          dumvar[2]:=p.nordvar[2]*sqrt(2);        {works, but it seems to.}
          getchangevelocity(1,0,p.meanvariance,p.meanvelocity,mv);
          repeat
            redo:=false;
            getchangevelocity(1,p.selcorr,dumvar,vel,v);
            nd^.dat[1]:=nd^.ancestor^.dat[1] + mv[1] + v[1];
            nd^.dat[2]:=nd^.ancestor^.dat[2] + mv[2] + v[2];
            if p.bounded <> 0 then
              bounds(nd,p,redo,abort);
          until not redo;
        end
      else
        nd^.dat:=nd^.ancestor^.dat;
    end;
  if (nd^.sibL<>NIL) and (not abort) then
    begin
      if random<0.5 then
        begin
          punctuatedbrownian(p,nd^.sibL,abort,true);
          punctuatedbrownian(p,nd^.sibR,abort,false);
        end
      else
        if not abort then
          begin
            punctuatedbrownian(p,nd^.sibL,abort,false);
            punctuatedbrownian(p,nd^.sibR,abort,true);
          end;
    end;
end;

{**************************************************************************}
{Here we put the trait data from the tree into a data structure in memory
where it will be stored until the data structure fills up.  When this
happens it is dumped to disk.  The data structure I speak of is a linked
list.  Diskbufptr points to the next available spot.  Diskbufnum keeps track
of the number of filled elements in the list +1.  Here we just put the data
where diskbufptr points and increment diskbufnum.
{**************************************************************************}
  procedure puttree(nd,rt:nodeptr;
                    var diskbufptr:datptr;
                    var diskbufnum:longint;
                    currentCount : integer);
    var
      ch :char;
    begin
      diskbufnum:=diskbufnum+1;
      diskbufptr^.name1:=nd^.name;
      diskbufptr^.dat1:=nd^.dat;
      diskbufptr^.simCount := currentCount;
      nd:=nd^.tipL;
      if nd<>nil then
        begin
          diskbufptr^.name2:=nd^.name;
          diskbufptr^.dat2:=nd^.dat;
          diskbufptr:=diskbufptr^.next;
          puttree(nd^.tipL,rt,diskbufptr,diskbufnum,currentCount)
        end
      else
        begin
          diskbufptr^.name2:='**';
          diskbufptr^.dat2[1]:=0.0;
          diskbufptr^.dat2[2]:=0.0;
          diskbufptr:=diskbufptr^.next;
        end;
    end;

{**************************************************************************}
{This procedure is in charge of taking the tree and extracting the results.
If there is room in the buffer for the data from this tree and one more then
the data is just put in the buffer.  Else it is put in the buffer and the
buffer is dumped to disk and reset.}
{**************************************************************************}
procedure dumpresults(var outfile:text; rt :nodeptr; tipnum, now,
         done :integer; var diskbufnum:longint; var diskbufptr :datptr;
            diskbufmax:longint; diskbufstart:datptr);
var
  tip,node: nodeptr;
  i,j,where, numsims2:integer;
  numsims :real;
  current : datptr;
begin
  puttree(rt^.tipR,rt,diskbufptr,diskbufnum,now);
  if (now=done) or ((diskbufnum + tipnum) > diskbufmax) then
    begin
      gotoxy(28,13);
      write('Writing results to file.');
      current:=diskbufstart;
      for i:=1 to diskbufnum do
        begin
          write(outfile,current^.name1,' ');
          extendWrite(outFile,current^.dat1[1]);
          write(outFile,' ');
          extendWrite(outFile,current^.dat1[2]);
          write(outFile,' ',current^.name2,' ');
          extendWrite(outFile,current^.dat2[1]);
          write(outFile,' ');
          extendWrite(outFile,current^.dat2[2]);
          writeln(outFile,' ',current^.simCount);
          current:=current^.next;
        end;
      diskbufnum:=0;
      diskbufptr:=diskbufstart;
      gotoxy(28,13);
      write('                             ');
    end;
end;

{**************************************************************************}
{This procedure initializes the buffer by taking all the available space
on the heap and allocating it to the disk buffer.  Again, note that the
disk buffer is a linked list.}
{**************************************************************************}
Procedure fillheap(var num:longint; var ptr:datptr);
var
  last :datptr;
begin
  new(ptr);
  last:=ptr;
  num:=1;
  repeat
    new(last^.next);
    last:=last^.next;
    if last <> NIL then
      num:=num+1;
  until last=nil;
end;

{**************************************************************************}
{This procedure frees all the space on the heap taken up by the diskbuffer.}
{**************************************************************************}
Procedure freeheap(var ptr:datptr);
var next :datptr;
begin
  repeat
    next:=ptr^.next;
    dispose(ptr);
    ptr:=next;
  until next=nil;
end;

{**************************************************************************}
{Here is the externally visible procedure.  This procedure prepares the
disk buffer with fillheap and initializes some more values.  It then does
a for loop to iterate through the number of simulations.  For each simulation
it decides the type of simulation and calls the appropriate procedure to do
the simulation.  It then saves the results with dumpresults.  Finally it
frees the memory used by the disk buffer.}
{**************************************************************************}
procedure simulate(var outfile: text; p :parameter);
var
  i :integer;
  abort:boolean;
  ch :char;
  diskbuffernum, diskbuffermax :longint;
  diskbufferptr, diskbufferstart :datptr;
begin
  diskbuffernum:=0;
  diskbuffermax:=0;
  fillheap(diskbuffermax,diskbufferptr);
  diskbufferstart:=diskbufferptr;
  p.root^.dat[1]:=p.starting[1];
  p.root^.dat[2]:=p.starting[2];
  clrscr;
  writeln('                            ',p.simnum,
                                  ' Simulations In Progress.');
  for i:=1 to p.simnum do
    begin
      gotoxy(25,12);
      writeln('Number of simulations completed: ',i);
      repeat
        abort:=false;
        if p.brownian=1 then
          begin
            if p.gradual=0 then
              gradualOU(p, p.root, p.peak, abort)
            else
              if p.gradual=1 then
                speciationalOU(p, p.root, p.peak, abort);
          end
        else
          begin
            if p.gradual=0 then
              gradualbrownian(p, p.root, abort)
            else
              if p.gradual=1 then
                speciationalbrownian(p, p.root, abort)
              else
                punctuatedbrownian(p, p.root,  abort, false);
          end;
      until not abort;
      dumpresults(outfile,p.root,p.tipnumber,i,p.simnum,
            diskbuffernum, diskbufferptr, diskbuffermax, diskbufferstart);
    end;
  freeheap(diskbufferptr);
  ret(chr(7)+'Simulations complete!  Please press any key to continue.'+
       chr(7));
end;

end.
