{************************************************************************}
{                              T_FELS_M.PAS                              }
{************************************************************************}
{This module is the entry point for the three modules that calculates,
 displays, and saves the phylogenetically
 independent contrasts as described by:

Felsenstein, J. 1985. Phylogenies and the comparative method.
     American Naturalist 125:1-15.

This module uses procedures from two other modules: T_FELS_D.PAS and
T_FELS_R.PAS.


11 different X-Y plots are displayed to the screen.  Each has a different
procedure to do it, but all are similar.  Each has a procedure that
calculates some statistics first.  This is called by the respective plot
procedure which then uses many procedures in t_util1 to do the plotting.
I wish I could have united them in some way so this file could be one tenth
the size it is, but the boss insisted that each had to be a little
different.  Oh well.  It's not pretty, but it seems to work.

In addition, Screen 9 is followed by Screens 9A, 9B, and 9C, which contain
additional information about the least-squares linear regression equation.
This information can be saved as three different files.}
{************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


unit t_fels_m;
interface
uses t_util1;

Procedure felsenstein(nd : nodeptr; livefile : boolean; filename : flstring);


implementation

uses crt,
     t_fels_d,
     t_fels_r;

{**************************************************************************}
{This procedure does two things; it checks that there are no
tip nodes with zero branch length, and it counts the number of tips.}
{**************************************************************************}
Procedure checkTwo(nd:nodePtr; Var zeroTip:boolean; Var numTips:integer);
   Begin
      If nd^.sibL<>nil Then Begin
         checkTwo(nd^.sibL,zeroTip,numTips);
         checkTwo(nd^.sibR,zeroTip,numTips);
      End {then}
      Else Begin
         inc(numTips);
         If (((nd^.ancestor^.sibL^.sibL=nil)
              And (nd^.ancestor^.sibR^.sibL=nil))
              And ((nd^.ancestor^.sibL^.length=0) And
                  (nd^.ancestor^.sibR^.length=0)))
         Then zeroTip:=true;
      End; {else}
   End; {procedure checktwo}


{************************************************************************}
{This function is another refugee from t_felsen.  It just lists the lines}
{in a .FIC file.                                                         }
{************************************************************************}
Procedure listFICLines;
   Begin
      writeln;
      writeln('Your .FIC file will contain the following fifteen columns:');
      writeln;
      writeln('     1.  Name of Contrast as Node-Node');
      writeln('     2.  Name of Node at Base of Contrast');
      writeln('     3.  Unstandardized Contrast for Trait 1');
      writeln('     4.  Unstandardized Contrast for Trait 2');
      writeln('     5.  Standard Deviation of Contrast');
      writeln('              = square root of sum of corrected branch lengths');
      writeln('     6.  Nodal Value for Trait 1');
      writeln('     7.  Nodal Value for Trait 2');
      writeln('     8.  Uncorrected Height of Base Node from Root of Tree');
      writeln('     9.  Corrected Height of Base Node from Root of Tree');
      writeln('    10.  Average Uncorrected Height of Daughter Nodes');
      writeln('    11.  Average Corrected Height of Daughter Nodes');
      writeln('    12.  Name of First Daughter Branch (Descendant-Ancestor)');
      writeln('    13.  Corrected Length of First Daughter Branch');
      writeln('    14.  Name of Second Daughter Branch (Descendant-Ancestor)');
      writeln('    15.  Corrected Length of Second Daughter Branch');
   End;{procedure listficlines}

{************************************************************************}
{This procedure dumps the statistics into a file.  It calls the various
statistics calculating procedures, and the various node value calculating
procedures to get the stats.  It then calls a 'flsts' procedure below to
write the stats collected to the file.}
{************************************************************************}
Procedure fileStats(nd:nodePtr;
                    Var fl:text;
                    ficName :flString;
                    constantIndex :integer;
                    star, tooLong:boolean);
   Var
      stpk:statPak;
      m,y0:extended;
      t,t2:integer;
      dfReduce : integer;
   {**********************************************************************}
   {This procedure writes the stats given to it to the file with the
   word zero for the means and the intercepts, and with some special
   stats at the end.}
   {**********************************************************************}
   Procedure flsts9(Var s:statPak;Var fl:text);
      {*******************************************************************}
      { This procedure mostly exists to save space.                       }
      {*******************************************************************}
      Procedure quickWrite(x : extended);
         Begin
             extendWrite(fl,x);
             write(fl,',');
         End; {local procedure quickwrite}
      Begin {*************************************************************}
         writeln(fl,'Number of contrasts: ',s.number);
         write(fl,'Minimum of X,Y coordinates: ');
         quickWrite(s.min[1]);
         extendWrite(fl,s.min[2]);
         writeln(fl);
         write(fl,'Maximum of X,Y coordinates: ');
         quickWrite(s.max[1]);
         extendWrite(fl,s.max[2]);
         writeln(fl);
         writeln(fl,'Mean of X,Y coordinates: zero, zero');
         write(fl,'Variance of X,Y coordinates: ');
         quickWrite(s.variance[1]);
         extendWrite(fl,s.variance[2]);
         writeln(fl);
         write(fl,'Covariance: ');
         extendWrite(fl,s.cov);
         writeln(fl);
         write(fl,'Pearson Product-Moment Correlation Coefficient: ');
         extendWrite(fl,s.corr);
         writeln(fl);
         writeln(fl,'Reduced Major Axis.');
         If (s.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.rma);
            writeln(fl);
            writeln(fl,'     Y intercept: zero');
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
         End; {else}
         writeln(fl,'Major Axis.');
         write(fl,'     Slope: ');
         extendWrite(fl,s.ma);
         writeln(fl);
         writeln(fl,'     Y intercept: zero');
         writeln(fl,'Least Squares Regression.');
         If (s.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.ls);
            writeln(fl);
            writeln(fl,'     Y intercept: zero');
            write(fl,'     R squared: ');
            extendWrite(fl,s.r2);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
            writeln(fl,'     R squared: undefined'); 
         End; {else}
         If (abs(s.corr)<>1) Then Begin
            write(fl,'T statistic: ');
            extendWrite(fl,s.ts);
            writeln(fl);
            write(fl,'F statistic: ');
            extendWrite(fl,s.f);
            writeln(fl);
            writeln(fl,'Degrees of Freedom: ',stpk.df);
         End {then}
         Else Begin
            writeln(fl,'T statistic: undefined');
            writeln(fl,'F statistic: undefined');
         End; {else}
         writeln(fl,'Number of Y contrasts > 0 : ',s.conypos);
         writeln(fl,'Number of Y contrasts < 0 : ',s.conyneg);
         writeln(fl,'Number of Y contrasts = 0 : ',s.cony0);
      End; {procedure flsts9}
   {**********************************************************************}
   {This procedure writes the stats given it to a file, omiting the stat
   that don't apply to the tenth plot.}
   {**********************************************************************}
   Procedure flsts10(Var s:statPak;Var fl:text);
      Begin
         writeln(fl,'Number of contrasts: ',s.number);
         write(fl,'Minimum of X,Y coordinates: ');
         extendWrite(fl,s.min[1]);
         write(fl,',');
         extendWrite(fl,s.min[2]);
         writeln(fl);
         write(fl,'Maximum of X,Y coordinates: ');
         extendWrite(fl,s.max[1]);
         write(fl,',');
         extendWrite(fl,s.max[2]);
         writeln(fl);
         write(fl,'Mean of X,Y coordinates: zero, ');
         extendWrite(fl,stpk.mean[2]);
         writeln(fl);
         write(fl,'Variance of X,Y coordinates: ');
         extendWrite(fl,s.variance[1]);
         write(fl,',');
         extendWrite(fl,s.variance[2]);
         writeln(fl);
         write(fl,'Covariance: ');
         extendWrite(fl,s.cov);
         writeln(fl);
         write(fl,'Pearson Product-Moment Correlation Coefficient: ');
         extendWrite(fl,s.corr);
         writeln(fl);
         writeln(fl);
      End; {procedure flsts10}
   {**********************************************************************}
   {This procedure dumps the stats given it to a file with one mean having
   a zero value and all the incercepts having zero values.}
   {**********************************************************************}
   Procedure flsts11(Var s:statPak;Var fl:text);
      Begin
         writeln(fl,'Number of contrasts: ',s.number);
         write(fl,'Minimum of X,Y coordinates: ');
         extendWrite(fl,s.min[1]);
         write(fl,',');
         extendwrite(fl,s.min[2]);
         writeln(fl);
         write(fl,'Maximum of X,Y coordinates: ');
         extendWrite(fl,s.max[1]);
         write(fl,',');
         extendWrite(fl,s.max[2]);
         writeln(fl);
         write(fl,'Mean of X,Y coordinates: ');
         extendWrite(fl,s.mean[1]);
         write(fl,',');
         extendWrite(fl,s.mean[2]);
         writeln(fl);
         write(fl,'Variance of X,Y coordinates: ');
         extendWrite(fl,s.variance[1]);
         write(fl,',');
         extendWrite(fl,s.variance[2]);
         writeln(fl);
         write(fl,'Covariance: ');
         extendWrite(fl,s.cov);
         writeln(fl);
         write(fl,'Pearson Product-Moment Correlation Coefficient: ');
         extendWrite(fl,s.corr);
         writeln(fl);
         writeln(fl,'Reduced Major Axis.');
         If (s.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.rma);
            writeln(fl);
            write(fl,'     Y intercept: ');
            extendWrite(fl,s.rmay);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
         End; {else}
         writeln(fl,'Major Axis.');
         write(fl,'     Slope: ');
         extendWrite(fl,s.ma);
         writeln(fl);
         write(fl,'     Y intercept: ');
         extendWrite(fl,s.may);
         writeln(fl);
         writeln(fl,'Least Squares Regression.');
         If (s.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.ls);
            writeln(fl);
            write(fl,'     Y intercept: ');
            extendWrite(fl,s.lsy);
            writeln(fl);
            write(fl,'     R squared: ');
            extendWrite(fl,s.r2);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
            writeln(fl,'     R squared: undefined'); 
         End; {else}
         If (abs(s.corr)<>1) Then Begin
            write(fl,'T statistic: ');
            extendWrite(fl,s.ts);
            writeln(fl);
            write(fl,'F statistic: ');
            extendWrite(fl,s.f);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'T statistic: undefined');
            writeln(fl,'F statistic: undefined');
         End; {else}
      End; {procedure flsts11}
   {**********************************************************************}
   {Here is the most general procedure that dumps stats to a file.  It
   outputs what we consider to be the normal or default set of stats to the
   file.  It is called to do 8 plots worth of stats.}
   {**********************************************************************}
   Procedure flsts(Var s:statPak;Var fl:text;contrasts : boolean);
      Begin
         If contrasts
         Then writeln(fl,'Number of contrasts: ',s.number)
         Else writeln(fl,'Number of data points: ',s.number);
         write(fl,'Minimum of X,Y coordinates: ');
         extendWrite(fl,s.min[1]);
         write(fl,',');
         extendWrite(fl,s.min[2]);
         writeln(fl);
         write(fl,'Maximum of X,Y coordinates: ');
         extendWrite(fl,s.max[1]);
         write(fl,',');
         extendWrite(fl,s.max[2]);
         writeln(fl);
         write(fl,'Mean of X,Y coordinates: ');
         extendWrite(fl,s.mean[1]);
         write(fl,',');
         extendWrite(fl,s.mean[2]);
         writeln(fl);
         write(fl,'Variance of X,Y coordinates: ');
         extendWrite(fl,s.variance[1]);
         write(fl,',');
         extendWrite(fl,s.variance[2]);
         writeln(fl);
         write(fl,'Covariance: ');
         extendWrite(fl,s.cov);
         writeln(fl);
         write(fl,'Pearson Product-Moment Correlation Coefficient: ');
         extendWrite(fl,s.corr);
         writeln(fl);
         writeln(fl,'Reduced Major Axis.');
         If (s.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.rma);
            writeln(fl);
            write(fl,'     Y intercept: ');
            extendWrite(fl,s.rmay);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
         End; {else}
         writeln(fl,'Major Axis.');
         write(fl,'     Slope: ');
         extendWrite(fl,s.ma);
         writeln(fl);
         write(fl,'     Y intercept: ');
         extendWrite(fl,s.may);
         writeln(fl);
         writeln(fl,'Least Squares Regression.');
         If (stpk.lsregok) Then Begin
            write(fl,'     Slope: ');
            extendWrite(fl,s.ls);
            writeln(fl);
            write(fl,'     Y intercept: ');
            extendWrite(fl,s.lsy);
            writeln(fl);
            write(fl,'     R squared: ');
            extendWrite(fl,s.r2);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'     Slope: undefined');
            writeln(fl,'     Y intercept: undefined');
            writeln(fl,'     R squared: undefined');
         End; {else}
         If (abs(s.corr)<>1) Then Begin
            write(fl,'T statistic: ');
            extendWrite(fl,s.ts);
            writeln(fl);
            write(fl,'F statistic: ');
            extendWrite(fl,s.f);
            writeln(fl);
         End {then}
         Else Begin
            writeln(fl,'T statistic: undefined');
            writeln(fl,'F statistic: undefined');
         End; {Else}
         writeln(fl,'Degrees of Freedom: ',s.df);
      End; {procedure flsts}
   {**********************************************************************}
   Begin {fileStats}
      tipStat(nd,stpk);
      writeln(fl);
      writeln(fl,'The Tip data statistics with ');
      If (stpk.n1=1)
      Then Begin
         writeln(fl,'X = Trait1');
         writeln(fl,'Y = Trait2');
      End
      Else Begin
         writeln(fl,'X = Trait2');
         writeln(fl,'Y = Trait1');
      End; {else}
      flsts(stpk,fl,false);
      statStdVsD(nd,1,stpk);
      writeln(fl);
      writeln(fl,'SCREEN 1:');
      writeln(fl,'The statistics for the absolute value of the standardized ',
                 'contrast for ');
      writeln(fl,'    Trait1 vs. its standard deviation (= square root of ',
                 'sum of ');
      writeln(fl,'    corrected branch lengths)');
      flsts(stpk,fl,true);
      statStdVsD(nd,2,stpk);
      writeln(fl);
      writeln(fl,'SCREEN 2:');
      writeln(fl,'The statistics for the absolute value of the standardized ',
                 'contrast for ');
      writeln(fl,'    Trait2 vs. its standard deviation (= square root of ',
                 'sum of ');
      writeln(fl,'    corrected branch lengths)');
      flsts(stpk,fl,true);
      If star
      Then
         writeln(fl,'The tree is a star.  Skipping statistics regarding',
                    ' estimated nodal values.')
      Else Begin
         If toolong Then Begin
            writeln(fl,'Some branch lengths too large (greater than ',
                     '1,000,000,000).');
            writeln(fl,'Skipping statistics regarding estimated nodal ',
                      'values.');
         End {then}
         Else Begin
            statNodeCon(nd,1,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 3:');
            writeln(fl,'The statistics for the absolute value of the standardized ',
                         'contrast for ');
            writeln(fl,'    Trait1 vs. the estimated nodal value.');
            flsts(stpk,fl,true);
            statNodeCon(nd,2,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 4:');
            writeln(fl,'The statistics for the absolute value of the standardized ',
                         'contrast for ');
            writeln(fl,'    Trait2 vs. the estimated nodal value.');
            flsts(stpk,fl,true);
            temp3CorrectedHeight(nd,nd,0);
            statHeightCon(nd,1,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 5:');
            writeln(fl,'The statistics for the absolute value of the standardized ',
                         'contrast for ');
            writeln(fl,'    Trait1 vs. the corrected height of the base node of the contrast.');
            flsts(stpk,fl,true);
            statHeightCon(nd,2,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 6:');
            writeln(fl,'The statistics for the absolute value of the standardized ',
                       'contrast for ');
            writeln(fl,'    Trait2 vs. the corrected height of the base node of the contrast.');
            flsts(stpk,fl,true);
            temp3CorrectedHeight(nd,nd,0);
            statHeightNode(nd,1,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 7:');
            writeln(fl,'The statistics for estimated base node value for Trait1 vs. its base');
            writeln(fl,'    node''s corrected height.');
            flsts(stpk,fl,false);
            statHeightNode(nd,2,stpk);
            writeln(fl);
            writeln(fl,'SCREEN 8:');
            writeln(fl,'The statistics for estimated base node value for Trait2 vs. its base');
            writeln(fl,'    node''s corrected height.');
            flsts(stpk,fl,false);
         End; {else}
      End; {else}
      If (constantIndex > 0)
      Then Begin
         writeln(fl);
         write(fl,'Trait ',constantIndex,' is constant, skipping statistics regarding');
         writeln(fl,' tip values.');
      End
      Else Begin
         statCon(nd,stpk);
         writeln(fl);
         writeln(fl,'SCREEN 9:');
         writeln(fl,'The positivized standardized contrasts statistics with ');
         If (stpk.n1=1)
         Then Begin
            writeln(fl,'X = Trait1');
            writeln(fl,'Y = Trait2');
         End
         Else Begin
            writeln(fl,'X = Trait2');
            writeln(fl,'Y = Trait1');
         End; {else}
         flsts9(stpk,fl);
         { protect these important values from overwriting by statRoot}
         t:=stpk.n1;
         t2:=stpk.n2;
         m:=stpk.ls;
         y0:=stpk.lsy;
         statRoot(nd,stpk,stpk.ls,stpk.n1,stpk.n2);
         writeln(fl);
         writeln(fl,'SCREEN 9A:');
         writeln(fl,'The statistics pertaining to estimated root node values.');
         writeln(fl);
         writeln(fl,'Note: the following statistics can also be saved as a .RT file.');
         writeln(fl);
         dumpStat9A(fl,stpk);
         writeln(fl);
         writeln(fl);
         writeln(fl,'SCREEN 9B:');
         writeln(fl);
         writeln(fl,'Prediction intervals for hypothetical species added at the');
         writeln(fl,'root, with a branch length equal to the mean height of');
         writeln(fl,'of the tree.');
         writeln(fl);
         dumpStat9B(fl,stpk,nd);
         writeln(fl);
         writeln(fl);
         tempResidual(nd,m,y0,t,t2);
         statResidualX(nd,t,stpk);
         If star
         Then  
            writeln(fl,'The tree is a star.  Skipping statistics regarding',
                        ' estimated nodal values.')
         Else Begin
            writeln(fl,'SCREEN 10:');
            writeln(fl,'The statistics of the residuals vs. X plot (from Screen 9).');
            flsts10(stpk,fl);
            temp3CorrectedHeight(nd,nd,0);
            tempAbsResidual(nd,m,y0,t,t2);
            If (not toolong) Then Begin
               statResidualHeight(nd,stpk);
               writeln(fl);
               writeln(fl,'SCREEN 11:');
               writeln(fl,'The statistics of the absolute value of residuals verses the');
               writeln(fl,'    height of their base node.');
               flsts11(stpk,fl);
            End; {then}
         End; {else}
      End; {else}
      writeln(fl);
      writeln(fl);
      writeln(fl,'The following is ',ficName);
      {The following lines write a header for a table that is written by
      another procedure}
      writeln(fl,'                                                       ',
              '                                                          ',
              '                            UnCorrected       Corrected   ',
              '     First',
              '                        Second');
      writeln(fl,'                                                       ',
              '                                                          ',
              '                            Average           Average      ',
              '    Daughter  Corrected          ',
              'Daughter  Corrected');
      writeln(fl,'Contrast Node  Contrast          Contrast          S.D.',
              '              Node              Node              ',
              'UnCorrected       Corrected         Daughter          ',
              'Daughter         Branch    First Daughter     Branch    ',
              'Second Daughter');
      writeln(fl,'Name     Name  Trait 1           Trait 2               ',
              '              Trait 1           Trait 2           ',
              'Node Height       Node Height       Node Height       ',
              'Node Height      Name      Branch Length',
              '      Name      Branch Length');
   End;{procedure fileStats}


{**************************************************************************}
{This procedure implements the menu for '(A)nalyze by Felsenstein's contrasts.'
nd must be the root, and the filename is expected to have the extension
intact.}
{**************************************************************************}
Procedure felsenstein(nd:nodeptr; liveFile:boolean; fileName:flString);
   Var
      fl:text;
      specialNode :nodePtr;
      numTips:integer;
      constantIndex : integer;
      fln,flnm:flString;
      ch:char;
      zeroTip:boolean;
      bail:boolean;
      conOK :boolean;
      wholeClade : boolean;
      allZero,tooLong:boolean;
   Begin
    {  width1 := defaultWidth1;
      width2 := defaultWidth2;  }
      If not livefile Then Begin
         clrscr;
         writeln('************************ Analyze with Felsenstein Module ',
                   '***********************');
         writeln('You must have a tree in memory to analyze the tree with',
                  ' Felsenstein''s contrasts.');
         writeln('To put a tree in memory you must read it in from a file, or');
         writeln('you must create a new one.');
         ret('Please press any key to return to the Main Menu.');  
      End {then}
      Else Begin
         zeroTip:=false;
         conOK := true;
         numTips:=0;
         checkTwo(nd,zerotip,numtips);
         {contrastCheck(nd,conOK);}    {Doesn't work PEM 10 Oct 1996}
         If (zeroTip Or Not(conOK)) Then Begin
            clrscr;
            writeln('************************ Analyze with Felsenstein Module ',
                      '***********************');
            writeln;
            writeln;
            writeln('To analyze with Felsenstein''s contrasts, no more than',
                      ' one of two sister');
            writeln('tip branches can have zero length!');
            writeln;
            writeln;
            writeln('Internal branches CAN have 0 length, which is how you',
                      ' create polytomies.');
            writeln;
            writeln;
            ret('Please press any key to return to the Main Menu.');
         End {then}
         Else
            If numtips<3 Then Begin
               clrscr;
               writeln('************************ Analyze with Felsenstein Module ',
                         '***********************');
               writeln;
               writeln;
               writeln('This implementation of Felsenstein''s method requires');
               writeln('there to be more than 2 tips.');
               writeln;
               writeln;
               ret('Please press any key to return to the Main Menu.');
            End {then}
            Else Begin
               fln:=fileName;
               delete(fileName,length(fileName)-3,4);
               Repeat
                  Repeat
                     clrscr;
                     writeln('** Analyze with Felsenstein''s Phylogenetically Independent Contrasts Module **');
                     writeln;
                     writeln('Please choose one of the following:');
                     writeln;
                     writeln('(V)iew diagnostic graphs and statistics (screens 1-8).');
                     writeln;
                     writeln('(W) view regression graphs and statistics (screens 9-11).');
                     writeln('      This includes option to save regression confidence and prediction');
                     writeln('      intervals.');
                     writeln;
                     writeln('(S)ave .STA file {screen display statistics and contrasts}.');
                     writeln('      This writes the statistics from both sets of stats (V and W in');
                     writeln('      sequence).');
                     writeln;
                     writeln('(J)ust save .FIC file {just contrasts}.');
                     writeln('      Use this command to produce a clean ASCII file containing only');
                     writeln('      contrasts, their standard deviations, their names (by tips and by ');
                     writeln('      base nodes), etc., which you can then enter into your favorite');
                     writeln('      graphics or statistics program.  If your overall analysis (e.g.,');
                     writeln('      multiple regression or principal components analysis) involves more');
                     writeln('      than two traits, then you will need to save more than one .FIC file');
                     writeln('      and then merge them into the external statistics program.');
                     writeln;
                     writeln('(R)eturn to Main Menu.');
                     ch:=UpCase(readkey);
                  Until (ch in ['V','W','S','J','R']);
                  If ch='V' Then showDiagnostics(nd,fln);
                  If ch='W' Then Begin
                     computeCon(nd);      {make sure contrasts are set up!}
                     showRegressionStats(nd,fln);
                  End; 
                  If ch='S' Then Begin
                     clrscr;
                     flnm:=filename+'.STA';
                     writeln('Saving data to file ',flnm);
                     computeCon(nd);
                     assign(fl,flnm);
                     rewrite(fl);
                     allZero:=true;
                     checkAllZero(nd,allzero);
                     tooLong:=false;
                     {checkBranchTooLong(nd,toolong); }
                     checkConstantTips(nd,constantIndex);
                     writeln(fl,'File = ',flnm);
                     fileStats(nd,fl,filename+'.FIC',constantIndex,allzero,toolong);
                     temp2Height(nd,0);
                     tempAveHeight(nd,0);
                     temp3CorrectedHeight(nd,nd,0);
                     temp4AveCorrectedHeight(nd,nd,0);
                     fileContrasts(nd,fl);
                     close(fl);
                     ret('Finished.  Press any key to continue.');
                  End; {then}
                  If ch='J' Then Begin
                     flnm:=filename+'.FIC';
                     clrscr;
                     writeln;
                     writeln('Saving data to file ',flnm);
                     listFICLines;
                     computeCon(nd);
                     assign(fl,flnm);
                     rewrite(fl);
                     temp2Height(nd,0);
                     tempAveHeight(nd,0);
                     temp3CorrectedHeight(nd,nd,0);
                     temp4AveCorrectedHeight(nd,nd,0);
                     fileContrasts(nd,fl);
                     close(fl);
                     ret('Finished.  Press any key to continue.');
                  End; {then}
               Until (ch='R');
          End; {else}
        End; {else}
    End; {procedure felsenstein}

End. {unit t_fels_m}
